@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -29,33 +29,27 @@
 /*************************************************************************/
 
 #include "camera_matrix.h"
-#include "math_funcs.h"
-#include "print_string.h"
+
+#include "core/math/math_funcs.h"
+#include "core/print_string.h"
 
 void CameraMatrix::set_identity() {
-
 	for (int i = 0; i < 4; i++) {
-
 		for (int j = 0; j < 4; j++) {
-
 			matrix[i][j] = (i == j) ? 1 : 0;
 		}
 	}
 }
 
 void CameraMatrix::set_zero() {
-
 	for (int i = 0; i < 4; i++) {
-
 		for (int j = 0; j < 4; j++) {
-
 			matrix[i][j] = 0;
 		}
 	}
 }
 
 Plane CameraMatrix::xform4(const Plane &p_vec4) const {
-
 	Plane ret;
 
 	ret.normal.x = matrix[0][0] * p_vec4.normal.x + matrix[1][0] * p_vec4.normal.y + matrix[2][0] * p_vec4.normal.z + matrix[3][0] * p_vec4.d;
@@ -66,7 +60,6 @@ Plane CameraMatrix::xform4(const Plane &p_vec4) const {
 }
 
 void CameraMatrix::set_perspective(real_t p_fovy_degrees, real_t p_aspect, real_t p_z_near, real_t p_z_far, bool p_flip_fov) {
-
 	if (p_flip_fov) {
 		p_fovy_degrees = get_fovy(p_fovy_degrees, 1.0 / p_aspect);
 	}
@@ -144,7 +137,7 @@ void CameraMatrix::set_for_hmd(int p_eye, real_t p_aspect, real_t p_intraocular_
 	f3 *= p_oversample;
 
 	// always apply KEEP_WIDTH aspect ratio
-	f3 *= p_aspect;
+	f3 /= p_aspect;
 
 	switch (p_eye) {
 		case 1: { // left eye
@@ -159,7 +152,6 @@ void CameraMatrix::set_for_hmd(int p_eye, real_t p_aspect, real_t p_intraocular_
 };
 
 void CameraMatrix::set_orthogonal(real_t p_left, real_t p_right, real_t p_bottom, real_t p_top, real_t p_znear, real_t p_zfar) {
-
 	set_identity();
 
 	matrix[0][0] = 2.0 / (p_right - p_left);
@@ -172,7 +164,6 @@ void CameraMatrix::set_orthogonal(real_t p_left, real_t p_right, real_t p_bottom
 }
 
 void CameraMatrix::set_orthogonal(real_t p_size, real_t p_aspect, real_t p_znear, real_t p_zfar, bool p_flip_fov) {
-
 	if (!p_flip_fov) {
 		p_size *= p_aspect;
 	}
@@ -181,6 +172,9 @@ void CameraMatrix::set_orthogonal(real_t p_size, real_t p_aspect, real_t p_znear
 }
 
 void CameraMatrix::set_frustum(real_t p_left, real_t p_right, real_t p_bottom, real_t p_top, real_t p_near, real_t p_far) {
+	ERR_FAIL_COND(p_right <= p_left);
+	ERR_FAIL_COND(p_top <= p_bottom);
+	ERR_FAIL_COND(p_far <= p_near);
 
 	real_t *te = &matrix[0][0];
 	real_t x = 2 * p_near / (p_right - p_left);
@@ -209,8 +203,15 @@ void CameraMatrix::set_frustum(real_t p_left, real_t p_right, real_t p_bottom, r
 	te[15] = 0;
 }
 
-real_t CameraMatrix::get_z_far() const {
+void CameraMatrix::set_frustum(real_t p_size, real_t p_aspect, Vector2 p_offset, real_t p_near, real_t p_far, bool p_flip_fov) {
+	if (!p_flip_fov) {
+		p_size *= p_aspect;
+	}
 
+	set_frustum(-p_size / 2 + p_offset.x, +p_size / 2 + p_offset.x, -p_size / p_aspect / 2 + p_offset.y, +p_size / p_aspect / 2 + p_offset.y, p_near, p_far);
+}
+
+real_t CameraMatrix::get_z_far() const {
 	const real_t *matrix = (const real_t *)this->matrix;
 	Plane new_plane = Plane(matrix[3] - matrix[2],
 			matrix[7] - matrix[6],
@@ -223,7 +224,6 @@ real_t CameraMatrix::get_z_far() const {
 	return new_plane.d;
 }
 real_t CameraMatrix::get_z_near() const {
-
 	const real_t *matrix = (const real_t *)this->matrix;
 	Plane new_plane = Plane(matrix[3] + matrix[2],
 			matrix[7] + matrix[6],
@@ -234,8 +234,7 @@ real_t CameraMatrix::get_z_near() const {
 	return new_plane.d;
 }
 
-void CameraMatrix::get_viewport_size(real_t &r_width, real_t &r_height) const {
-
+Vector2 CameraMatrix::get_viewport_half_extents() const {
 	const real_t *matrix = (const real_t *)this->matrix;
 	///////--- Near Plane ---///////
 	Plane near_plane = Plane(matrix[3] + matrix[2],
@@ -260,12 +259,10 @@ void CameraMatrix::get_viewport_size(real_t &r_width, real_t &r_height) const {
 	Vector3 res;
 	near_plane.intersect_3(right_plane, top_plane, &res);
 
-	r_width = res.x;
-	r_height = res.y;
+	return Vector2(res.x, res.y);
 }
 
 bool CameraMatrix::get_endpoints(const Transform &p_transform, Vector3 *p_8points) const {
-
 	Vector<Plane> planes = get_projection_planes(Transform());
 	const Planes intersections[8][3] = {
 		{ PLANE_FAR, PLANE_LEFT, PLANE_TOP },
@@ -279,7 +276,6 @@ bool CameraMatrix::get_endpoints(const Transform &p_transform, Vector3 *p_8point
 	};
 
 	for (int i = 0; i < 8; i++) {
-
 		Vector3 point;
 		bool res = planes[intersections[i][0]].intersect_3(planes[intersections[i][1]], planes[intersections[i][2]], &point);
 		ERR_FAIL_COND_V(!res, false);
@@ -290,11 +286,10 @@ bool CameraMatrix::get_endpoints(const Transform &p_transform, Vector3 *p_8point
 }
 
 Vector<Plane> CameraMatrix::get_projection_planes(const Transform &p_transform) const {
-
 	/** Fast Plane Extraction from combined modelview/projection matrices.
 	 * References:
-	 * http://www.markmorley.com/opengl/frustumculling.html
-	 * http://www2.ravensoft.com/users/ggribb/plane%20extraction.pdf
+	 * https://web.archive.org/web/20011221205252/http://www.markmorley.com/opengl/frustumculling.html
+	 * https://web.archive.org/web/20061020020112/http://www2.ravensoft.com/users/ggribb/plane%20extraction.pdf
 	 */
 
 	Vector<Plane> planes;
@@ -373,14 +368,12 @@ Vector<Plane> CameraMatrix::get_projection_planes(const Transform &p_transform)
 }
 
 CameraMatrix CameraMatrix::inverse() const {
-
 	CameraMatrix cm = *this;
 	cm.invert();
 	return cm;
 }
 
 void CameraMatrix::invert() {
-
 	int i, j, k;
 	int pvt_i[4], pvt_j[4]; /* Locations of pivot matrix */
 	real_t pvt_val; /* Value of current pivot element */
@@ -431,20 +424,26 @@ void CameraMatrix::invert() {
 
 		/** Divide column by minus pivot value **/
 		for (i = 0; i < 4; i++) {
-			if (i != k) matrix[i][k] /= (-pvt_val);
+			if (i != k) {
+				matrix[i][k] /= (-pvt_val);
+			}
 		}
 
 		/** Reduce the matrix **/
 		for (i = 0; i < 4; i++) {
 			hold = matrix[i][k];
 			for (j = 0; j < 4; j++) {
-				if (i != k && j != k) matrix[i][j] += hold * matrix[k][j];
+				if (i != k && j != k) {
+					matrix[i][j] += hold * matrix[k][j];
+				}
 			}
 		}
 
 		/** Divide row by pivot **/
 		for (j = 0; j < 4; j++) {
-			if (j != k) matrix[k][j] /= pvt_val;
+			if (j != k) {
+				matrix[k][j] /= pvt_val;
+			}
 		}
 
 		/** Replace pivot by reciprocal (at last we can touch it). **/
@@ -464,29 +463,29 @@ void CameraMatrix::invert() {
 		}
 
 		j = pvt_i[k]; /* Columns to swap correspond to pivot ROW */
-		if (j != k) /* If columns are different */
+		if (j != k) { /* If columns are different */
 			for (i = 0; i < 4; i++) {
 				hold = matrix[i][k];
 				matrix[i][k] = -matrix[i][j];
 				matrix[i][j] = hold;
 			}
+		}
 	}
 }
 
 CameraMatrix::CameraMatrix() {
-
 	set_identity();
 }
 
 CameraMatrix CameraMatrix::operator*(const CameraMatrix &p_matrix) const {
-
 	CameraMatrix new_matrix;
 
 	for (int j = 0; j < 4; j++) {
 		for (int i = 0; i < 4; i++) {
 			real_t ab = 0;
-			for (int k = 0; k < 4; k++)
+			for (int k = 0; k < 4; k++) {
 				ab += matrix[k][i] * p_matrix.matrix[j][k];
+			}
 			new_matrix.matrix[j][i] = ab;
 		}
 	}
@@ -495,75 +494,70 @@ CameraMatrix CameraMatrix::operator*(const CameraMatrix &p_matrix) const {
 }
 
 void CameraMatrix::set_light_bias() {
-
 	real_t *m = &matrix[0][0];
 
-	m[0] = 0.5,
-	m[1] = 0.0,
-	m[2] = 0.0,
-	m[3] = 0.0,
-	m[4] = 0.0,
-	m[5] = 0.5,
-	m[6] = 0.0,
-	m[7] = 0.0,
-	m[8] = 0.0,
-	m[9] = 0.0,
-	m[10] = 0.5,
-	m[11] = 0.0,
-	m[12] = 0.5,
-	m[13] = 0.5,
-	m[14] = 0.5,
+	m[0] = 0.5;
+	m[1] = 0.0;
+	m[2] = 0.0;
+	m[3] = 0.0;
+	m[4] = 0.0;
+	m[5] = 0.5;
+	m[6] = 0.0;
+	m[7] = 0.0;
+	m[8] = 0.0;
+	m[9] = 0.0;
+	m[10] = 0.5;
+	m[11] = 0.0;
+	m[12] = 0.5;
+	m[13] = 0.5;
+	m[14] = 0.5;
 	m[15] = 1.0;
 }
 
 void CameraMatrix::set_light_atlas_rect(const Rect2 &p_rect) {
-
 	real_t *m = &matrix[0][0];
 
-	m[0] = p_rect.size.width,
-	m[1] = 0.0,
-	m[2] = 0.0,
-	m[3] = 0.0,
-	m[4] = 0.0,
-	m[5] = p_rect.size.height,
-	m[6] = 0.0,
-	m[7] = 0.0,
-	m[8] = 0.0,
-	m[9] = 0.0,
-	m[10] = 1.0,
-	m[11] = 0.0,
-	m[12] = p_rect.position.x,
-	m[13] = p_rect.position.y,
-	m[14] = 0.0,
+	m[0] = p_rect.size.width;
+	m[1] = 0.0;
+	m[2] = 0.0;
+	m[3] = 0.0;
+	m[4] = 0.0;
+	m[5] = p_rect.size.height;
+	m[6] = 0.0;
+	m[7] = 0.0;
+	m[8] = 0.0;
+	m[9] = 0.0;
+	m[10] = 1.0;
+	m[11] = 0.0;
+	m[12] = p_rect.position.x;
+	m[13] = p_rect.position.y;
+	m[14] = 0.0;
 	m[15] = 1.0;
 }
 
 CameraMatrix::operator String() const {
-
 	String str;
-	for (int i = 0; i < 4; i++)
-		for (int j = 0; j < 4; j++)
+	for (int i = 0; i < 4; i++) {
+		for (int j = 0; j < 4; j++) {
 			str += String((j > 0) ? ", " : "\n") + rtos(matrix[i][j]);
+		}
+	}
 
 	return str;
 }
 
 real_t CameraMatrix::get_aspect() const {
-
-	real_t w, h;
-	get_viewport_size(w, h);
-	return w / h;
+	Vector2 vp_he = get_viewport_half_extents();
+	return vp_he.x / vp_he.y;
 }
 
 int CameraMatrix::get_pixels_per_meter(int p_for_pixel_width) const {
-
 	Vector3 result = xform(Vector3(1, 0, -1));
 
 	return int((result.x * 0.5 + 0.5) * p_for_pixel_width);
 }
 
 bool CameraMatrix::is_orthogonal() const {
-
 	return matrix[3][3] == 1.0;
 }
 
@@ -591,7 +585,6 @@ real_t CameraMatrix::get_fov() const {
 }
 
 void CameraMatrix::make_scale(const Vector3 &p_scale) {
-
 	set_identity();
 	matrix[0][0] = p_scale.x;
 	matrix[1][1] = p_scale.y;
@@ -599,7 +592,6 @@ void CameraMatrix::make_scale(const Vector3 &p_scale) {
 }
 
 void CameraMatrix::scale_translate_to_fit(const AABB &p_aabb) {
-
 	Vector3 min = p_aabb.position;
 	Vector3 max = p_aabb.position + p_aabb.size;
 
@@ -625,7 +617,6 @@ void CameraMatrix::scale_translate_to_fit(const AABB &p_aabb) {
 }
 
 CameraMatrix::operator Transform() const {
-
 	Transform tr;
 	const real_t *m = &matrix[0][0];
 
@@ -649,7 +640,6 @@ CameraMatrix::operator Transform() const {
 }
 
 CameraMatrix::CameraMatrix(const Transform &p_transform) {
-
 	const Transform &tr = p_transform;
 	real_t *m = &matrix[0][0];
 
