@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -31,9 +31,9 @@
 #ifndef METHOD_PTRCALL_H
 #define METHOD_PTRCALL_H
 
-#include "math_2d.h"
-#include "typedefs.h"
-#include "variant.h"
+#include "core/math/transform_2d.h"
+#include "core/typedefs.h"
+#include "core/variant.h"
 
 #ifdef PTRCALL_ENABLED
 
@@ -141,28 +141,22 @@ MAKE_PTRARG_BY_REFERENCE(Variant);
 
 template <class T>
 struct PtrToArg<T *> {
-
 	_FORCE_INLINE_ static T *convert(const void *p_ptr) {
-
 		return const_cast<T *>(reinterpret_cast<const T *>(p_ptr));
 	}
 
 	_FORCE_INLINE_ static void encode(T *p_var, void *p_ptr) {
-
 		*((T **)p_ptr) = p_var;
 	}
 };
 
 template <class T>
 struct PtrToArg<const T *> {
-
 	_FORCE_INLINE_ static const T *convert(const void *p_ptr) {
-
 		return reinterpret_cast<const T *>(p_ptr);
 	}
 
 	_FORCE_INLINE_ static void encode(T *p_var, void *p_ptr) {
-
 		*((T **)p_ptr) = p_var;
 	}
 };
@@ -171,7 +165,7 @@ struct PtrToArg<const T *> {
 
 #define MAKE_VECARG(m_type)                                                                      \
 	template <>                                                                                  \
-	struct PtrToArg<Vector<m_type> > {                                                           \
+	struct PtrToArg<Vector<m_type>> {                                                            \
 		_FORCE_INLINE_ static Vector<m_type> convert(const void *p_ptr) {                        \
 			const PoolVector<m_type> *dvs = reinterpret_cast<const PoolVector<m_type> *>(p_ptr); \
 			Vector<m_type> ret;                                                                  \
@@ -180,7 +174,7 @@ struct PtrToArg<const T *> {
 			{                                                                                    \
 				PoolVector<m_type>::Read r = dvs->read();                                        \
 				for (int i = 0; i < len; i++) {                                                  \
-					ret[i] = r[i];                                                               \
+					ret.write[i] = r[i];                                                         \
 				}                                                                                \
 			}                                                                                    \
 			return ret;                                                                          \
@@ -207,13 +201,57 @@ struct PtrToArg<const T *> {
 			{                                                                                    \
 				PoolVector<m_type>::Read r = dvs->read();                                        \
 				for (int i = 0; i < len; i++) {                                                  \
-					ret[i] = r[i];                                                               \
+					ret.write[i] = r[i];                                                         \
 				}                                                                                \
 			}                                                                                    \
 			return ret;                                                                          \
 		}                                                                                        \
 	}
 
+#define MAKE_VECARG_ALT(m_type, m_type_alt)                                                      \
+	template <>                                                                                  \
+	struct PtrToArg<Vector<m_type_alt>> {                                                        \
+		_FORCE_INLINE_ static Vector<m_type_alt> convert(const void *p_ptr) {                    \
+			const PoolVector<m_type> *dvs = reinterpret_cast<const PoolVector<m_type> *>(p_ptr); \
+			Vector<m_type_alt> ret;                                                              \
+			int len = dvs->size();                                                               \
+			ret.resize(len);                                                                     \
+			{                                                                                    \
+				PoolVector<m_type>::Read r = dvs->read();                                        \
+				for (int i = 0; i < len; i++) {                                                  \
+					ret.write[i] = r[i];                                                         \
+				}                                                                                \
+			}                                                                                    \
+			return ret;                                                                          \
+		}                                                                                        \
+		_FORCE_INLINE_ static void encode(Vector<m_type_alt> p_vec, void *p_ptr) {               \
+			PoolVector<m_type> *dv = reinterpret_cast<PoolVector<m_type> *>(p_ptr);              \
+			int len = p_vec.size();                                                              \
+			dv->resize(len);                                                                     \
+			{                                                                                    \
+				PoolVector<m_type>::Write w = dv->write();                                       \
+				for (int i = 0; i < len; i++) {                                                  \
+					w[i] = p_vec[i];                                                             \
+				}                                                                                \
+			}                                                                                    \
+		}                                                                                        \
+	};                                                                                           \
+	template <>                                                                                  \
+	struct PtrToArg<const Vector<m_type_alt> &> {                                                \
+		_FORCE_INLINE_ static Vector<m_type_alt> convert(const void *p_ptr) {                    \
+			const PoolVector<m_type> *dvs = reinterpret_cast<const PoolVector<m_type> *>(p_ptr); \
+			Vector<m_type_alt> ret;                                                              \
+			int len = dvs->size();                                                               \
+			ret.resize(len);                                                                     \
+			{                                                                                    \
+				PoolVector<m_type>::Read r = dvs->read();                                        \
+				for (int i = 0; i < len; i++) {                                                  \
+					ret.write[i] = r[i];                                                         \
+				}                                                                                \
+			}                                                                                    \
+			return ret;                                                                          \
+		}                                                                                        \
+	}
 MAKE_VECARG(String);
 MAKE_VECARG(uint8_t);
 MAKE_VECARG(int);
@@ -221,18 +259,19 @@ MAKE_VECARG(float);
 MAKE_VECARG(Vector2);
 MAKE_VECARG(Vector3);
 MAKE_VECARG(Color);
+MAKE_VECARG_ALT(String, StringName);
 
 //for stuff that gets converted to Array vectors
 #define MAKE_VECARR(m_type)                                                    \
 	template <>                                                                \
-	struct PtrToArg<Vector<m_type> > {                                         \
+	struct PtrToArg<Vector<m_type>> {                                          \
 		_FORCE_INLINE_ static Vector<m_type> convert(const void *p_ptr) {      \
 			const Array *arr = reinterpret_cast<const Array *>(p_ptr);         \
 			Vector<m_type> ret;                                                \
 			int len = arr->size();                                             \
 			ret.resize(len);                                                   \
 			for (int i = 0; i < len; i++) {                                    \
-				ret[i] = (*arr)[i];                                            \
+				ret.write[i] = (*arr)[i];                                      \
 			}                                                                  \
 			return ret;                                                        \
 		}                                                                      \
@@ -253,7 +292,7 @@ MAKE_VECARG(Color);
 			int len = arr->size();                                             \
 			ret.resize(len);                                                   \
 			for (int i = 0; i < len; i++) {                                    \
-				ret[i] = (*arr)[i];                                            \
+				ret.write[i] = (*arr)[i];                                      \
 			}                                                                  \
 			return ret;                                                        \
 		}                                                                      \
@@ -265,7 +304,7 @@ MAKE_VECARR(Plane);
 
 #define MAKE_DVECARR(m_type)                                                       \
 	template <>                                                                    \
-	struct PtrToArg<PoolVector<m_type> > {                                         \
+	struct PtrToArg<PoolVector<m_type>> {                                          \
 		_FORCE_INLINE_ static PoolVector<m_type> convert(const void *p_ptr) {      \
 			const Array *arr = reinterpret_cast<const Array *>(p_ptr);             \
 			PoolVector<m_type> ret;                                                \
@@ -357,7 +396,7 @@ MAKE_STRINGCONV(StringName);
 MAKE_STRINGCONV_BY_REFERENCE(IP_Address);
 
 template <>
-struct PtrToArg<PoolVector<Face3> > {
+struct PtrToArg<PoolVector<Face3>> {
 	_FORCE_INLINE_ static PoolVector<Face3> convert(const void *p_ptr) {
 		const PoolVector<Vector3> *dvs = reinterpret_cast<const PoolVector<Vector3> *>(p_ptr);
 		PoolVector<Face3> ret;
