@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -30,17 +30,18 @@
 
 #include "translation.h"
 
-#include "io/resource_loader.h"
-#include "os/os.h"
-#include "project_settings.h"
+#include "core/io/resource_loader.h"
+#include "core/os/os.h"
+#include "core/project_settings.h"
 
-// ISO 639-1 language codes, with the addition of glibc locales with their
-// regional identifiers. This list must match the language names (in English)
-// of locale_names.
+// ISO 639-1 language codes (and a couple of three-letter ISO 639-2 codes),
+// with the addition of glibc locales with their regional identifiers.
+// This list must match the language names (in English) of locale_names.
 //
 // References:
 // - https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes
 // - https://lh.2xlibre.net/locales/
+// - https://iso639-3.sil.org/
 
 static const char *locale_list[] = {
 	"aa", //  Afar
@@ -78,6 +79,7 @@ static const char *locale_list[] = {
 	"ast_ES", //  Asturian (Spain)
 	"ayc_PE", //  Southern Aymara (Peru)
 	"ay_PE", //  Aymara (Peru)
+	"az", //  Azerbaijani
 	"az_AZ", //  Azerbaijani (Azerbaijan)
 	"be", //  Belarusian
 	"be_BY", //  Belarusian (Belarus)
@@ -95,6 +97,7 @@ static const char *locale_list[] = {
 	"bo", //  Tibetan
 	"bo_CN", //  Tibetan (China)
 	"bo_IN", //  Tibetan (India)
+	"br", //  Breton
 	"br_FR", //  Breton (France)
 	"brx_IN", //  Bodo (India)
 	"bs_BA", //  Bosnian (Bosnia and Herzegovina)
@@ -179,6 +182,7 @@ static const char *locale_list[] = {
 	"ff_SN", //  Fulah (Senegal)
 	"fi", //  Finnish
 	"fi_FI", //  Finnish (Finland)
+	"fil", //  Filipino
 	"fil_PH", //  Filipino (Philippines)
 	"fo_FO", //  Faroese (Faroe Islands)
 	"fr", //  French
@@ -195,6 +199,7 @@ static const char *locale_list[] = {
 	"gd_GB", //  Scottish Gaelic (United Kingdom)
 	"gez_ER", //  Geez (Eritrea)
 	"gez_ET", //  Geez (Ethiopia)
+	"gl", //  Galician
 	"gl_ES", //  Galician (Spain)
 	"gu_IN", //  Gujarati (India)
 	"gv_GB", //  Manx (United Kingdom)
@@ -227,9 +232,11 @@ static const char *locale_list[] = {
 	"ja", //  Japanese
 	"ja_JP", //  Japanese (Japan)
 	"kab_DZ", //  Kabyle (Algeria)
+	"ka", //  Georgian
 	"ka_GE", //  Georgian (Georgia)
 	"kk_KZ", //  Kazakh (Kazakhstan)
 	"kl_GL", //  Kalaallisut (Greenland)
+	"km", //  Central Khmer
 	"km_KH", //  Central Khmer (Cambodia)
 	"kn_IN", //  Kannada (India)
 	"kok_IN", //  Konkani (India)
@@ -257,13 +264,16 @@ static const char *locale_list[] = {
 	"mg_MG", //  Malagasy (Madagascar)
 	"mh_MH", //  Marshallese (Marshall Islands)
 	"mhr_RU", //  Eastern Mari (Russia)
-	"mi_NZ", //  Maori (New Zealand)
+	"mi", //  Māori
+	"mi_NZ", //  Māori (New Zealand)
 	"miq_NI", //  Mískito (Nicaragua)
 	"mk", //  Macedonian
 	"mk_MK", //  Macedonian (Macedonia)
+	"ml", //  Malayalam
 	"ml_IN", //  Malayalam (India)
 	"mni_IN", //  Manipuri (India)
 	"mn_MN", //  Mongolian (Mongolia)
+	"mr", //  Marathi
 	"mr_IN", //  Marathi (India)
 	"ms", //  Malay
 	"ms_MY", //  Malay (Malaysia)
@@ -293,6 +303,7 @@ static const char *locale_list[] = {
 	"om", //  Oromo
 	"om_ET", //  Oromo (Ethiopia)
 	"om_KE", //  Oromo (Kenya)
+	"or", //  Oriya
 	"or_IN", //  Oriya (India)
 	"os_RU", //  Ossetian (Russia)
 	"pa_IN", //  Panjabi (India)
@@ -326,6 +337,7 @@ static const char *locale_list[] = {
 	"sgs_LT", //  Samogitian (Lithuania)
 	"shs_CA", //  Shuswap (Canada)
 	"sid_ET", //  Sidamo (Ethiopia)
+	"si", //  Sinhala
 	"si_LK", //  Sinhala (Sri Lanka)
 	"sk", //  Slovak
 	"sk_SK", //  Slovak (Slovakia)
@@ -343,6 +355,7 @@ static const char *locale_list[] = {
 	"sq_MK", //  Albanian (Macedonia)
 	"sr", //  Serbian
 	"sr_Cyrl", //  Serbian (Cyrillic)
+	"sr_Latn", //  Serbian (Latin)
 	"sr_ME", //  Serbian (Montenegro)
 	"sr_RS", //  Serbian (Serbia)
 	"ss_ZA", //  Swati (South Africa)
@@ -357,6 +370,7 @@ static const char *locale_list[] = {
 	"ta_IN", //  Tamil (India)
 	"ta_LK", //  Tamil (Sri Lanka)
 	"tcy_IN", //  Tulu (India)
+	"te", //  Telugu
 	"te_IN", //  Telugu (India)
 	"tg_TJ", //  Tajik (Tajikistan)
 	"the_NP", //  Chitwania Tharu (Nepal)
@@ -373,7 +387,10 @@ static const char *locale_list[] = {
 	"tr_CY", //  Turkish (Cyprus)
 	"tr_TR", //  Turkish (Turkey)
 	"ts_ZA", //  Tsonga (South Africa)
+	"tt", //  Tatar
 	"tt_RU", //  Tatar (Russia)
+	"tzm", // Central Atlas Tamazight
+	"tzm_MA", // Central Atlas Tamazight (Marrocos)
 	"ug_CN", //  Uighur (China)
 	"uk", //  Ukrainian
 	"uk_UA", //  Ukrainian (Ukraine)
@@ -400,7 +417,7 @@ static const char *locale_list[] = {
 	"zh_SG", //  Chinese (Singapore)
 	"zh_TW", //  Chinese (Taiwan)
 	"zu_ZA", //  Zulu (South Africa)
-	0
+	nullptr
 };
 
 static const char *locale_names[] = {
@@ -439,6 +456,7 @@ static const char *locale_names[] = {
 	"Asturian (Spain)",
 	"Southern Aymara (Peru)",
 	"Aymara (Peru)",
+	"Azerbaijani",
 	"Azerbaijani (Azerbaijan)",
 	"Belarusian",
 	"Belarusian (Belarus)",
@@ -456,6 +474,7 @@ static const char *locale_names[] = {
 	"Tibetan",
 	"Tibetan (China)",
 	"Tibetan (India)",
+	"Breton",
 	"Breton (France)",
 	"Bodo (India)",
 	"Bosnian (Bosnia and Herzegovina)",
@@ -540,6 +559,7 @@ static const char *locale_names[] = {
 	"Fulah (Senegal)",
 	"Finnish",
 	"Finnish (Finland)",
+	"Filipino",
 	"Filipino (Philippines)",
 	"Faroese (Faroe Islands)",
 	"French",
@@ -556,6 +576,7 @@ static const char *locale_names[] = {
 	"Scottish Gaelic (United Kingdom)",
 	"Geez (Eritrea)",
 	"Geez (Ethiopia)",
+	"Galician",
 	"Galician (Spain)",
 	"Gujarati (India)",
 	"Manx (United Kingdom)",
@@ -588,9 +609,11 @@ static const char *locale_names[] = {
 	"Japanese",
 	"Japanese (Japan)",
 	"Kabyle (Algeria)",
+	"Georgian",
 	"Georgian (Georgia)",
 	"Kazakh (Kazakhstan)",
 	"Kalaallisut (Greenland)",
+	"Central Khmer",
 	"Central Khmer (Cambodia)",
 	"Kannada (India)",
 	"Konkani (India)",
@@ -618,13 +641,16 @@ static const char *locale_names[] = {
 	"Malagasy (Madagascar)",
 	"Marshallese (Marshall Islands)",
 	"Eastern Mari (Russia)",
-	"Maori (New Zealand)",
+	"Māori",
+	"Māori (New Zealand)",
 	"Mískito (Nicaragua)",
 	"Macedonian",
 	"Macedonian (Macedonia)",
+	"Malayalam",
 	"Malayalam (India)",
 	"Manipuri (India)",
 	"Mongolian (Mongolia)",
+	"Marathi",
 	"Marathi (India)",
 	"Malay",
 	"Malay (Malaysia)",
@@ -654,6 +680,7 @@ static const char *locale_names[] = {
 	"Oromo",
 	"Oromo (Ethiopia)",
 	"Oromo (Kenya)",
+	"Oriya",
 	"Oriya (India)",
 	"Ossetian (Russia)",
 	"Panjabi (India)",
@@ -687,6 +714,7 @@ static const char *locale_names[] = {
 	"Samogitian (Lithuania)",
 	"Shuswap (Canada)",
 	"Sidamo (Ethiopia)",
+	"Sinhala",
 	"Sinhala (Sri Lanka)",
 	"Slovak",
 	"Slovak (Slovakia)",
@@ -704,6 +732,7 @@ static const char *locale_names[] = {
 	"Albanian (Macedonia)",
 	"Serbian",
 	"Serbian (Cyrillic)",
+	"Serbian (Latin)",
 	"Serbian (Montenegro)",
 	"Serbian (Serbia)",
 	"Swati (South Africa)",
@@ -718,6 +747,7 @@ static const char *locale_names[] = {
 	"Tamil (India)",
 	"Tamil (Sri Lanka)",
 	"Tulu (India)",
+	"Telugu",
 	"Telugu (India)",
 	"Tajik (Tajikistan)",
 	"Chitwania Tharu (Nepal)",
@@ -734,7 +764,10 @@ static const char *locale_names[] = {
 	"Turkish (Cyprus)",
 	"Turkish (Turkey)",
 	"Tsonga (South Africa)",
+	"Tatar",
 	"Tatar (Russia)",
+	"Central Atlas Tamazight",
+	"Central Atlas Tamazight (Marrocos)",
 	"Uighur (China)",
 	"Ukrainian",
 	"Ukrainian (Ukraine)",
@@ -761,7 +794,7 @@ static const char *locale_names[] = {
 	"Chinese (Singapore)",
 	"Chinese (Taiwan)",
 	"Zulu (South Africa)",
-	0
+	nullptr
 };
 
 // Windows has some weird locale identifiers which do not honor the ISO 639-1
@@ -772,26 +805,22 @@ static const char *locale_names[] = {
 // - https://msdn.microsoft.com/en-us/library/windows/desktop/ms693062(v=vs.85).aspx
 
 static const char *locale_renames[][2] = {
-	{ "in", "id" }, //  Indonesian
-	{ "iw", "he" }, //  Hebrew
-	{ "no", "nb" }, //  Norwegian Bokmål
-	{ NULL, NULL }
+	{ "in", "id" }, // Indonesian
+	{ "iw", "he" }, // Hebrew
+	{ "no", "nb" }, // Norwegian Bokmål
+	{ "C", "en" }, // "C" is the simple/default/untranslated Computer locale.
+	// ASCII-only, English, no currency symbols. Godot treats this as "en".
+	// See https://unix.stackexchange.com/a/87763/164141 "The C locale is"...
+	{ nullptr, nullptr }
 };
 
-static String get_trimmed_locale(const String &p_locale) {
-
-	return p_locale.substr(0, 2);
-}
-
 ///////////////////////////////////////////////
 
 PoolVector<String> Translation::_get_messages() const {
-
 	PoolVector<String> msgs;
 	msgs.resize(translation_map.size() * 2);
 	int idx = 0;
 	for (const Map<StringName, StringName>::Element *E = translation_map.front(); E; E = E->next()) {
-
 		msgs.set(idx + 0, E->key());
 		msgs.set(idx + 1, E->get());
 		idx += 2;
@@ -801,12 +830,10 @@ PoolVector<String> Translation::_get_messages() const {
 }
 
 PoolVector<String> Translation::_get_message_list() const {
-
 	PoolVector<String> msgs;
 	msgs.resize(translation_map.size());
 	int idx = 0;
 	for (const Map<StringName, StringName>::Element *E = translation_map.front(); E; E = E->next()) {
-
 		msgs.set(idx, E->key());
 		idx += 1;
 	}
@@ -815,27 +842,23 @@ PoolVector<String> Translation::_get_message_list() const {
 }
 
 void Translation::_set_messages(const PoolVector<String> &p_messages) {
-
 	int msg_count = p_messages.size();
 	ERR_FAIL_COND(msg_count % 2);
 
 	PoolVector<String>::Read r = p_messages.read();
 
 	for (int i = 0; i < msg_count; i += 2) {
-
 		add_message(r[i + 0], r[i + 1]);
 	}
 }
 
 void Translation::set_locale(const String &p_locale) {
-
 	String univ_locale = TranslationServer::standardize_locale(p_locale);
 
 	if (!TranslationServer::is_locale_valid(univ_locale)) {
-		String trimmed_locale = get_trimmed_locale(univ_locale);
+		String trimmed_locale = TranslationServer::get_language_code(univ_locale);
 
-		ERR_EXPLAIN("Invalid locale: " + trimmed_locale);
-		ERR_FAIL_COND(!TranslationServer::is_locale_valid(trimmed_locale));
+		ERR_FAIL_COND_MSG(!TranslationServer::is_locale_valid(trimmed_locale), "Invalid locale: " + trimmed_locale + ".");
 
 		locale = trimmed_locale;
 	} else {
@@ -848,38 +871,36 @@ void Translation::set_locale(const String &p_locale) {
 }
 
 void Translation::add_message(const StringName &p_src_text, const StringName &p_xlated_text) {
-
 	translation_map[p_src_text] = p_xlated_text;
 }
 StringName Translation::get_message(const StringName &p_src_text) const {
+	if (get_script_instance()) {
+		return get_script_instance()->call("_get_message", p_src_text);
+	}
 
 	const Map<StringName, StringName>::Element *E = translation_map.find(p_src_text);
-	if (!E)
+	if (!E) {
 		return StringName();
+	}
 
 	return E->get();
 }
 
 void Translation::erase_message(const StringName &p_src_text) {
-
 	translation_map.erase(p_src_text);
 }
 
 void Translation::get_message_list(List<StringName> *r_messages) const {
-
 	for (const Map<StringName, StringName>::Element *E = translation_map.front(); E; E = E->next()) {
-
 		r_messages->push_back(E->key());
 	}
 }
 
 int Translation::get_message_count() const {
-
 	return translation_map.size();
 };
 
 void Translation::_bind_methods() {
-
 	ClassDB::bind_method(D_METHOD("set_locale", "locale"), &Translation::set_locale);
 	ClassDB::bind_method(D_METHOD("get_locale"), &Translation::get_locale);
 	ClassDB::bind_method(D_METHOD("add_message", "src_message", "xlated_message"), &Translation::add_message);
@@ -890,6 +911,8 @@ void Translation::_bind_methods() {
 	ClassDB::bind_method(D_METHOD("_set_messages"), &Translation::_set_messages);
 	ClassDB::bind_method(D_METHOD("_get_messages"), &Translation::_get_messages);
 
+	BIND_VMETHOD(MethodInfo(Variant::STRING, "_get_message", PropertyInfo(Variant::STRING, "src_message")));
+
 	ADD_PROPERTY(PropertyInfo(Variant::POOL_STRING_ARRAY, "messages", PROPERTY_HINT_NONE, "", PROPERTY_USAGE_NOEDITOR | PROPERTY_USAGE_INTERNAL), "_set_messages", "_get_messages");
 	ADD_PROPERTY(PropertyInfo(Variant::STRING, "locale"), "set_locale", "get_locale");
 }
@@ -901,13 +924,12 @@ Translation::Translation() :
 ///////////////////////////////////////////////
 
 bool TranslationServer::is_locale_valid(const String &p_locale) {
-
 	const char **ptr = locale_list;
 
 	while (*ptr) {
-
-		if (*ptr == p_locale)
+		if (*ptr == p_locale) {
 			return true;
+		}
 		ptr++;
 	}
 
@@ -915,13 +937,12 @@ bool TranslationServer::is_locale_valid(const String &p_locale) {
 }
 
 String TranslationServer::standardize_locale(const String &p_locale) {
-
 	// Replaces '-' with '_' for macOS Sierra-style locales
 	String univ_locale = p_locale.replace("-", "_");
 
 	// Handles known non-ISO locale names used e.g. on Windows
 	int idx = 0;
-	while (locale_renames[idx][0] != NULL) {
+	while (locale_renames[idx][0] != nullptr) {
 		if (locale_renames[idx][0] == univ_locale) {
 			univ_locale = locale_renames[idx][1];
 			break;
@@ -932,17 +953,35 @@ String TranslationServer::standardize_locale(const String &p_locale) {
 	return univ_locale;
 }
 
-void TranslationServer::set_locale(const String &p_locale) {
+String TranslationServer::get_language_code(const String &p_locale) {
+	ERR_FAIL_COND_V_MSG(p_locale.length() < 2, p_locale, "Invalid locale '" + p_locale + "'.");
+	// Most language codes are two letters, but some are three,
+	// so we have to look for a regional code separator ('_' or '-')
+	// to extract the left part.
+	// For example we get 'nah_MX' as input and should return 'nah'.
+	int split = p_locale.find("_");
+	if (split == -1) {
+		split = p_locale.find("-");
+	}
+	if (split == -1) { // No separator, so the locale is already only a language code.
+		return p_locale;
+	}
+	return p_locale.left(split);
+}
 
+void TranslationServer::set_locale(const String &p_locale) {
 	String univ_locale = standardize_locale(p_locale);
 
 	if (!is_locale_valid(univ_locale)) {
-		String trimmed_locale = get_trimmed_locale(univ_locale);
+		String trimmed_locale = get_language_code(univ_locale);
+		print_verbose(vformat("Unsupported locale '%s', falling back to '%s'.", p_locale, trimmed_locale));
 
-		ERR_EXPLAIN("Invalid locale: " + trimmed_locale);
-		ERR_FAIL_COND(!is_locale_valid(trimmed_locale));
-
-		locale = trimmed_locale;
+		if (!is_locale_valid(trimmed_locale)) {
+			ERR_PRINT(vformat("Unsupported locale '%s', falling back to 'en'.", trimmed_locale));
+			locale = "en";
+		} else {
+			locale = trimmed_locale;
+		}
 	} else {
 		locale = univ_locale;
 	}
@@ -955,18 +994,33 @@ void TranslationServer::set_locale(const String &p_locale) {
 }
 
 String TranslationServer::get_locale() const {
-
 	return locale;
 }
 
 String TranslationServer::get_locale_name(const String &p_locale) const {
-
-	if (!locale_name_map.has(p_locale)) return String();
+	if (!locale_name_map.has(p_locale)) {
+		return String();
+	}
 	return locale_name_map[p_locale];
 }
 
-Vector<String> TranslationServer::get_all_locales() {
+Array TranslationServer::get_loaded_locales() const {
+	Array locales;
+	for (const Set<Ref<Translation>>::Element *E = translations.front(); E; E = E->next()) {
+		const Ref<Translation> &t = E->get();
+		ERR_FAIL_COND_V(t.is_null(), Array());
+		String l = t->get_locale();
 
+		if (!locales.has(l)) {
+			locales.push_back(l);
+		}
+	}
+
+	locales.sort();
+	return locales;
+}
+
+Vector<String> TranslationServer::get_all_locales() {
 	Vector<String> locales;
 
 	const char **ptr = locale_list;
@@ -980,7 +1034,6 @@ Vector<String> TranslationServer::get_all_locales() {
 }
 
 Vector<String> TranslationServer::get_all_locale_names() {
-
 	Vector<String> locales;
 
 	const char **ptr = locale_names;
@@ -994,100 +1047,111 @@ Vector<String> TranslationServer::get_all_locale_names() {
 }
 
 void TranslationServer::add_translation(const Ref<Translation> &p_translation) {
-
 	translations.insert(p_translation);
 }
 void TranslationServer::remove_translation(const Ref<Translation> &p_translation) {
-
 	translations.erase(p_translation);
 }
 
 void TranslationServer::clear() {
-
 	translations.clear();
 };
 
 StringName TranslationServer::translate(const StringName &p_message) const {
+	// Match given message against the translation catalog for the project locale.
 
-	//translate using locale
-
-	if (!enabled)
+	if (!enabled) {
 		return p_message;
-
-	StringName res;
-	bool near_match = false;
-	const CharType *lptr = &locale[0];
-
-	for (const Set<Ref<Translation> >::Element *E = translations.front(); E; E = E->next()) {
-
-		const Ref<Translation> &t = E->get();
-		String l = t->get_locale();
-		if (lptr[0] != l[0] || lptr[1] != l[1])
-			continue; // locale not match
-
-		//near match
-		bool match = (l != locale);
-
-		if (near_match && !match)
-			continue; //only near-match once
-
-		StringName r = t->get_message(p_message);
-
-		if (!r)
-			continue;
-
-		res = r;
-
-		if (match)
-			break;
-		else
-			near_match = true;
 	}
 
-	if (!res) {
-		//try again with fallback
-		if (fallback.length() >= 2) {
+	ERR_FAIL_COND_V_MSG(locale.length() < 2, p_message, "Could not translate message as configured locale '" + locale + "' is invalid.");
 
-			const CharType *fptr = &fallback[0];
-			bool near_match = false;
-			for (const Set<Ref<Translation> >::Element *E = translations.front(); E; E = E->next()) {
+	// Locale can be of the form 'll_CC', i.e. language code and regional code,
+	// e.g. 'en_US', 'en_GB', etc. It might also be simply 'll', e.g. 'en'.
+	// To find the relevant translation, we look for those with locale starting
+	// with the language code, and then if any is an exact match for the long
+	// form. If not found, we fall back to a near match (another locale with
+	// same language code).
 
-				const Ref<Translation> &t = E->get();
-				String l = t->get_locale();
-				if (fptr[0] != l[0] || fptr[1] != l[1])
-					continue; // locale not match
+	// Note: ResourceLoader::_path_remap reproduces this locale near matching
+	// logic, so be sure to propagate changes there when changing things here.
 
-				//near match
-				bool match = (l != fallback);
+	StringName res;
+	String lang = get_language_code(locale);
+	bool near_match = false;
 
-				if (near_match && !match)
-					continue; //only near-match once
+	for (const Set<Ref<Translation>>::Element *E = translations.front(); E; E = E->next()) {
+		const Ref<Translation> &t = E->get();
+		ERR_FAIL_COND_V(t.is_null(), p_message);
+		String l = t->get_locale();
 
-				StringName r = t->get_message(p_message);
+		bool exact_match = (l == locale);
+		if (!exact_match) {
+			if (near_match) {
+				continue; // Only near-match once, but keep looking for exact matches.
+			}
+			if (get_language_code(l) != lang) {
+				continue; // Language code does not match.
+			}
+		}
 
-				if (!r)
-					continue;
+		StringName r = t->get_message(p_message);
+		if (!r) {
+			continue;
+		}
+		res = r;
 
-				res = r;
+		if (exact_match) {
+			break;
+		} else {
+			near_match = true;
+		}
+	}
 
-				if (match)
-					break;
-				else
-					near_match = true;
+	if (!res && fallback.length() >= 2) {
+		// Try again with the fallback locale.
+		String fallback_lang = get_language_code(fallback);
+		near_match = false;
+
+		for (const Set<Ref<Translation>>::Element *E = translations.front(); E; E = E->next()) {
+			const Ref<Translation> &t = E->get();
+			ERR_FAIL_COND_V(t.is_null(), p_message);
+			String l = t->get_locale();
+
+			bool exact_match = (l == fallback);
+			if (!exact_match) {
+				if (near_match) {
+					continue; // Only near-match once, but keep looking for exact matches.
+				}
+				if (get_language_code(l) != fallback_lang) {
+					continue; // Language code does not match.
+				}
+			}
+
+			StringName r = t->get_message(p_message);
+			if (!r) {
+				continue;
+			}
+			res = r;
+
+			if (exact_match) {
+				break;
+			} else {
+				near_match = true;
 			}
 		}
 	}
 
-	if (!res)
+	if (!res) {
 		return p_message;
+	}
 
 	return res;
 }
 
-TranslationServer *TranslationServer::singleton = NULL;
+TranslationServer *TranslationServer::singleton = nullptr;
 
 bool TranslationServer::_load_translations(const String &p_from) {
-
 	if (ProjectSettings::get_singleton()->has_setting(p_from)) {
 		PoolVector<String> translations = ProjectSettings::get_singleton()->get(p_from);
 
@@ -1097,11 +1161,10 @@ bool TranslationServer::_load_translations(const String &p_from) {
 			PoolVector<String>::Read r = translations.read();
 
 			for (int i = 0; i < tcount; i++) {
-
-				//print_line( "Loading translation from " + r[i] );
 				Ref<Translation> tr = ResourceLoader::load(r[i]);
-				if (tr.is_valid())
+				if (tr.is_valid()) {
 					add_translation(tr);
+				}
 			}
 		}
 		return true;
@@ -1111,29 +1174,28 @@ bool TranslationServer::_load_translations(const String &p_from) {
 }
 
 void TranslationServer::setup() {
-
 	String test = GLOBAL_DEF("locale/test", "");
 	test = test.strip_edges();
-	if (test != "")
+	if (test != "") {
 		set_locale(test);
-	else
+	} else {
 		set_locale(OS::get_singleton()->get_locale());
+	}
 	fallback = GLOBAL_DEF("locale/fallback", "en");
 #ifdef TOOLS_ENABLED
-
 	{
 		String options = "";
 		int idx = 0;
 		while (locale_list[idx]) {
-			if (idx > 0)
+			if (idx > 0) {
 				options += ",";
+			}
 			options += locale_list[idx];
 			idx++;
 		}
 		ProjectSettings::get_singleton()->set_custom_property_info("locale/fallback", PropertyInfo(Variant::STRING, "locale/fallback", PROPERTY_HINT_ENUM, options));
 	}
 #endif
-	//load translations
 }
 
 void TranslationServer::set_tool_translation(const Ref<Translation> &p_translation) {
@@ -1141,20 +1203,30 @@ void TranslationServer::set_tool_translation(const Ref<Translation> &p_translati
 }
 
 StringName TranslationServer::tool_translate(const StringName &p_message) const {
-
 	if (tool_translation.is_valid()) {
 		StringName r = tool_translation->get_message(p_message);
-
 		if (r) {
 			return r;
 		}
 	}
+	return p_message;
+}
 
+void TranslationServer::set_doc_translation(const Ref<Translation> &p_translation) {
+	doc_translation = p_translation;
+}
+
+StringName TranslationServer::doc_translate(const StringName &p_message) const {
+	if (doc_translation.is_valid()) {
+		StringName r = doc_translation->get_message(p_message);
+		if (r) {
+			return r;
+		}
+	}
 	return p_message;
 }
 
 void TranslationServer::_bind_methods() {
-
 	ClassDB::bind_method(D_METHOD("set_locale", "locale"), &TranslationServer::set_locale);
 	ClassDB::bind_method(D_METHOD("get_locale"), &TranslationServer::get_locale);
 
@@ -1166,18 +1238,17 @@ void TranslationServer::_bind_methods() {
 	ClassDB::bind_method(D_METHOD("remove_translation", "translation"), &TranslationServer::remove_translation);
 
 	ClassDB::bind_method(D_METHOD("clear"), &TranslationServer::clear);
+
+	ClassDB::bind_method(D_METHOD("get_loaded_locales"), &TranslationServer::get_loaded_locales);
 }
 
 void TranslationServer::load_translations() {
-
 	String locale = get_locale();
-	bool found = _load_translations("locale/translations"); //all
+	_load_translations("locale/translations"); //all
+	_load_translations("locale/translations_" + locale.substr(0, 2));
 
-	if (_load_translations("locale/translations_" + locale.substr(0, 2)))
-		found = true;
 	if (locale.substr(0, 2) != locale) {
-		if (_load_translations("locale/translations_" + locale))
-			found = true;
+		_load_translations("locale/translations_" + locale);
 	}
 }
 
@@ -1187,7 +1258,6 @@ TranslationServer::TranslationServer() :
 	singleton = this;
 
 	for (int i = 0; locale_list[i]; ++i) {
-
 		locale_name_map.insert(locale_list[i], String::utf8(locale_names[i]));
 	}
 }
