@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -31,11 +31,13 @@
 #ifndef RASTERIZERSTORAGEGLES3_H
 #define RASTERIZERSTORAGEGLES3_H
 
-#include "self_list.h"
+#include "core/self_list.h"
+#include "drivers/gles_common/rasterizer_asserts.h"
 #include "servers/visual/rasterizer.h"
 #include "servers/visual/shader_language.h"
 #include "shader_compiler_gles3.h"
 #include "shader_gles3.h"
+
 #include "shaders/blend_shape.glsl.gen.h"
 #include "shaders/canvas.glsl.gen.h"
 #include "shaders/copy.glsl.gen.h"
@@ -63,10 +65,11 @@ public:
 	};
 
 	struct Config {
-
 		bool shrink_textures_x2;
 		bool use_fast_texture_filter;
 		bool use_anisotropic_filter;
+		bool use_lightmap_filter_bicubic;
+		bool use_physical_light_attenuation;
 
 		bool s3tc_supported;
 		bool latc_supported;
@@ -76,16 +79,20 @@ public:
 		bool etc2_supported;
 		bool pvrtc_supported;
 
-		bool hdr_supported;
-
 		bool srgb_decode_supported;
 
+		bool support_npot_repeat_mipmap;
+		bool texture_float_linear_supported;
+		bool framebuffer_float_supported;
+		bool framebuffer_half_float_supported;
+
 		bool use_rgba_2d_shadows;
 
 		float anisotropic_level;
 
 		int max_texture_image_units;
 		int max_texture_size;
+		int max_cubemap_texture_size;
 
 		bool generate_wireframes;
 
@@ -95,12 +102,15 @@ public:
 
 		bool keep_original_textures;
 
-		bool no_depth_prepass;
+		bool use_depth_prepass;
 		bool force_vertex_shading;
+
+		// in some cases the legacy render didn't orphan. We will mark these
+		// so the user can switch orphaning off for them.
+		bool should_orphan;
 	} config;
 
 	mutable struct Shaders {
-
 		CopyShaderGLES3 copy;
 
 		ShaderCompilerGLES3 compiler;
@@ -117,12 +127,14 @@ public:
 	} shaders;
 
 	struct Resources {
-
 		GLuint white_tex;
 		GLuint black_tex;
 		GLuint normal_tex;
 		GLuint aniso_tex;
 
+		GLuint white_tex_3d;
+		GLuint white_tex_array;
+
 		GLuint quadie;
 		GLuint quadie_array;
 
@@ -132,7 +144,6 @@ public:
 	} resources;
 
 	struct Info {
-
 		uint64_t texture_mem;
 		uint64_t vertex_mem;
 
@@ -143,6 +154,8 @@ public:
 			uint32_t surface_switch_count;
 			uint32_t shader_rebind_count;
 			uint32_t vertices_count;
+			uint32_t _2d_item_count;
+			uint32_t _2d_draw_call_count;
 
 			void reset() {
 				object_count = 0;
@@ -151,11 +164,12 @@ public:
 				surface_switch_count = 0;
 				shader_rebind_count = 0;
 				vertices_count = 0;
+				_2d_item_count = 0;
+				_2d_draw_call_count = 0;
 			}
 		} render, render_final, snap;
 
 		Info() {
-
 			texture_mem = 0;
 			vertex_mem = 0;
 			render.reset();
@@ -169,25 +183,12 @@ public:
 	/////////////////////////////////////////////////////////////////////////////////////////
 
 	struct Instantiable : public RID_Data {
-
 		SelfList<RasterizerScene::InstanceBase>::List instance_list;
 
-		_FORCE_INLINE_ void instance_change_notify() {
-
+		_FORCE_INLINE_ void instance_change_notify(bool p_aabb, bool p_materials) {
 			SelfList<RasterizerScene::InstanceBase> *instances = instance_list.first();
 			while (instances) {
-
-				instances->self()->base_changed();
-				instances = instances->next();
-			}
-		}
-
-		_FORCE_INLINE_ void instance_material_change_notify() {
-
-			SelfList<RasterizerScene::InstanceBase> *instances = instance_list.first();
-			while (instances) {
-
-				instances->self()->base_material_changed();
+				instances->self()->base_changed(p_aabb, p_materials);
 				instances = instances->next();
 			}
 		}
@@ -195,7 +196,6 @@ public:
 		_FORCE_INLINE_ void instance_remove_deps() {
 			SelfList<RasterizerScene::InstanceBase> *instances = instance_list.first();
 			while (instances) {
-
 				SelfList<RasterizerScene::InstanceBase> *next = instances->next();
 				instances->self()->base_removed();
 				instances = next;
@@ -208,11 +208,9 @@ public:
 	};
 
 	struct GeometryOwner : public Instantiable {
-
 		virtual ~GeometryOwner() {}
 	};
 	struct Geometry : Instantiable {
-
 		enum Type {
 			GEOMETRY_INVALID,
 			GEOMETRY_SURFACE,
@@ -242,15 +240,15 @@ public:
 	struct RenderTarget;
 
 	struct Texture : public RID_Data {
-
 		Texture *proxy;
 		Set<Texture *> proxy_owners;
 
 		String path;
 		uint32_t flags;
-		int width, height;
-		int alloc_width, alloc_height;
+		int width, height, depth;
+		int alloc_width, alloc_height, alloc_depth;
 		Image::Format format;
+		VS::TextureType type;
 
 		GLenum target;
 		GLenum gl_format_cache;
@@ -264,16 +262,19 @@ public:
 
 		int mipmaps;
 
+		bool is_npot_repeat_mipmap;
+
 		bool active;
 		GLuint tex_id;
 
 		bool using_srgb;
+		bool redraw_if_visible;
 
 		uint16_t stored_cube_sides;
 
 		RenderTarget *render_target;
 
-		Ref<Image> images[6];
+		Vector<Ref<Image>> images;
 
 		VisualServer::TextureDetectCallback detect_3d;
 		void *detect_3d_ud;
@@ -284,28 +285,32 @@ public:
 		VisualServer::TextureDetectCallback detect_normal;
 		void *detect_normal_ud;
 
-		Texture() {
-
-			using_srgb = false;
-			stored_cube_sides = 0;
-			ignore_mipmaps = false;
-			render_target = NULL;
-			flags = width = height = 0;
-			tex_id = 0;
-			data_size = 0;
-			format = Image::FORMAT_L8;
-			active = false;
-			compressed = false;
-			total_data_size = 0;
-			target = GL_TEXTURE_2D;
-			mipmaps = 0;
-			detect_3d = NULL;
-			detect_3d_ud = NULL;
-			detect_srgb = NULL;
-			detect_srgb_ud = NULL;
-			detect_normal = NULL;
-			detect_normal_ud = NULL;
-			proxy = NULL;
+		Texture() :
+				proxy(nullptr),
+				flags(0),
+				width(0),
+				height(0),
+				format(Image::FORMAT_L8),
+				type(VS::TEXTURE_TYPE_2D),
+				target(GL_TEXTURE_2D),
+				data_size(0),
+				compressed(false),
+				srgb(false),
+				total_data_size(0),
+				ignore_mipmaps(false),
+				mipmaps(0),
+				active(false),
+				tex_id(0),
+				using_srgb(false),
+				redraw_if_visible(false),
+				stored_cube_sides(0),
+				render_target(nullptr),
+				detect_3d(nullptr),
+				detect_3d_ud(nullptr),
+				detect_srgb(nullptr),
+				detect_srgb_ud(nullptr),
+				detect_normal(nullptr),
+				detect_normal_ud(nullptr) {
 		}
 
 		_ALWAYS_INLINE_ Texture *get_ptr() {
@@ -317,14 +322,12 @@ public:
 		}
 
 		~Texture() {
-
 			if (tex_id != 0) {
-
 				glDeleteTextures(1, &tex_id);
 			}
 
 			for (Set<Texture *>::Element *E = proxy_owners.front(); E; E = E->next()) {
-				E->get()->proxy = NULL;
+				E->get()->proxy = nullptr;
 			}
 
 			if (proxy) {
@@ -335,19 +338,23 @@ public:
 
 	mutable RID_Owner<Texture> texture_owner;
 
-	Ref<Image> _get_gl_image_and_format(const Ref<Image> &p_image, Image::Format p_format, uint32_t p_flags, GLenum &r_gl_format, GLenum &r_gl_internal_format, GLenum &r_gl_type, bool &r_compressed, bool &srgb);
+	Ref<Image> _get_gl_image_and_format(const Ref<Image> &p_image, Image::Format p_format, uint32_t p_flags, Image::Format &r_real_format, GLenum &r_gl_format, GLenum &r_gl_internal_format, GLenum &r_gl_type, bool &r_compressed, bool &r_srgb, bool p_force_decompress) const;
 
 	virtual RID texture_create();
-	virtual void texture_allocate(RID p_texture, int p_width, int p_height, Image::Format p_format, uint32_t p_flags = VS::TEXTURE_FLAGS_DEFAULT);
-	virtual void texture_set_data(RID p_texture, const Ref<Image> &p_image, VS::CubeMapSide p_cube_side = VS::CUBEMAP_LEFT);
-	virtual Ref<Image> texture_get_data(RID p_texture, VS::CubeMapSide p_cube_side = VS::CUBEMAP_LEFT) const;
+	virtual void texture_allocate(RID p_texture, int p_width, int p_height, int p_depth_3d, Image::Format p_format, VS::TextureType p_type, uint32_t p_flags = VS::TEXTURE_FLAGS_DEFAULT);
+	virtual void texture_set_data(RID p_texture, const Ref<Image> &p_image, int p_layer = 0);
+	virtual void texture_set_data_partial(RID p_texture, const Ref<Image> &p_image, int src_x, int src_y, int src_w, int src_h, int dst_x, int dst_y, int p_dst_mip, int p_layer = 0);
+	virtual Ref<Image> texture_get_data(RID p_texture, int p_layer = 0) const;
 	virtual void texture_set_flags(RID p_texture, uint32_t p_flags);
 	virtual uint32_t texture_get_flags(RID p_texture) const;
 	virtual Image::Format texture_get_format(RID p_texture) const;
+	virtual VS::TextureType texture_get_type(RID p_texture) const;
 	virtual uint32_t texture_get_texid(RID p_texture) const;
 	virtual uint32_t texture_get_width(RID p_texture) const;
 	virtual uint32_t texture_get_height(RID p_texture) const;
-	virtual void texture_set_size_override(RID p_texture, int p_width, int p_height);
+	virtual uint32_t texture_get_depth(RID p_texture) const;
+	virtual void texture_set_size_override(RID p_texture, int p_width, int p_height, int p_depth);
+	virtual void texture_bind(RID p_texture, uint32_t p_texture_no);
 
 	virtual void texture_set_path(RID p_texture, const String &p_path);
 	virtual String texture_get_path(RID p_texture) const;
@@ -365,13 +372,16 @@ public:
 	virtual void texture_set_detect_normal_callback(RID p_texture, VisualServer::TextureDetectCallback p_callback, void *p_userdata);
 
 	virtual void texture_set_proxy(RID p_texture, RID p_proxy);
+	virtual Size2 texture_size_with_proxy(RID p_texture) const;
+
+	virtual void texture_set_force_redraw_if_visible(RID p_texture, bool p_enable);
 
 	/* SKY API */
 
 	struct Sky : public RID_Data {
-
 		RID panorama;
 		GLuint radiance;
+		GLuint irradiance;
 		int radiance_size;
 	};
 
@@ -385,7 +395,6 @@ public:
 	struct Material;
 
 	struct Shader : public RID_Data {
-
 		RID self;
 
 		VS::ShaderMode mode;
@@ -406,6 +415,7 @@ public:
 
 		Map<StringName, RID> default_textures;
 
+		Vector<ShaderLanguage::DataType> texture_types;
 		Vector<ShaderLanguage::ShaderNode::Uniform::Hint> texture_hints;
 
 		bool valid;
@@ -413,13 +423,13 @@ public:
 		String path;
 
 		struct CanvasItem {
-
 			enum BlendMode {
 				BLEND_MODE_MIX,
 				BLEND_MODE_ADD,
 				BLEND_MODE_SUB,
 				BLEND_MODE_MUL,
 				BLEND_MODE_PMALPHA,
+				BLEND_MODE_DISABLED,
 			};
 
 			int blend_mode;
@@ -431,14 +441,29 @@ public:
 			};
 
 			int light_mode;
+
+			// these flags are specifically for batching
+			// some of the logic is thus in rasterizer_storage.cpp
+			// we could alternatively set bitflags for each 'uses' and test on the fly
+			// defined in RasterizerStorageCommon::BatchFlags
+			unsigned int batch_flags;
+
 			bool uses_screen_texture;
 			bool uses_screen_uv;
 			bool uses_time;
+			bool uses_modulate;
+			bool uses_color;
+			bool uses_vertex;
+
+			// all these should disable item joining if used in a custom shader
+			bool uses_world_matrix;
+			bool uses_extra_matrix;
+			bool uses_projection_matrix;
+			bool uses_instance_custom;
 
 		} canvas_item;
 
 		struct Spatial {
-
 			enum BlendMode {
 				BLEND_MODE_MIX,
 				BLEND_MODE_ADD,
@@ -475,6 +500,8 @@ public:
 			bool uses_screen_texture;
 			bool uses_depth_texture;
 			bool uses_time;
+			bool uses_tangent;
+			bool uses_ensure_correct_normals;
 			bool writes_modelview_or_projection;
 			bool uses_vertex_lighting;
 			bool uses_world_coordinates;
@@ -482,7 +509,6 @@ public:
 		} spatial;
 
 		struct Particles {
-
 		} particles;
 
 		bool uses_vertex_time;
@@ -490,8 +516,7 @@ public:
 
 		Shader() :
 				dirty_list(this) {
-
-			shader = NULL;
+			shader = nullptr;
 			ubo_size = 0;
 			valid = false;
 			custom_code_id = 0;
@@ -513,6 +538,10 @@ public:
 	virtual void shader_set_default_texture_param(RID p_shader, const StringName &p_name, RID p_texture);
 	virtual RID shader_get_default_texture_param(RID p_shader, const StringName &p_name) const;
 
+	virtual void shader_add_custom_define(RID p_shader, const String &p_define);
+	virtual void shader_get_custom_defines(RID p_shader, Vector<String> *p_defines) const;
+	virtual void shader_remove_custom_define(RID p_shader, const String &p_define);
+
 	void _update_shader(Shader *p_shader) const;
 
 	void update_dirty_shaders();
@@ -520,13 +549,13 @@ public:
 	/* COMMON MATERIAL API */
 
 	struct Material : public RID_Data {
-
 		Shader *shader;
 		GLuint ubo_id;
 		uint32_t ubo_size;
 		Map<StringName, Variant> params;
 		SelfList<Material> list;
 		SelfList<Material> dirty_list;
+		Vector<bool> texture_is_3d;
 		Vector<RID> textures;
 		float line_width;
 		int render_priority;
@@ -543,16 +572,16 @@ public:
 		bool is_animated_cache;
 
 		Material() :
+				shader(nullptr),
+				ubo_id(0),
+				ubo_size(0),
 				list(this),
-				dirty_list(this) {
-			can_cast_shadow_cache = false;
-			is_animated_cache = false;
-			shader = NULL;
-			line_width = 1.0;
-			ubo_id = 0;
-			ubo_size = 0;
-			last_pass = 0;
-			render_priority = 0;
+				dirty_list(this),
+				line_width(1.0),
+				render_priority(0),
+				last_pass(0),
+				can_cast_shadow_cache(false),
+				is_animated_cache(false) {
 		}
 	};
 
@@ -570,12 +599,15 @@ public:
 
 	virtual void material_set_param(RID p_material, const StringName &p_param, const Variant &p_value);
 	virtual Variant material_get_param(RID p_material, const StringName &p_param) const;
+	virtual Variant material_get_param_default(RID p_material, const StringName &p_param) const;
 
 	virtual void material_set_line_width(RID p_material, float p_width);
 	virtual void material_set_next_pass(RID p_material, RID p_next_material);
 
 	virtual bool material_is_animated(RID p_material);
 	virtual bool material_casts_shadows(RID p_material);
+	virtual bool material_uses_tangents(RID p_material);
+	virtual bool material_uses_ensure_correct_normals(RID p_material);
 
 	virtual void material_add_instance_owner(RID p_material, RasterizerScene::InstanceBase *p_instance);
 	virtual void material_remove_instance_owner(RID p_material, RasterizerScene::InstanceBase *p_instance);
@@ -590,9 +622,7 @@ public:
 
 	struct Mesh;
 	struct Surface : public Geometry {
-
 		struct Attrib {
-
 			bool enabled;
 			bool integer;
 			GLuint index;
@@ -644,65 +674,60 @@ public:
 		bool active;
 
 		virtual void material_changed_notify() {
-			mesh->instance_material_change_notify();
+			mesh->instance_change_notify(false, true);
 			mesh->update_multimeshes();
 		}
 
 		int total_data_size;
 
-		Surface() {
-
-			array_byte_size = 0;
-			index_array_byte_size = 0;
-			mesh = NULL;
-			format = 0;
-			array_id = 0;
-			vertex_id = 0;
-			index_id = 0;
-			array_len = 0;
+		Surface() :
+				mesh(nullptr),
+				format(0),
+				array_id(0),
+				vertex_id(0),
+				index_id(0),
+				index_wireframe_id(0),
+				array_wireframe_id(0),
+				instancing_array_wireframe_id(0),
+				index_wireframe_len(0),
+				array_len(0),
+				index_array_len(0),
+				array_byte_size(0),
+				index_array_byte_size(0),
+				primitive(VS::PRIMITIVE_POINTS),
+				active(false),
+				total_data_size(0) {
 			type = GEOMETRY_SURFACE;
-			primitive = VS::PRIMITIVE_POINTS;
-			index_array_len = 0;
-			active = false;
-
-			total_data_size = 0;
-
-			index_wireframe_id = 0;
-			array_wireframe_id = 0;
-			instancing_array_wireframe_id = 0;
-			index_wireframe_len = 0;
 		}
 
 		~Surface() {
 		}
 	};
 
-	class MultiMesh;
+	struct MultiMesh;
 
 	struct Mesh : public GeometryOwner {
-
 		bool active;
 		Vector<Surface *> surfaces;
 		int blend_shape_count;
 		VS::BlendShapeMode blend_shape_mode;
+		PoolRealArray blend_shape_values;
 		AABB custom_aabb;
 		mutable uint64_t last_pass;
 		SelfList<MultiMesh>::List multimeshes;
-
 		_FORCE_INLINE_ void update_multimeshes() {
-
 			SelfList<MultiMesh> *mm = multimeshes.first();
 			while (mm) {
-				mm->self()->instance_material_change_notify();
+				mm->self()->instance_change_notify(false, true);
 				mm = mm->next();
 			}
 		}
 
-		Mesh() {
-			blend_shape_mode = VS::BLEND_SHAPE_MODE_NORMALIZED;
-			blend_shape_count = 0;
-			last_pass = 0;
-			active = false;
+		Mesh() :
+				active(false),
+				blend_shape_count(0),
+				blend_shape_mode(VS::BLEND_SHAPE_MODE_NORMALIZED),
+				last_pass(0) {
 		}
 	};
 
@@ -710,7 +735,7 @@ public:
 
 	virtual RID mesh_create();
 
-	virtual void mesh_add_surface(RID p_mesh, uint32_t p_format, VS::PrimitiveType p_primitive, const PoolVector<uint8_t> &p_array, int p_vertex_count, const PoolVector<uint8_t> &p_index_array, int p_index_count, const AABB &p_aabb, const Vector<PoolVector<uint8_t> > &p_blend_shapes = Vector<PoolVector<uint8_t> >(), const Vector<AABB> &p_bone_aabbs = Vector<AABB>());
+	virtual void mesh_add_surface(RID p_mesh, uint32_t p_format, VS::PrimitiveType p_primitive, const PoolVector<uint8_t> &p_array, int p_vertex_count, const PoolVector<uint8_t> &p_index_array, int p_index_count, const AABB &p_aabb, const Vector<PoolVector<uint8_t>> &p_blend_shapes = Vector<PoolVector<uint8_t>>(), const Vector<AABB> &p_bone_aabbs = Vector<AABB>());
 
 	virtual void mesh_set_blend_shape_count(RID p_mesh, int p_amount);
 	virtual int mesh_get_blend_shape_count(RID p_mesh) const;
@@ -718,6 +743,9 @@ public:
 	virtual void mesh_set_blend_shape_mode(RID p_mesh, VS::BlendShapeMode p_mode);
 	virtual VS::BlendShapeMode mesh_get_blend_shape_mode(RID p_mesh) const;
 
+	virtual void mesh_set_blend_shape_values(RID p_mesh, PoolVector<float> p_values);
+	virtual PoolVector<float> mesh_get_blend_shape_values(RID p_mesh) const;
+
 	virtual void mesh_surface_update_region(RID p_mesh, int p_surface, int p_offset, const PoolVector<uint8_t> &p_data);
 
 	virtual void mesh_surface_set_material(RID p_mesh, int p_surface, RID p_material);
@@ -733,7 +761,7 @@ public:
 	virtual VS::PrimitiveType mesh_surface_get_primitive_type(RID p_mesh, int p_surface) const;
 
 	virtual AABB mesh_surface_get_aabb(RID p_mesh, int p_surface) const;
-	virtual Vector<PoolVector<uint8_t> > mesh_surface_get_blend_shapes(RID p_mesh, int p_surface) const;
+	virtual Vector<PoolVector<uint8_t>> mesh_surface_get_blend_shapes(RID p_mesh, int p_surface) const;
 	virtual Vector<AABB> mesh_surface_get_skeleton_aabb(RID p_mesh, int p_surface) const;
 
 	virtual void mesh_remove_surface(RID p_mesh, int p_surface);
@@ -754,6 +782,7 @@ public:
 		int size;
 		VS::MultimeshTransformFormat transform_format;
 		VS::MultimeshColorFormat color_format;
+		VS::MultimeshCustomDataFormat custom_data_format;
 		Vector<float> data;
 		AABB aabb;
 		SelfList<MultiMesh> update_list;
@@ -763,22 +792,25 @@ public:
 
 		int xform_floats;
 		int color_floats;
+		int custom_data_floats;
 
 		bool dirty_aabb;
 		bool dirty_data;
 
 		MultiMesh() :
+				size(0),
+				transform_format(VS::MULTIMESH_TRANSFORM_2D),
+				color_format(VS::MULTIMESH_COLOR_NONE),
+				custom_data_format(VS::MULTIMESH_CUSTOM_DATA_NONE),
 				update_list(this),
-				mesh_list(this) {
-			dirty_aabb = true;
-			dirty_data = true;
-			xform_floats = 0;
-			color_floats = 0;
-			visible_instances = -1;
-			size = 0;
-			buffer = 0;
-			transform_format = VS::MULTIMESH_TRANSFORM_2D;
-			color_format = VS::MULTIMESH_COLOR_NONE;
+				mesh_list(this),
+				buffer(0),
+				visible_instances(-1),
+				xform_floats(0),
+				color_floats(0),
+				custom_data_floats(0),
+				dirty_aabb(true),
+				dirty_data(true) {
 		}
 	};
 
@@ -790,19 +822,23 @@ public:
 
 	virtual RID multimesh_create();
 
-	virtual void multimesh_allocate(RID p_multimesh, int p_instances, VS::MultimeshTransformFormat p_transform_format, VS::MultimeshColorFormat p_color_format);
+	virtual void multimesh_allocate(RID p_multimesh, int p_instances, VS::MultimeshTransformFormat p_transform_format, VS::MultimeshColorFormat p_color_format, VS::MultimeshCustomDataFormat p_data_format = VS::MULTIMESH_CUSTOM_DATA_NONE);
 	virtual int multimesh_get_instance_count(RID p_multimesh) const;
 
 	virtual void multimesh_set_mesh(RID p_multimesh, RID p_mesh);
 	virtual void multimesh_instance_set_transform(RID p_multimesh, int p_index, const Transform &p_transform);
 	virtual void multimesh_instance_set_transform_2d(RID p_multimesh, int p_index, const Transform2D &p_transform);
 	virtual void multimesh_instance_set_color(RID p_multimesh, int p_index, const Color &p_color);
+	virtual void multimesh_instance_set_custom_data(RID p_multimesh, int p_index, const Color &p_custom_data);
 
 	virtual RID multimesh_get_mesh(RID p_multimesh) const;
 
 	virtual Transform multimesh_instance_get_transform(RID p_multimesh, int p_index) const;
 	virtual Transform2D multimesh_instance_get_transform_2d(RID p_multimesh, int p_index) const;
 	virtual Color multimesh_instance_get_color(RID p_multimesh, int p_index) const;
+	virtual Color multimesh_instance_get_custom_data(RID p_multimesh, int p_index) const;
+
+	virtual void multimesh_set_as_bulk_array(RID p_multimesh, const PoolVector<float> &p_array);
 
 	virtual void multimesh_set_visible_instances(RID p_multimesh, int p_visible);
 	virtual int multimesh_get_visible_instances(RID p_multimesh) const;
@@ -812,9 +848,7 @@ public:
 	/* IMMEDIATE API */
 
 	struct Immediate : public Geometry {
-
 		struct Chunk {
-
 			RID texture;
 			VS::PrimitiveType primitive;
 			Vector<Vector3> vertices;
@@ -846,7 +880,7 @@ public:
 	mutable RID_Owner<Immediate> immediate_owner;
 
 	virtual RID immediate_create();
-	virtual void immediate_begin(RID p_immediate, VS::PrimitiveType p_rimitive, RID p_texture = RID());
+	virtual void immediate_begin(RID p_immediate, VS::PrimitiveType p_primitive, RID p_texture = RID());
 	virtual void immediate_vertex(RID p_immediate, const Vector3 &p_vertex);
 	virtual void immediate_normal(RID p_immediate, const Vector3 &p_normal);
 	virtual void immediate_tangent(RID p_immediate, const Plane &p_tangent);
@@ -868,13 +902,13 @@ public:
 		GLuint texture;
 		SelfList<Skeleton> update_list;
 		Set<RasterizerScene::InstanceBase *> instances; //instances using skeleton
+		Transform2D base_transform_2d;
 
 		Skeleton() :
+				use_2d(false),
+				size(0),
+				texture(0),
 				update_list(this) {
-			size = 0;
-
-			use_2d = false;
-			texture = 0;
 		}
 	};
 
@@ -891,11 +925,11 @@ public:
 	virtual Transform skeleton_bone_get_transform(RID p_skeleton, int p_bone) const;
 	virtual void skeleton_bone_set_transform_2d(RID p_skeleton, int p_bone, const Transform2D &p_transform);
 	virtual Transform2D skeleton_bone_get_transform_2d(RID p_skeleton, int p_bone) const;
+	virtual void skeleton_set_base_transform_2d(RID p_skeleton, const Transform2D &p_base_transform);
 
 	/* Light API */
 
 	struct Light : Instantiable {
-
 		VS::LightType type;
 		float param[VS::LIGHT_PARAM_MAX];
 		Color color;
@@ -904,6 +938,7 @@ public:
 		bool shadow;
 		bool negative;
 		bool reverse_cull;
+		VS::LightBakeMode bake_mode;
 		uint32_t cull_mask;
 		VS::LightOmniShadowMode omni_shadow_mode;
 		VS::LightOmniShadowDetail omni_shadow_detail;
@@ -925,6 +960,8 @@ public:
 	virtual void light_set_negative(RID p_light, bool p_enable);
 	virtual void light_set_cull_mask(RID p_light, uint32_t p_mask);
 	virtual void light_set_reverse_cull_face_mode(RID p_light, bool p_enabled);
+	virtual void light_set_use_gi(RID p_light, bool p_enabled);
+	virtual void light_set_bake_mode(RID p_light, VS::LightBakeMode p_bake_mode);
 
 	virtual void light_omni_set_shadow_mode(RID p_light, VS::LightOmniShadowMode p_mode);
 	virtual void light_omni_set_shadow_detail(RID p_light, VS::LightOmniShadowDetail p_detail);
@@ -944,6 +981,8 @@ public:
 	virtual VS::LightType light_get_type(RID p_light) const;
 	virtual float light_get_param(RID p_light, VS::LightParam p_param);
 	virtual Color light_get_color(RID p_light);
+	virtual bool light_get_use_gi(RID p_light);
+	virtual VS::LightBakeMode light_get_bake_mode(RID p_light);
 
 	virtual AABB light_get_aabb(RID p_light) const;
 	virtual uint64_t light_get_version(RID p_light) const;
@@ -951,7 +990,6 @@ public:
 	/* PROBE API */
 
 	struct ReflectionProbe : Instantiable {
-
 		VS::ReflectionProbeUpdateMode update_mode;
 		float intensity;
 		Color interior_ambient;
@@ -982,6 +1020,7 @@ public:
 	virtual void reflection_probe_set_enable_box_projection(RID p_probe, bool p_enable);
 	virtual void reflection_probe_set_enable_shadows(RID p_probe, bool p_enable);
 	virtual void reflection_probe_set_cull_mask(RID p_probe, uint32_t p_layers);
+	virtual void reflection_probe_set_resolution(RID p_probe, int p_resolution);
 
 	virtual AABB reflection_probe_get_aabb(RID p_probe) const;
 	virtual VS::ReflectionProbeUpdateMode reflection_probe_get_update_mode(RID p_probe) const;
@@ -995,7 +1034,6 @@ public:
 	/* GI PROBE API */
 
 	struct GIProbe : public Instantiable {
-
 		AABB bounds;
 		Transform to_cell;
 		float cell_size;
@@ -1053,7 +1091,6 @@ public:
 	virtual uint32_t gi_probe_get_version(RID p_probe);
 
 	struct GIProbeData : public RID_Data {
-
 		int width;
 		int height;
 		int depth;
@@ -1067,7 +1104,6 @@ public:
 
 	mutable RID_Owner<GIProbeData> gi_probe_data_owner;
 
-	virtual GIProbeCompression gi_probe_get_dynamic_data_get_preferred_compression() const;
 	virtual RID gi_probe_dynamic_data_create(int p_width, int p_height, int p_depth, GIProbeCompression p_compression);
 	virtual void gi_probe_dynamic_data_update(RID p_gi_probe_data, int p_depth_slice, int p_slice_count, int p_mipmap, const void *p_data);
 
@@ -1085,28 +1121,38 @@ public:
 
 	virtual void lightmap_capture_set_energy(RID p_capture, float p_energy);
 	virtual float lightmap_capture_get_energy(RID p_capture) const;
+	virtual void lightmap_capture_set_interior(RID p_capture, bool p_interior);
+	virtual bool lightmap_capture_is_interior(RID p_capture) const;
 
 	virtual const PoolVector<LightmapCaptureOctree> *lightmap_capture_get_octree_ptr(RID p_capture) const;
 
 	struct LightmapCapture : public Instantiable {
-
 		PoolVector<LightmapCaptureOctree> octree;
 		AABB bounds;
 		Transform cell_xform;
 		int cell_subdiv;
 		float energy;
-		LightmapCapture() {
+		bool interior;
+
+		SelfList<LightmapCapture> update_list;
+
+		LightmapCapture() :
+				update_list(this) {
 			energy = 1.0;
 			cell_subdiv = 1;
+			interior = false;
 		}
 	};
 
+	SelfList<LightmapCapture>::List capture_update_list;
+
+	void update_dirty_captures();
+
 	mutable RID_Owner<LightmapCapture> lightmap_capture_data_owner;
 
 	/* PARTICLES */
 
 	struct Particles : public GeometryOwner {
-
 		bool inactive;
 		float inactive_time;
 		bool emitting;
@@ -1153,43 +1199,36 @@ public:
 		Transform emission_transform;
 
 		Particles() :
-				particle_element(this) {
-			cycle_number = 0;
-			emitting = false;
-			one_shot = false;
-			amount = 0;
-			lifetime = 1.0;
-			pre_process_time = 0.0;
-			explosiveness = 0.0;
-			randomness = 0.0;
-			use_local_coords = true;
-			fixed_fps = 0;
-			fractional_delta = false;
-			frame_remainder = 0;
-			histories_enabled = false;
-			speed_scale = 1.0;
-			random_seed = 0;
-
-			restart_request = false;
-
-			custom_aabb = AABB(Vector3(-4, -4, -4), Vector3(8, 8, 8));
-
-			draw_order = VS::PARTICLES_DRAW_ORDER_INDEX;
+				inactive(true),
+				inactive_time(0.0),
+				emitting(false),
+				one_shot(false),
+				amount(0),
+				lifetime(1.0),
+				pre_process_time(0.0),
+				explosiveness(0.0),
+				randomness(0.0),
+				restart_request(false),
+				custom_aabb(AABB(Vector3(-4, -4, -4), Vector3(8, 8, 8))),
+				use_local_coords(true),
+				draw_order(VS::PARTICLES_DRAW_ORDER_INDEX),
+				histories_enabled(false),
+				particle_element(this),
+				prev_ticks(0),
+				random_seed(0),
+				cycle_number(0),
+				speed_scale(1.0),
+				fixed_fps(0),
+				fractional_delta(false),
+				frame_remainder(0),
+				clear(true) {
 			particle_buffers[0] = 0;
 			particle_buffers[1] = 0;
-
-			prev_ticks = 0;
-
-			clear = true;
-			inactive = true;
-			inactive_time = 0.0;
-
 			glGenBuffers(2, particle_buffers);
 			glGenVertexArrays(2, particle_vaos);
 		}
 
 		~Particles() {
-
 			glDeleteBuffers(2, particle_buffers);
 			glDeleteVertexArrays(2, particle_vaos);
 			if (histories_enabled) {
@@ -1240,6 +1279,8 @@ public:
 	virtual int particles_get_draw_passes(RID p_particles) const;
 	virtual RID particles_get_draw_pass_mesh(RID p_particles, int p_pass) const;
 
+	virtual bool particles_is_inactive(RID p_particles) const;
+
 	/* INSTANCE */
 
 	virtual void instance_add_skeleton(RID p_skeleton, RasterizerScene::InstanceBase *p_instance);
@@ -1251,13 +1292,11 @@ public:
 	/* RENDER TARGET */
 
 	struct RenderTarget : public RID_Data {
-
 		GLuint fbo;
 		GLuint color;
 		GLuint depth;
 
 		struct Buffers {
-
 			bool active;
 			bool effects_active;
 			GLuint fbo;
@@ -1273,9 +1312,7 @@ public:
 		} buffers;
 
 		struct Effects {
-
 			struct MipMaps {
-
 				struct Size {
 					GLuint fbo;
 					int width;
@@ -1286,9 +1323,9 @@ public:
 				GLuint color;
 				int levels;
 
-				MipMaps() {
-					color = 0;
-					levels = 0;
+				MipMaps() :
+						color(0),
+						levels(0) {
 				}
 			};
 
@@ -1303,10 +1340,10 @@ public:
 
 				Vector<GLuint> depth_mipmap_fbos; //fbos for depth mipmapsla ver
 
-				SSAO() {
+				SSAO() :
+						linear_depth(0) {
 					blur_fbo[0] = 0;
 					blur_fbo[1] = 0;
-					linear_depth = 0;
 				}
 			} ssao;
 
@@ -1318,9 +1355,24 @@ public:
 			GLuint fbo;
 			GLuint color;
 
-			Exposure() { fbo = 0; }
+			Exposure() :
+					fbo(0) {}
 		} exposure;
 
+		// External FBO to render our final result to (mostly used for ARVR)
+		struct External {
+			GLuint fbo;
+			GLuint color;
+			GLuint depth;
+			RID texture;
+
+			External() :
+					fbo(0),
+					color(0),
+					depth(0) {
+			}
+		} external;
+
 		uint64_t last_exposure_tick;
 
 		int width, height;
@@ -1329,29 +1381,32 @@ public:
 
 		bool used_in_frame;
 		VS::ViewportMSAA msaa;
+		bool use_fxaa;
+		bool use_debanding;
+		float sharpen_intensity;
 
 		RID texture;
 
-		RenderTarget() {
-
-			msaa = VS::VIEWPORT_MSAA_DISABLED;
-			width = 0;
-			height = 0;
-			depth = 0;
-			fbo = 0;
+		RenderTarget() :
+				fbo(0),
+				depth(0),
+				last_exposure_tick(0),
+				width(0),
+				height(0),
+				used_in_frame(false),
+				msaa(VS::VIEWPORT_MSAA_DISABLED),
+				use_fxaa(false),
+				use_debanding(false),
+				sharpen_intensity(0.0) {
 			exposure.fbo = 0;
 			buffers.fbo = 0;
-			used_in_frame = false;
-
+			external.fbo = 0;
 			for (int i = 0; i < RENDER_TARGET_FLAG_MAX; i++) {
 				flags[i] = false;
 			}
 			flags[RENDER_TARGET_HDR] = true;
-
 			buffers.active = false;
 			buffers.effects_active = false;
-
-			last_exposure_tick = 0;
 		}
 	};
 
@@ -1361,18 +1416,23 @@ public:
 	void _render_target_allocate(RenderTarget *rt);
 
 	virtual RID render_target_create();
+	virtual void render_target_set_position(RID p_render_target, int p_x, int p_y);
 	virtual void render_target_set_size(RID p_render_target, int p_width, int p_height);
 	virtual RID render_target_get_texture(RID p_render_target) const;
+	virtual uint32_t render_target_get_depth_texture_id(RID p_render_target) const;
+	virtual void render_target_set_external_texture(RID p_render_target, unsigned int p_texture_id, unsigned int p_depth_id);
 
 	virtual void render_target_set_flag(RID p_render_target, RenderTargetFlags p_flag, bool p_value);
 	virtual bool render_target_was_used(RID p_render_target);
 	virtual void render_target_clear_used(RID p_render_target);
 	virtual void render_target_set_msaa(RID p_render_target, VS::ViewportMSAA p_msaa);
+	virtual void render_target_set_use_fxaa(RID p_render_target, bool p_fxaa);
+	virtual void render_target_set_use_debanding(RID p_render_target, bool p_debanding);
+	virtual void render_target_set_sharpen_intensity(RID p_render_target, float p_intensity);
 
 	/* CANVAS SHADOW */
 
 	struct CanvasLightShadow : public RID_Data {
-
 		int size;
 		int height;
 		GLuint fbo;
@@ -1387,7 +1447,6 @@ public:
 	/* LIGHT SHADOW MAPPING */
 
 	struct CanvasOccluder : public RID_Data {
-
 		GLuint array_id; // 0 means, unconfigured
 		GLuint vertex_id; // 0 means, unconfigured
 		GLuint index_id; // 0 means, unconfigured
@@ -1405,15 +1464,12 @@ public:
 	virtual bool free(RID p_rid);
 
 	struct Frame {
-
 		RenderTarget *current_rt;
 
 		bool clear_request;
 		Color clear_request_color;
-		int canvas_draw_commands;
 		float time[4];
 		float delta;
-		uint64_t prev_tick;
 		uint64_t count;
 
 	} frame;
@@ -1431,9 +1487,54 @@ public:
 	virtual void render_info_end_capture();
 	virtual int get_captured_render_info(VS::RenderInfo p_info);
 
-	virtual int get_render_info(VS::RenderInfo p_info);
+	virtual uint64_t get_render_info(VS::RenderInfo p_info);
+	virtual String get_video_adapter_name() const;
+	virtual String get_video_adapter_vendor() const;
+
+	// NOTE : THESE SIZES ARE IN BYTES. BUFFER SIZES MAY NOT BE SPECIFIED IN BYTES SO REMEMBER TO CONVERT THEM WHEN CALLING.
+	void buffer_orphan_and_upload(unsigned int p_buffer_size_bytes, unsigned int p_offset_bytes, unsigned int p_data_size_bytes, const void *p_data, GLenum p_target = GL_ARRAY_BUFFER, GLenum p_usage = GL_DYNAMIC_DRAW, bool p_optional_orphan = false) const;
+	bool safe_buffer_sub_data(unsigned int p_total_buffer_size_bytes, GLenum p_target, unsigned int p_offset_bytes, unsigned int p_data_size_bytes, const void *p_data, unsigned int &r_offset_after_bytes) const;
 
 	RasterizerStorageGLES3();
 };
 
+inline bool RasterizerStorageGLES3::safe_buffer_sub_data(unsigned int p_total_buffer_size_bytes, GLenum p_target, unsigned int p_offset_bytes, unsigned int p_data_size_bytes, const void *p_data, unsigned int &r_offset_after_bytes) const {
+	r_offset_after_bytes = p_offset_bytes + p_data_size_bytes;
+#ifdef DEBUG_ENABLED
+	// we are trying to write across the edge of the buffer
+	if (r_offset_after_bytes > p_total_buffer_size_bytes) {
+		return false;
+	}
+#endif
+	glBufferSubData(p_target, p_offset_bytes, p_data_size_bytes, p_data);
+	return true;
+}
+
+// standardize the orphan / upload in one place so it can be changed per platform as necessary, and avoid future
+// bugs causing pipeline stalls
+// NOTE : THESE SIZES ARE IN BYTES. BUFFER SIZES MAY NOT BE SPECIFIED IN BYTES SO REMEMBER TO CONVERT THEM WHEN CALLING.
+inline void RasterizerStorageGLES3::buffer_orphan_and_upload(unsigned int p_buffer_size_bytes, unsigned int p_offset_bytes, unsigned int p_data_size_bytes, const void *p_data, GLenum p_target, GLenum p_usage, bool p_optional_orphan) const {
+	// Orphan the buffer to avoid CPU/GPU sync points caused by glBufferSubData
+	// Was previously #ifndef GLES_OVER_GL however this causes stalls on desktop mac also (and possibly other)
+	if (!p_optional_orphan || (config.should_orphan)) {
+		glBufferData(p_target, p_buffer_size_bytes, nullptr, p_usage);
+#ifdef RASTERIZER_EXTRA_CHECKS
+		// fill with garbage off the end of the array
+		if (p_buffer_size_bytes) {
+			unsigned int start = p_offset_bytes + p_data_size_bytes;
+			unsigned int end = start + 1024;
+			if (end < p_buffer_size) {
+				uint8_t *garbage = (uint8_t *)alloca(1024);
+				for (int n = 0; n < 1024; n++) {
+					garbage[n] = Math::random(0, 255);
+				}
+				glBufferSubData(p_target, start, 1024, garbage);
+			}
+		}
+#endif
+	}
+	ERR_FAIL_COND((p_offset_bytes + p_data_size_bytes) > p_buffer_size_bytes);
+	glBufferSubData(p_target, p_offset_bytes, p_data_size_bytes, p_data);
+}
+
 #endif // RASTERIZERSTORAGEGLES3_H
