@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -30,28 +30,27 @@
 
 #include "editor_profiler.h"
 
+#include "core/os/os.h"
 #include "editor_scale.h"
 #include "editor_settings.h"
-#include "os/os.h"
 
 void EditorProfiler::_make_metric_ptrs(Metric &m) {
-
 	for (int i = 0; i < m.categories.size(); i++) {
-		m.category_ptrs[m.categories[i].signature] = &m.categories[i];
+		m.category_ptrs[m.categories[i].signature] = &m.categories.write[i];
 		for (int j = 0; j < m.categories[i].items.size(); j++) {
-			m.item_ptrs[m.categories[i].items[j].signature] = &m.categories[i].items[j];
+			m.item_ptrs[m.categories[i].items[j].signature] = &m.categories.write[i].items.write[j];
 		}
 	}
 }
 
 void EditorProfiler::add_frame_metric(const Metric &p_metric, bool p_final) {
-
 	++last_metric;
-	if (last_metric >= frame_metrics.size())
+	if (last_metric >= frame_metrics.size()) {
 		last_metric = 0;
+	}
 
-	frame_metrics[last_metric] = p_metric;
-	_make_metric_ptrs(frame_metrics[last_metric]);
+	frame_metrics.write[last_metric] = p_metric;
+	_make_metric_ptrs(frame_metrics.write[last_metric]);
 
 	updating_frame = true;
 	cursor_metric_edit->set_max(frame_metrics[last_metric].frame_number);
@@ -68,34 +67,31 @@ void EditorProfiler::add_frame_metric(const Metric &p_metric, bool p_final) {
 	}
 	updating_frame = false;
 
-	if (!frame_delay->is_processing()) {
-
+	if (frame_delay->is_stopped()) {
 		frame_delay->set_wait_time(p_final ? 0.1 : 1);
 		frame_delay->start();
 	}
 
-	if (!plot_delay->is_processing()) {
+	if (plot_delay->is_stopped()) {
 		plot_delay->set_wait_time(0.1);
 		plot_delay->start();
 	}
 }
 
 void EditorProfiler::clear() {
-
 	int metric_size = EditorSettings::get_singleton()->get("debugger/profiler_frame_history_size");
 	metric_size = CLAMP(metric_size, 60, 1024);
 	frame_metrics.clear();
 	frame_metrics.resize(metric_size);
 	last_metric = -1;
 	variables->clear();
-	//activate->set_pressed(false);
 	plot_sigs.clear();
 	plot_sigs.insert("physics_frame_time");
 	plot_sigs.insert("category_frame_time");
 
 	updating_frame = true;
 	cursor_metric_edit->set_min(0);
-	cursor_metric_edit->set_max(0);
+	cursor_metric_edit->set_max(100); // Doesn't make much sense, but we can't have min == max. Doesn't hurt.
 	cursor_metric_edit->set_value(0);
 	updating_frame = false;
 	hover_metric = -1;
@@ -103,26 +99,27 @@ void EditorProfiler::clear() {
 }
 
 static String _get_percent_txt(float p_value, float p_total) {
-	if (p_total == 0)
+	if (p_total == 0) {
 		p_total = 0.00001;
+	}
+
 	return String::num((p_value / p_total) * 100, 1) + "%";
 }
 
-String EditorProfiler::_get_time_as_text(Metric &m, float p_time, int p_calls) {
-
-	int dmode = display_mode->get_selected();
+String EditorProfiler::_get_time_as_text(const Metric &m, float p_time, int p_calls) {
+	const int dmode = display_mode->get_selected();
 
 	if (dmode == DISPLAY_FRAME_TIME) {
-		return rtos(p_time);
+		return rtos(p_time * 1000).pad_decimals(2) + " ms";
 	} else if (dmode == DISPLAY_AVERAGE_TIME) {
-		if (p_calls == 0)
-			return "0";
-		else
-			return rtos(p_time / p_calls);
+		if (p_calls == 0) {
+			return "0.00 ms";
+		} else {
+			return rtos((p_time / p_calls) * 1000).pad_decimals(2) + " ms";
+		}
 	} else if (dmode == DISPLAY_FRAME_PERCENT) {
 		return _get_percent_txt(p_time, m.frame_time);
 	} else if (dmode == DISPLAY_PHYSICS_FRAME_PERCENT) {
-
 		return _get_percent_txt(p_time, m.physics_frame_time);
 	}
 
@@ -130,7 +127,6 @@ String EditorProfiler::_get_time_as_text(Metric &m, float p_time, int p_calls) {
 }
 
 Color EditorProfiler::_get_color_from_signature(const StringName &p_signature) const {
-
 	Color bc = get_color("error_color", "Editor");
 	double rot = ABS(double(p_signature.hash()) / double(0x7FFFFFFF));
 	Color c;
@@ -139,20 +135,22 @@ Color EditorProfiler::_get_color_from_signature(const StringName &p_signature) c
 }
 
 void EditorProfiler::_item_edited() {
-
-	if (updating_frame)
+	if (updating_frame) {
 		return;
+	}
 
 	TreeItem *item = variables->get_edited();
-	if (!item)
+	if (!item) {
 		return;
+	}
 	StringName signature = item->get_metadata(0);
 	bool checked = item->is_checked(0);
 
-	if (checked)
+	if (checked) {
 		plot_sigs.insert(signature);
-	else
+	} else {
 		plot_sigs.erase(signature);
+	}
 
 	if (!frame_delay->is_processing()) {
 		frame_delay->set_wait_time(0.1);
@@ -163,13 +161,10 @@ void EditorProfiler::_item_edited() {
 }
 
 void EditorProfiler::_update_plot() {
-
-	int w = graph->get_size().width;
-	int h = graph->get_size().height;
-
+	const int w = graph->get_size().width;
+	const int h = graph->get_size().height;
 	bool reset_texture = false;
-
-	int desired_len = w * h * 4;
+	const int desired_len = w * h * 4;
 
 	if (graph_image.size() != desired_len) {
 		reset_texture = true;
@@ -177,33 +172,34 @@ void EditorProfiler::_update_plot() {
 	}
 
 	PoolVector<uint8_t>::Write wr = graph_image.write();
+	const Color background_color = get_color("dark_color_2", "Editor");
 
-	//clear
+	// Clear the previous frame and set the background color.
 	for (int i = 0; i < desired_len; i += 4) {
-		wr[i + 0] = 0;
-		wr[i + 1] = 0;
-		wr[i + 2] = 0;
+		wr[i + 0] = Math::fast_ftoi(background_color.r * 255);
+		wr[i + 1] = Math::fast_ftoi(background_color.g * 255);
+		wr[i + 2] = Math::fast_ftoi(background_color.b * 255);
 		wr[i + 3] = 255;
 	}
 
 	//find highest value
 
-	bool use_self = display_time->get_selected() == DISPLAY_SELF_TIME;
+	const bool use_self = display_time->get_selected() == DISPLAY_SELF_TIME;
 	float highest = 0;
 
 	for (int i = 0; i < frame_metrics.size(); i++) {
-		Metric &m = frame_metrics[i];
-		if (!m.valid)
+		const Metric &m = frame_metrics[i];
+		if (!m.valid) {
 			continue;
+		}
 
 		for (Set<StringName>::Element *E = plot_sigs.front(); E; E = E->next()) {
-
-			Map<StringName, Metric::Category *>::Element *F = m.category_ptrs.find(E->get());
+			const Map<StringName, Metric::Category *>::Element *F = m.category_ptrs.find(E->get());
 			if (F) {
 				highest = MAX(F->get()->total_time, highest);
 			}
 
-			Map<StringName, Metric::Category::Item *>::Element *G = m.item_ptrs.find(E->get());
+			const Map<StringName, Metric::Category::Item *>::Element *G = m.item_ptrs.find(E->get());
 			if (G) {
 				if (use_self) {
 					highest = MAX(G->get()->self, highest);
@@ -227,10 +223,7 @@ void EditorProfiler::_update_plot() {
 		Map<StringName, int> plot_prev;
 		//Map<StringName,int> plot_max;
 
-		uint64_t time = OS::get_singleton()->get_ticks_usec();
-
 		for (int i = 0; i < w; i++) {
-
 			for (int j = 0; j < h * 4; j++) {
 				column[j] = 0;
 			}
@@ -240,15 +233,14 @@ void EditorProfiler::_update_plot() {
 			if (next > frame_metrics.size()) {
 				next = frame_metrics.size();
 			}
-			if (next == current)
+			if (next == current) {
 				next = current + 1; //just because for loop must work
+			}
 
 			for (Set<StringName>::Element *E = plot_sigs.front(); E; E = E->next()) {
-
 				int plot_pos = -1;
 
 				for (int j = current; j < next; j++) {
-
 					//wrap
 					int idx = last_metric + 1 + j;
 					while (idx >= frame_metrics.size()) {
@@ -256,18 +248,19 @@ void EditorProfiler::_update_plot() {
 					}
 
 					//get
-					Metric &m = frame_metrics[idx];
-					if (m.valid == false)
+					const Metric &m = frame_metrics[idx];
+					if (!m.valid) {
 						continue; //skip because invalid
+					}
 
 					float value = 0;
 
-					Map<StringName, Metric::Category *>::Element *F = m.category_ptrs.find(E->get());
+					const Map<StringName, Metric::Category *>::Element *F = m.category_ptrs.find(E->get());
 					if (F) {
 						value = F->get()->total_time;
 					}
 
-					Map<StringName, Metric::Category::Item *>::Element *G = m.item_ptrs.find(E->get());
+					const Map<StringName, Metric::Category::Item *>::Element *G = m.item_ptrs.find(E->get());
 					if (G) {
 						if (use_self) {
 							value = G->get()->self;
@@ -294,7 +287,6 @@ void EditorProfiler::_update_plot() {
 				}
 
 				if (prev_plot != -1 && plot_pos == -1) {
-
 					plot_pos = prev_plot;
 				}
 
@@ -312,7 +304,6 @@ void EditorProfiler::_update_plot() {
 				Color col = _get_color_from_signature(E->get());
 
 				for (int j = prev_plot; j <= plot_pos; j++) {
-
 					column[j * 4 + 0] += Math::fast_ftoi(CLAMP(col.r * 255, 0, 255));
 					column[j * 4 + 1] += Math::fast_ftoi(CLAMP(col.g * 255, 0, 255));
 					column[j * 4 + 2] += Math::fast_ftoi(CLAMP(col.b * 255, 0, 255));
@@ -321,38 +312,35 @@ void EditorProfiler::_update_plot() {
 			}
 
 			for (int j = 0; j < h * 4; j += 4) {
-
-				int a = column[j + 3];
+				const int a = column[j + 3];
 				if (a > 0) {
 					column[j + 0] /= a;
 					column[j + 1] /= a;
 					column[j + 2] /= a;
 				}
 
-				uint8_t r = uint8_t(column[j + 0]);
-				uint8_t g = uint8_t(column[j + 1]);
-				uint8_t b = uint8_t(column[j + 2]);
+				const uint8_t red = uint8_t(column[j + 0]);
+				const uint8_t green = uint8_t(column[j + 1]);
+				const uint8_t blue = uint8_t(column[j + 2]);
+				const bool is_filled = red >= 1 || green >= 1 || blue >= 1;
+				const int widx = ((j >> 2) * w + i) * 4;
 
-				int widx = ((j >> 2) * w + i) * 4;
-				wr[widx + 0] = r;
-				wr[widx + 1] = g;
-				wr[widx + 2] = b;
+				// If the pixel isn't filled by any profiler line, apply the background color instead.
+				wr[widx + 0] = is_filled ? red : Math::fast_ftoi(background_color.r * 255);
+				wr[widx + 1] = is_filled ? green : Math::fast_ftoi(background_color.g * 255);
+				wr[widx + 2] = is_filled ? blue : Math::fast_ftoi(background_color.b * 255);
 				wr[widx + 3] = 255;
 			}
 		}
-
-		time = OS::get_singleton()->get_ticks_usec() - time;
-		//print_line("Taken: "+rtos(USEC_TO_SEC(time)));
 	}
 
-	wr = PoolVector<uint8_t>::Write();
+	wr.release();
 
 	Ref<Image> img;
 	img.instance();
-	img->create(w, h, 0, Image::FORMAT_RGBA8, graph_image);
+	img->create(w, h, false, Image::FORMAT_RGBA8, graph_image);
 
 	if (reset_texture) {
-
 		if (graph_texture.is_null()) {
 			graph_texture.instance();
 		}
@@ -366,7 +354,6 @@ void EditorProfiler::_update_plot() {
 }
 
 void EditorProfiler::_update_frame() {
-
 	int cursor_metric = _get_cursor_index();
 
 	ERR_FAIL_INDEX(cursor_metric, frame_metrics.size());
@@ -375,12 +362,11 @@ void EditorProfiler::_update_frame() {
 	variables->clear();
 
 	TreeItem *root = variables->create_item();
-	Metric &m = frame_metrics[cursor_metric];
+	const Metric &m = frame_metrics[cursor_metric];
 
 	int dtime = display_time->get_selected();
 
 	for (int i = 0; i < m.categories.size(); i++) {
-
 		TreeItem *category = variables->create_item(root);
 		category->set_cell_mode(0, TreeItem::CELL_MODE_CHECK);
 		category->set_editable(0, true);
@@ -393,8 +379,8 @@ void EditorProfiler::_update_frame() {
 			category->set_custom_color(0, _get_color_from_signature(m.categories[i].signature));
 		}
 
-		for (int j = 0; j < m.categories[i].items.size(); j++) {
-			Metric::Category::Item &it = m.categories[i].items[j];
+		for (int j = m.categories[i].items.size() - 1; j >= 0; j--) {
+			const Metric::Category::Item &it = m.categories[i].items[j];
 
 			TreeItem *item = variables->create_item(category);
 			item->set_cell_mode(0, TreeItem::CELL_MODE_CHECK);
@@ -403,7 +389,8 @@ void EditorProfiler::_update_frame() {
 			item->set_metadata(0, it.signature);
 			item->set_metadata(1, it.script);
 			item->set_metadata(2, it.line);
-			item->set_tooltip(0, it.script + ":" + itos(it.line));
+			item->set_text_align(2, TreeItem::ALIGN_RIGHT);
+			item->set_tooltip(0, it.name + "\n" + it.script + ":" + itos(it.line));
 
 			float time = dtime == DISPLAY_SELF_TIME ? it.self : it.total;
 
@@ -422,35 +409,38 @@ void EditorProfiler::_update_frame() {
 }
 
 void EditorProfiler::_activate_pressed() {
-
 	if (activate->is_pressed()) {
-		clear();
 		activate->set_icon(get_icon("Stop", "EditorIcons"));
-		activate->set_text(TTR("Stop Profiling"));
+		activate->set_text(TTR("Stop"));
 	} else {
 		activate->set_icon(get_icon("Play", "EditorIcons"));
-		activate->set_text(TTR("Start Profiling"));
+		activate->set_text(TTR("Start"));
 	}
 	emit_signal("enable_profiling", activate->is_pressed());
 }
 
-void EditorProfiler::_notification(int p_what) {
+void EditorProfiler::_clear_pressed() {
+	clear();
+	_update_plot();
+}
 
+void EditorProfiler::_notification(int p_what) {
 	if (p_what == NOTIFICATION_ENTER_TREE) {
 		activate->set_icon(get_icon("Play", "EditorIcons"));
+		clear_button->set_icon(get_icon("Clear", "EditorIcons"));
 	}
 }
 
 void EditorProfiler::_graph_tex_draw() {
-
-	if (last_metric < 0)
+	if (last_metric < 0) {
 		return;
+	}
 	if (seeking) {
-
 		int max_frames = frame_metrics.size();
 		int frame = cursor_metric_edit->get_value() - (frame_metrics[last_metric].frame_number - max_frames + 1);
-		if (frame < 0)
+		if (frame < 0) {
 			frame = 0;
+		}
 
 		int cur_x = frame * graph->get_size().x / max_frames;
 
@@ -458,11 +448,11 @@ void EditorProfiler::_graph_tex_draw() {
 	}
 
 	if (hover_metric != -1 && frame_metrics[hover_metric].valid) {
-
 		int max_frames = frame_metrics.size();
 		int frame = frame_metrics[hover_metric].frame_number - (frame_metrics[last_metric].frame_number - max_frames + 1);
-		if (frame < 0)
+		if (frame < 0) {
 			frame = 0;
+		}
 
 		int cur_x = frame * graph->get_size().x / max_frames;
 
@@ -471,23 +461,23 @@ void EditorProfiler::_graph_tex_draw() {
 }
 
 void EditorProfiler::_graph_tex_mouse_exit() {
-
 	hover_metric = -1;
 	graph->update();
 }
 
 void EditorProfiler::_cursor_metric_changed(double) {
-	if (updating_frame)
+	if (updating_frame) {
 		return;
+	}
 
 	graph->update();
 	_update_frame();
 }
 
 void EditorProfiler::_graph_tex_input(const Ref<InputEvent> &p_ev) {
-
-	if (last_metric < 0)
+	if (last_metric < 0) {
 		return;
+	}
 
 	Ref<InputEventMouse> me = p_ev;
 	Ref<InputEventMouseButton> mb = p_ev;
@@ -496,7 +486,6 @@ void EditorProfiler::_graph_tex_input(const Ref<InputEvent> &p_ev) {
 	if (
 			(mb.is_valid() && mb->get_button_index() == BUTTON_LEFT && mb->is_pressed()) ||
 			(mm.is_valid())) {
-
 		int x = me->get_position().x;
 		x = x * frame_metrics.size() / graph->get_size().width;
 
@@ -517,7 +506,6 @@ void EditorProfiler::_graph_tex_input(const Ref<InputEvent> &p_ev) {
 		}
 
 		if (show_hover) {
-
 			hover_metric = metric;
 
 		} else {
@@ -531,19 +519,20 @@ void EditorProfiler::_graph_tex_input(const Ref<InputEvent> &p_ev) {
 			//metric may be invalid, so look for closest metric that is valid, this makes snap feel better
 			bool valid = false;
 			for (int i = 0; i < frame_metrics.size(); i++) {
-
 				if (frame_metrics[metric].valid) {
 					valid = true;
 					break;
 				}
 
 				metric++;
-				if (metric >= frame_metrics.size())
+				if (metric >= frame_metrics.size()) {
 					metric = 0;
+				}
 			}
 
-			if (valid)
+			if (valid) {
 				cursor_metric_edit->set_value(frame_metrics[metric].frame_number);
+			}
 
 			updating_frame = false;
 
@@ -566,11 +555,12 @@ void EditorProfiler::_graph_tex_input(const Ref<InputEvent> &p_ev) {
 }
 
 int EditorProfiler::_get_cursor_index() const {
-
-	if (last_metric < 0)
+	if (last_metric < 0) {
 		return 0;
-	if (!frame_metrics[last_metric].valid)
+	}
+	if (!frame_metrics[last_metric].valid) {
 		return 0;
+	}
 
 	int diff = (frame_metrics[last_metric].frame_number - cursor_metric_edit->get_value());
 
@@ -583,22 +573,20 @@ int EditorProfiler::_get_cursor_index() const {
 }
 
 void EditorProfiler::disable_seeking() {
-
 	seeking = false;
 	graph->update();
 }
 
 void EditorProfiler::_combo_changed(int) {
-
 	_update_frame();
 	_update_plot();
 }
 
 void EditorProfiler::_bind_methods() {
-
 	ClassDB::bind_method(D_METHOD("_update_frame"), &EditorProfiler::_update_frame);
 	ClassDB::bind_method(D_METHOD("_update_plot"), &EditorProfiler::_update_plot);
 	ClassDB::bind_method(D_METHOD("_activate_pressed"), &EditorProfiler::_activate_pressed);
+	ClassDB::bind_method(D_METHOD("_clear_pressed"), &EditorProfiler::_clear_pressed);
 	ClassDB::bind_method(D_METHOD("_graph_tex_draw"), &EditorProfiler::_graph_tex_draw);
 	ClassDB::bind_method(D_METHOD("_graph_tex_input"), &EditorProfiler::_graph_tex_input);
 	ClassDB::bind_method(D_METHOD("_graph_tex_mouse_exit"), &EditorProfiler::_graph_tex_mouse_exit);
@@ -611,7 +599,6 @@ void EditorProfiler::_bind_methods() {
 }
 
 void EditorProfiler::set_enabled(bool p_enable) {
-
 	activate->set_disabled(!p_enable);
 }
 
@@ -619,21 +606,94 @@ bool EditorProfiler::is_profiling() {
 	return activate->is_pressed();
 }
 
-EditorProfiler::EditorProfiler() {
+Vector<Vector<String>> EditorProfiler::get_data_as_csv() const {
+	Vector<Vector<String>> res;
 
+	if (frame_metrics.empty()) {
+		return res;
+	}
+
+	// Different metrics may contain different number of categories.
+	Set<StringName> possible_signatures;
+	for (int i = 0; i < frame_metrics.size(); i++) {
+		const Metric &m = frame_metrics[i];
+		if (!m.valid) {
+			continue;
+		}
+		for (Map<StringName, Metric::Category *>::Element *E = m.category_ptrs.front(); E; E = E->next()) {
+			possible_signatures.insert(E->key());
+		}
+		for (Map<StringName, Metric::Category::Item *>::Element *E = m.item_ptrs.front(); E; E = E->next()) {
+			possible_signatures.insert(E->key());
+		}
+	}
+
+	// Generate CSV header and cache indices.
+	Map<StringName, int> sig_map;
+	Vector<String> signatures;
+	signatures.resize(possible_signatures.size());
+	int sig_index = 0;
+	for (const Set<StringName>::Element *E = possible_signatures.front(); E; E = E->next()) {
+		signatures.write[sig_index] = E->get();
+		sig_map[E->get()] = sig_index;
+		sig_index++;
+	}
+	res.push_back(signatures);
+
+	// values
+	Vector<String> values;
+
+	int index = last_metric;
+
+	for (int i = 0; i < frame_metrics.size(); i++) {
+		++index;
+
+		if (index >= frame_metrics.size()) {
+			index = 0;
+		}
+
+		const Metric &m = frame_metrics[index];
+
+		if (!m.valid) {
+			continue;
+		}
+
+		// Don't keep old values since there may be empty cells.
+		values.clear();
+		values.resize(possible_signatures.size());
+
+		for (Map<StringName, Metric::Category *>::Element *E = m.category_ptrs.front(); E; E = E->next()) {
+			values.write[sig_map[E->key()]] = String::num_real(E->value()->total_time);
+		}
+		for (Map<StringName, Metric::Category::Item *>::Element *E = m.item_ptrs.front(); E; E = E->next()) {
+			values.write[sig_map[E->key()]] = String::num_real(E->value()->total);
+		}
+
+		res.push_back(values);
+	}
+
+	return res;
+}
+
+EditorProfiler::EditorProfiler() {
 	HBoxContainer *hb = memnew(HBoxContainer);
 	add_child(hb);
 	activate = memnew(Button);
 	activate->set_toggle_mode(true);
-	activate->set_text(TTR("Start Profiling"));
+	activate->set_text(TTR("Start"));
 	activate->connect("pressed", this, "_activate_pressed");
 	hb->add_child(activate);
 
+	clear_button = memnew(Button);
+	clear_button->set_text(TTR("Clear"));
+	clear_button->connect("pressed", this, "_clear_pressed");
+	hb->add_child(clear_button);
+
 	hb->add_child(memnew(Label(TTR("Measure:"))));
 
 	display_mode = memnew(OptionButton);
-	display_mode->add_item(TTR("Frame Time (sec)"));
-	display_mode->add_item(TTR("Average Time (sec)"));
+	display_mode->add_item(TTR("Frame Time (ms)"));
+	display_mode->add_item(TTR("Average Time (ms)"));
 	display_mode->add_item(TTR("Frame %"));
 	display_mode->add_item(TTR("Physics Frame %"));
 	display_mode->connect("item_selected", this, "_combo_changed");
@@ -645,6 +705,7 @@ EditorProfiler::EditorProfiler() {
 	display_time = memnew(OptionButton);
 	display_time->add_item(TTR("Inclusive"));
 	display_time->add_item(TTR("Self"));
+	display_time->set_tooltip(TTR("Inclusive: Includes time from other functions called by this function.\nUse this to spot bottlenecks.\n\nSelf: Only count the time spent in the function itself, not in other functions called by that function.\nUse this to find individual functions to optimize."));
 	display_time->connect("item_selected", this, "_combo_changed");
 
 	hb->add_child(display_time);
@@ -665,7 +726,7 @@ EditorProfiler::EditorProfiler() {
 	h_split->set_v_size_flags(SIZE_EXPAND_FILL);
 
 	variables = memnew(Tree);
-	variables->set_custom_minimum_size(Size2(300, 0) * EDSCALE);
+	variables->set_custom_minimum_size(Size2(320, 0) * EDSCALE);
 	variables->set_hide_folding(true);
 	h_split->add_child(variables);
 	variables->set_hide_root(true);
@@ -673,10 +734,10 @@ EditorProfiler::EditorProfiler() {
 	variables->set_column_titles_visible(true);
 	variables->set_column_title(0, TTR("Name"));
 	variables->set_column_expand(0, true);
-	variables->set_column_min_width(0, 60);
+	variables->set_column_min_width(0, 60 * EDSCALE);
 	variables->set_column_title(1, TTR("Time"));
 	variables->set_column_expand(1, false);
-	variables->set_column_min_width(1, 60 * EDSCALE);
+	variables->set_column_min_width(1, 100 * EDSCALE);
 	variables->set_column_title(2, TTR("Calls"));
 	variables->set_column_expand(2, false);
 	variables->set_column_min_width(2, 60 * EDSCALE);
@@ -685,7 +746,6 @@ EditorProfiler::EditorProfiler() {
 	graph = memnew(TextureRect);
 	graph->set_expand(true);
 	graph->set_mouse_filter(MOUSE_FILTER_STOP);
-	//graph->set_ignore_mouse(false);
 	graph->connect("draw", this, "_graph_tex_draw");
 	graph->connect("gui_input", this, "_graph_tex_input");
 	graph->connect("mouse_exited", this, "_graph_tex_mouse_exit");
@@ -696,12 +756,9 @@ EditorProfiler::EditorProfiler() {
 	int metric_size = CLAMP(int(EDITOR_DEF("debugger/profiler_frame_history_size", 600)), 60, 1024);
 	frame_metrics.resize(metric_size);
 	last_metric = -1;
-	//cursor_metric=-1;
 	hover_metric = -1;
 
-	EDITOR_DEF("debugger/profiler_frame_max_functions", 64);
-
-	//display_mode=DISPLAY_FRAME_TIME;
+	EDITOR_DEF("debugger/profiler_frame_max_functions", 512);
 
 	frame_delay = memnew(Timer);
 	frame_delay->set_wait_time(0.1);
@@ -720,6 +777,4 @@ EditorProfiler::EditorProfiler() {
 
 	seeking = false;
 	graph_height = 1;
-
-	//activate->set_disabled(true);
 }
