@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -30,18 +30,26 @@
 
 #include "gdnative/pool_arrays.h"
 
-#include "array.h"
+#include "core/array.h"
+#include "core/pool_vector.h"
 #include "core/variant.h"
-#include "dvector.h"
 
 #include "core/color.h"
-#include "core/math/math_2d.h"
+#include "core/math/vector2.h"
 #include "core/math/vector3.h"
 
 #ifdef __cplusplus
 extern "C" {
 #endif
 
+static_assert(sizeof(godot_pool_byte_array) == sizeof(PoolVector<uint8_t>), "PoolVector<uint8_t> size mismatch");
+static_assert(sizeof(godot_pool_int_array) == sizeof(PoolVector<godot_int>), "PoolVector<godot_int> size mismatch");
+static_assert(sizeof(godot_pool_real_array) == sizeof(PoolVector<godot_real>), "PoolVector<godot_real> size mismatch");
+static_assert(sizeof(godot_pool_string_array) == sizeof(PoolVector<String>), "PoolVector<String> size mismatch");
+static_assert(sizeof(godot_pool_vector2_array) == sizeof(PoolVector<Vector2>), "PoolVector<Vector2> size mismatch");
+static_assert(sizeof(godot_pool_vector3_array) == sizeof(PoolVector<Vector3>), "PoolVector<Vector3> size mismatch");
+static_assert(sizeof(godot_pool_color_array) == sizeof(PoolVector<Color>), "PoolVector<Color> size mismatch");
+
 #define memnew_placement_custom(m_placement, m_class, m_constr) _post_initialize(new (m_placement, sizeof(m_class), "") m_constr)
 
 // byte
@@ -129,6 +137,11 @@ godot_int GDAPI godot_pool_byte_array_size(const godot_pool_byte_array *p_self)
 	return self->size();
 }
 
+godot_bool GDAPI godot_pool_byte_array_empty(const godot_pool_byte_array *p_self) {
+	const PoolVector<uint8_t> *self = (const PoolVector<uint8_t> *)p_self;
+	return self->empty();
+}
+
 void GDAPI godot_pool_byte_array_destroy(godot_pool_byte_array *p_self) {
 	((PoolVector<uint8_t> *)p_self)->~PoolVector();
 }
@@ -218,6 +231,11 @@ godot_int GDAPI godot_pool_int_array_size(const godot_pool_int_array *p_self) {
 	return self->size();
 }
 
+godot_bool GDAPI godot_pool_int_array_empty(const godot_pool_int_array *p_self) {
+	const PoolVector<godot_int> *self = (const PoolVector<godot_int> *)p_self;
+	return self->empty();
+}
+
 void GDAPI godot_pool_int_array_destroy(godot_pool_int_array *p_self) {
 	((PoolVector<godot_int> *)p_self)->~PoolVector();
 }
@@ -307,6 +325,11 @@ godot_int GDAPI godot_pool_real_array_size(const godot_pool_real_array *p_self)
 	return self->size();
 }
 
+godot_bool GDAPI godot_pool_real_array_empty(const godot_pool_real_array *p_self) {
+	const PoolVector<godot_real> *self = (const PoolVector<godot_real> *)p_self;
+	return self->empty();
+}
+
 void GDAPI godot_pool_real_array_destroy(godot_pool_real_array *p_self) {
 	((PoolVector<godot_real> *)p_self)->~PoolVector();
 }
@@ -404,6 +427,11 @@ godot_int GDAPI godot_pool_string_array_size(const godot_pool_string_array *p_se
 	return self->size();
 }
 
+godot_bool GDAPI godot_pool_string_array_empty(const godot_pool_string_array *p_self) {
+	const PoolVector<String> *self = (const PoolVector<String> *)p_self;
+	return self->empty();
+}
+
 void GDAPI godot_pool_string_array_destroy(godot_pool_string_array *p_self) {
 	((PoolVector<String> *)p_self)->~PoolVector();
 }
@@ -500,6 +528,11 @@ godot_int GDAPI godot_pool_vector2_array_size(const godot_pool_vector2_array *p_
 	return self->size();
 }
 
+godot_bool GDAPI godot_pool_vector2_array_empty(const godot_pool_vector2_array *p_self) {
+	const PoolVector<Vector2> *self = (const PoolVector<Vector2> *)p_self;
+	return self->empty();
+}
+
 void GDAPI godot_pool_vector2_array_destroy(godot_pool_vector2_array *p_self) {
 	((PoolVector<Vector2> *)p_self)->~PoolVector();
 }
@@ -596,6 +629,11 @@ godot_int GDAPI godot_pool_vector3_array_size(const godot_pool_vector3_array *p_
 	return self->size();
 }
 
+godot_bool GDAPI godot_pool_vector3_array_empty(const godot_pool_vector3_array *p_self) {
+	const PoolVector<Vector3> *self = (const PoolVector<Vector3> *)p_self;
+	return self->empty();
+}
+
 void GDAPI godot_pool_vector3_array_destroy(godot_pool_vector3_array *p_self) {
 	((PoolVector<Vector3> *)p_self)->~PoolVector();
 }
@@ -692,6 +730,11 @@ godot_int GDAPI godot_pool_color_array_size(const godot_pool_color_array *p_self
 	return self->size();
 }
 
+godot_bool GDAPI godot_pool_color_array_empty(const godot_pool_color_array *p_self) {
+	const PoolVector<Color> *self = (const PoolVector<Color> *)p_self;
+	return self->empty();
+}
+
 void GDAPI godot_pool_color_array_destroy(godot_pool_color_array *p_self) {
 	((PoolVector<Color> *)p_self)->~PoolVector();
 }
