@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -30,7 +30,7 @@
 
 #include "gdnative/string.h"
 
-#include "core/string_db.h"
+#include "core/string_name.h"
 #include "core/ustring.h"
 #include "core/variant.h"
 
@@ -40,6 +40,10 @@
 extern "C" {
 #endif
 
+static_assert(sizeof(godot_char_string) == sizeof(CharString), "CharString size mismatch");
+static_assert(sizeof(godot_string) == sizeof(String), "String size mismatch");
+static_assert(sizeof(godot_char_type) == sizeof(CharType), "CharType size mismatch");
+
 godot_int GDAPI godot_char_string_length(const godot_char_string *p_cs) {
 	const CharString *cs = (const CharString *)p_cs;
 
@@ -74,7 +78,7 @@ void GDAPI godot_string_new_with_wide_string(godot_string *r_dest, const wchar_t
 	memnew_placement(dest, String(p_contents, p_size));
 }
 
-wchar_t GDAPI *godot_string_operator_index(godot_string *p_self, const godot_int p_idx) {
+const wchar_t GDAPI *godot_string_operator_index(godot_string *p_self, const godot_int p_idx) {
 	String *self = (String *)p_self;
 	return &(self->operator[](p_idx));
 }
@@ -186,6 +190,20 @@ godot_bool GDAPI godot_string_ends_with(const godot_string *p_self, const godot_
 	return self->ends_with(*string);
 }
 
+godot_int GDAPI godot_string_count(const godot_string *p_self, godot_string p_what, godot_int p_from, godot_int p_to) {
+	const String *self = (const String *)p_self;
+	String *what = (String *)&p_what;
+
+	return self->count(*what, p_from, p_to);
+}
+
+godot_int GDAPI godot_string_countn(const godot_string *p_self, godot_string p_what, godot_int p_from, godot_int p_to) {
+	const String *self = (const String *)p_self;
+	String *what = (String *)&p_what;
+
+	return self->countn(*what, p_from, p_to);
+}
+
 godot_int GDAPI godot_string_find(const godot_string *p_self, godot_string p_what) {
 	const String *self = (const String *)p_self;
 	String *what = (String *)&p_what;
@@ -207,7 +225,7 @@ godot_int GDAPI godot_string_findmk(const godot_string *p_self, const godot_arra
 	Array *keys_proxy = (Array *)p_keys;
 	keys.resize(keys_proxy->size());
 	for (int i = 0; i < keys_proxy->size(); i++) {
-		keys[i] = (*keys_proxy)[i];
+		keys.write[i] = (*keys_proxy)[i];
 	}
 
 	return self->findmk(keys);
@@ -220,7 +238,7 @@ godot_int GDAPI godot_string_findmk_from(const godot_string *p_self, const godot
 	Array *keys_proxy = (Array *)p_keys;
 	keys.resize(keys_proxy->size());
 	for (int i = 0; i < keys_proxy->size(); i++) {
-		keys[i] = (*keys_proxy)[i];
+		keys.write[i] = (*keys_proxy)[i];
 	}
 
 	return self->findmk(keys, p_from);
@@ -233,7 +251,7 @@ godot_int GDAPI godot_string_findmk_from_in_place(const godot_string *p_self, co
 	Array *keys_proxy = (Array *)p_keys;
 	keys.resize(keys_proxy->size());
 	for (int i = 0; i < keys_proxy->size(); i++) {
-		keys[i] = (*keys_proxy)[i];
+		keys.write[i] = (*keys_proxy)[i];
 	}
 
 	return self->findmk(keys, p_from, r_key);
@@ -696,7 +714,7 @@ godot_array GDAPI godot_string_split_floats_mk(const godot_string *p_self, const
 	Array *splitter_proxy = (Array *)p_splitters;
 	splitters.resize(splitter_proxy->size());
 	for (int i = 0; i < splitter_proxy->size(); i++) {
-		splitters[i] = (*splitter_proxy)[i];
+		splitters.write[i] = (*splitter_proxy)[i];
 	}
 
 	godot_array result;
@@ -719,7 +737,7 @@ godot_array GDAPI godot_string_split_floats_mk_allows_empty(const godot_string *
 	Array *splitter_proxy = (Array *)p_splitters;
 	splitters.resize(splitter_proxy->size());
 	for (int i = 0; i < splitter_proxy->size(); i++) {
-		splitters[i] = (*splitter_proxy)[i];
+		splitters.write[i] = (*splitter_proxy)[i];
 	}
 
 	godot_array result;
@@ -774,7 +792,7 @@ godot_array GDAPI godot_string_split_ints_mk(const godot_string *p_self, const g
 	Array *splitter_proxy = (Array *)p_splitters;
 	splitters.resize(splitter_proxy->size());
 	for (int i = 0; i < splitter_proxy->size(); i++) {
-		splitters[i] = (*splitter_proxy)[i];
+		splitters.write[i] = (*splitter_proxy)[i];
 	}
 
 	godot_array result;
@@ -797,7 +815,7 @@ godot_array GDAPI godot_string_split_ints_mk_allows_empty(const godot_string *p_
 	Array *splitter_proxy = (Array *)p_splitters;
 	splitters.resize(splitter_proxy->size());
 	for (int i = 0; i < splitter_proxy->size(); i++) {
-		splitters[i] = (*splitter_proxy)[i];
+		splitters.write[i] = (*splitter_proxy)[i];
 	}
 
 	godot_array result;
@@ -1086,7 +1104,7 @@ godot_string GDAPI godot_string_get_file(const godot_string *p_self) {
 	return result;
 }
 
-godot_string GDAPI godot_string_humanize_size(size_t p_size) {
+godot_string GDAPI godot_string_humanize_size(uint64_t p_size) {
 	godot_string result;
 	String return_value = String::humanize_size(p_size);
 	memnew_placement(&result, String(return_value));
@@ -1285,6 +1303,64 @@ godot_bool GDAPI godot_string_is_valid_ip_address(const godot_string *p_self) {
 	return self->is_valid_ip_address();
 }
 
+godot_string GDAPI godot_string_dedent(const godot_string *p_self) {
+	const String *self = (const String *)p_self;
+	godot_string result;
+	String return_value = self->dedent();
+	memnew_placement(&result, String(return_value));
+
+	return result;
+}
+
+godot_string GDAPI godot_string_trim_prefix(const godot_string *p_self, const godot_string *p_prefix) {
+	const String *self = (const String *)p_self;
+	String *prefix = (String *)p_prefix;
+	godot_string result;
+	String return_value = self->trim_prefix(*prefix);
+	memnew_placement(&result, String(return_value));
+
+	return result;
+}
+
+godot_string GDAPI godot_string_trim_suffix(const godot_string *p_self, const godot_string *p_suffix) {
+	const String *self = (const String *)p_self;
+	String *suffix = (String *)p_suffix;
+	godot_string result;
+	String return_value = self->trim_suffix(*suffix);
+	memnew_placement(&result, String(return_value));
+
+	return result;
+}
+
+godot_string GDAPI godot_string_rstrip(const godot_string *p_self, const godot_string *p_chars) {
+	const String *self = (const String *)p_self;
+	String *chars = (String *)p_chars;
+	godot_string result;
+	String return_value = self->rstrip(*chars);
+	memnew_placement(&result, String(return_value));
+
+	return result;
+}
+
+godot_pool_string_array GDAPI godot_string_rsplit(const godot_string *p_self, const godot_string *p_divisor,
+		const godot_bool p_allow_empty, const godot_int p_maxsplit) {
+	const String *self = (const String *)p_self;
+	String *divisor = (String *)p_divisor;
+
+	godot_pool_string_array result;
+	memnew_placement(&result, PoolStringArray);
+	PoolStringArray *proxy = (PoolStringArray *)&result;
+	PoolStringArray::Write proxy_writer = proxy->write();
+	Vector<String> tmp_result = self->rsplit(*divisor, p_allow_empty, p_maxsplit);
+	proxy->resize(tmp_result.size());
+
+	for (int i = 0; i < tmp_result.size(); i++) {
+		proxy_writer[i] = tmp_result[i];
+	}
+
+	return result;
+}
+
 #ifdef __cplusplus
 }
 #endif
