@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -31,7 +31,6 @@
 #include "visual_script_expression.h"
 
 bool VisualScriptExpression::_set(const StringName &p_name, const Variant &p_value) {
-
 	if (String(p_name) == "expression") {
 		expression = p_value;
 		expression_dirty = true;
@@ -52,15 +51,14 @@ bool VisualScriptExpression::_set(const StringName &p_name, const Variant &p_val
 	}
 
 	if (String(p_name) == "input_count") {
-
 		int from = inputs.size();
 		inputs.resize(int(p_value));
 		for (int i = from; i < inputs.size(); i++) {
-			inputs[i].name = String::chr('a' + i);
+			inputs.write[i].name = String::chr('a' + i);
 			if (from == 0) {
-				inputs[i].type = output_type;
+				inputs.write[i].type = output_type;
 			} else {
-				inputs[i].type = inputs[from - 1].type;
+				inputs.write[i].type = inputs[from - 1].type;
 			}
 		}
 		expression_dirty = true;
@@ -70,18 +68,15 @@ bool VisualScriptExpression::_set(const StringName &p_name, const Variant &p_val
 	}
 
 	if (String(p_name).begins_with("input_")) {
-
 		int idx = String(p_name).get_slicec('_', 1).get_slicec('/', 0).to_int();
 		ERR_FAIL_INDEX_V(idx, inputs.size(), false);
 
 		String what = String(p_name).get_slice("/", 1);
 
 		if (what == "type") {
-
-			inputs[idx].type = Variant::Type(int(p_value));
+			inputs.write[idx].type = Variant::Type(int(p_value));
 		} else if (what == "name") {
-
-			inputs[idx].name = p_value;
+			inputs.write[idx].name = p_value;
 		} else {
 			return false;
 		}
@@ -95,7 +90,6 @@ bool VisualScriptExpression::_set(const StringName &p_name, const Variant &p_val
 }
 
 bool VisualScriptExpression::_get(const StringName &p_name, Variant &r_ret) const {
-
 	if (String(p_name) == "expression") {
 		r_ret = expression;
 		return true;
@@ -117,17 +111,14 @@ bool VisualScriptExpression::_get(const StringName &p_name, Variant &r_ret) cons
 	}
 
 	if (String(p_name).begins_with("input_")) {
-
 		int idx = String(p_name).get_slicec('_', 1).get_slicec('/', 0).to_int();
 		ERR_FAIL_INDEX_V(idx, inputs.size(), false);
 
 		String what = String(p_name).get_slice("/", 1);
 
 		if (what == "type") {
-
 			r_ret = inputs[idx].type;
 		} else if (what == "name") {
-
 			r_ret = inputs[idx].name;
 		} else {
 			return false;
@@ -139,7 +130,6 @@ bool VisualScriptExpression::_get(const StringName &p_name, Variant &r_ret) cons
 	return false;
 }
 void VisualScriptExpression::_get_property_list(List<PropertyInfo> *p_list) const {
-
 	String argt = "Any";
 	for (int i = 1; i < Variant::VARIANT_MAX; i++) {
 		argt += "," + Variant::get_type_name(Variant::Type(i));
@@ -151,55 +141,44 @@ void VisualScriptExpression::_get_property_list(List<PropertyInfo> *p_list) cons
 	p_list->push_back(PropertyInfo(Variant::BOOL, "sequenced"));
 
 	for (int i = 0; i < inputs.size(); i++) {
-
 		p_list->push_back(PropertyInfo(Variant::INT, "input_" + itos(i) + "/type", PROPERTY_HINT_ENUM, argt));
 		p_list->push_back(PropertyInfo(Variant::STRING, "input_" + itos(i) + "/name"));
 	}
 }
 
 int VisualScriptExpression::get_output_sequence_port_count() const {
-
 	return sequenced ? 1 : 0;
 }
 bool VisualScriptExpression::has_input_sequence_port() const {
-
 	return sequenced;
 }
 
 String VisualScriptExpression::get_output_sequence_port_text(int p_port) const {
-
 	return String();
 }
 
 int VisualScriptExpression::get_input_value_port_count() const {
-
 	return inputs.size();
 }
 int VisualScriptExpression::get_output_value_port_count() const {
-
 	return 1;
 }
 
 PropertyInfo VisualScriptExpression::get_input_value_port_info(int p_idx) const {
-
 	return PropertyInfo(inputs[p_idx].type, inputs[p_idx].name);
 }
 PropertyInfo VisualScriptExpression::get_output_value_port_info(int p_idx) const {
-
 	return PropertyInfo(output_type, "result");
 }
 
 String VisualScriptExpression::get_caption() const {
-
 	return "Expression";
 }
 String VisualScriptExpression::get_text() const {
-
 	return expression;
 }
 
 Error VisualScriptExpression::_get_token(Token &r_token) {
-
 	while (true) {
 #define GET_CHAR() (str_ofs >= expression.length() ? 0 : expression[str_ofs++])
 
@@ -210,58 +189,47 @@ Error VisualScriptExpression::_get_token(Token &r_token) {
 		}
 
 		switch (cchar) {
-
 			case 0: {
 				r_token.type = TK_EOF;
 				return OK;
 			} break;
 			case '{': {
-
 				r_token.type = TK_CURLY_BRACKET_OPEN;
 				return OK;
 			};
 			case '}': {
-
 				r_token.type = TK_CURLY_BRACKET_CLOSE;
 				return OK;
 			};
 			case '[': {
-
 				r_token.type = TK_BRACKET_OPEN;
 				return OK;
 			};
 			case ']': {
-
 				r_token.type = TK_BRACKET_CLOSE;
 				return OK;
 			};
 			case '(': {
-
 				r_token.type = TK_PARENTHESIS_OPEN;
 				return OK;
 			};
 			case ')': {
-
 				r_token.type = TK_PARENTHESIS_CLOSE;
 				return OK;
 			};
 			case ',': {
-
 				r_token.type = TK_COMMA;
 				return OK;
 			};
 			case ':': {
-
 				r_token.type = TK_COLON;
 				return OK;
 			};
 			case '.': {
-
 				r_token.type = TK_PERIOD;
 				return OK;
 			};
 			case '=': {
-
 				cchar = GET_CHAR();
 				if (cchar == '=') {
 					r_token.type = TK_OP_EQUAL;
@@ -273,7 +241,6 @@ Error VisualScriptExpression::_get_token(Token &r_token) {
 				return OK;
 			};
 			case '!': {
-
 				if (expression[str_ofs] == '=') {
 					r_token.type = TK_OP_NOT_EQUAL;
 					str_ofs++;
@@ -283,7 +250,6 @@ Error VisualScriptExpression::_get_token(Token &r_token) {
 				return OK;
 			};
 			case '>': {
-
 				if (expression[str_ofs] == '=') {
 					r_token.type = TK_OP_GREATER_EQUAL;
 					str_ofs++;
@@ -296,7 +262,6 @@ Error VisualScriptExpression::_get_token(Token &r_token) {
 				return OK;
 			};
 			case '<': {
-
 				if (expression[str_ofs] == '=') {
 					r_token.type = TK_OP_LESS_EQUAL;
 					str_ofs++;
@@ -329,7 +294,6 @@ Error VisualScriptExpression::_get_token(Token &r_token) {
 				return OK;
 			};
 			case '&': {
-
 				if (expression[str_ofs] == '&') {
 					r_token.type = TK_OP_AND;
 					str_ofs++;
@@ -339,7 +303,6 @@ Error VisualScriptExpression::_get_token(Token &r_token) {
 				return OK;
 			};
 			case '|': {
-
 				if (expression[str_ofs] == '|') {
 					r_token.type = TK_OP_OR;
 					str_ofs++;
@@ -349,22 +312,18 @@ Error VisualScriptExpression::_get_token(Token &r_token) {
 				return OK;
 			};
 			case '^': {
-
 				r_token.type = TK_OP_BIT_XOR;
 
 				return OK;
 			};
 			case '~': {
-
 				r_token.type = TK_OP_BIT_INVERT;
 
 				return OK;
 			};
 			case '"': {
-
 				String str;
 				while (true) {
-
 					CharType ch = GET_CHAR();
 
 					if (ch == 0) {
@@ -385,12 +344,21 @@ Error VisualScriptExpression::_get_token(Token &r_token) {
 						CharType res = 0;
 
 						switch (next) {
-
-							case 'b': res = 8; break;
-							case 't': res = 9; break;
-							case 'n': res = 10; break;
-							case 'f': res = 12; break;
-							case 'r': res = 13; break;
+							case 'b':
+								res = 8;
+								break;
+							case 't':
+								res = 9;
+								break;
+							case 'n':
+								res = 10;
+								break;
+							case 'f':
+								res = 12;
+								break;
+							case 'r':
+								res = 13;
+								break;
 							case 'u': {
 								//hexnumbarh - oct is deprecated
 
@@ -403,7 +371,6 @@ Error VisualScriptExpression::_get_token(Token &r_token) {
 										return ERR_PARSE_ERROR;
 									}
 									if (!((c >= '0' && c <= '9') || (c >= 'a' && c <= 'f') || (c >= 'A' && c <= 'F'))) {
-
 										_set_error("Malformed hex constant in string");
 										r_token.type = TK_ERROR;
 										return ERR_PARSE_ERROR;
@@ -450,12 +417,11 @@ Error VisualScriptExpression::_get_token(Token &r_token) {
 
 			} break;
 			default: {
-
 				if (cchar <= 32) {
 					break;
 				}
 
-				if (cchar == '-' || (cchar >= '0' && cchar <= '9')) {
+				if (cchar >= '0' && cchar <= '9') {
 					//a number
 
 					String num;
@@ -466,21 +432,14 @@ Error VisualScriptExpression::_get_token(Token &r_token) {
 #define READING_DONE 4
 					int reading = READING_INT;
 
-					if (cchar == '-') {
-						num += '-';
-						cchar = GET_CHAR();
-					}
-
 					CharType c = cchar;
 					bool exp_sign = false;
 					bool exp_beg = false;
 					bool is_float = false;
 
 					while (true) {
-
 						switch (reading) {
 							case READING_INT: {
-
 								if (c >= '0' && c <= '9') {
 									//pass
 								} else if (c == '.') {
@@ -494,9 +453,7 @@ Error VisualScriptExpression::_get_token(Token &r_token) {
 
 							} break;
 							case READING_DEC: {
-
 								if (c >= '0' && c <= '9') {
-
 								} else if (c == 'e') {
 									reading = READING_EXP;
 
@@ -506,13 +463,13 @@ Error VisualScriptExpression::_get_token(Token &r_token) {
 
 							} break;
 							case READING_EXP: {
-
 								if (c >= '0' && c <= '9') {
 									exp_beg = true;
 
 								} else if ((c == '-' || c == '+') && !exp_sign && !exp_beg) {
-									if (c == '-')
+									if (c == '-') {
 										is_float = true;
+									}
 									exp_sign = true;
 
 								} else {
@@ -521,8 +478,9 @@ Error VisualScriptExpression::_get_token(Token &r_token) {
 							} break;
 						}
 
-						if (reading == READING_DONE)
+						if (reading == READING_DONE) {
 							break;
+						}
 						num += String::chr(c);
 						c = GET_CHAR();
 					}
@@ -531,19 +489,18 @@ Error VisualScriptExpression::_get_token(Token &r_token) {
 
 					r_token.type = TK_CONSTANT;
 
-					if (is_float)
+					if (is_float) {
 						r_token.value = num.to_double();
-					else
+					} else {
 						r_token.value = num.to_int();
+					}
 					return OK;
 
 				} else if ((cchar >= 'A' && cchar <= 'Z') || (cchar >= 'a' && cchar <= 'z') || cchar == '_') {
-
 					String id;
 					bool first = true;
 
 					while ((cchar >= 'A' && cchar <= 'Z') || (cchar >= 'a' && cchar <= 'z') || cchar == '_' || (!first && cchar >= '0' && cchar <= '9')) {
-
 						id += String::chr(cchar);
 						cchar = GET_CHAR();
 						first = false;
@@ -583,7 +540,6 @@ Error VisualScriptExpression::_get_token(Token &r_token) {
 					} else if (id == "self") {
 						r_token.type = TK_SELF;
 					} else {
-
 						for (int i = 0; i < Variant::VARIANT_MAX; i++) {
 							if (id == Variant::get_type_name(Variant::Type(i))) {
 								r_token.type = TK_BASIC_TYPE;
@@ -658,17 +614,17 @@ const char *VisualScriptExpression::token_name[TK_MAX] = {
 };
 
 VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
-
 	Vector<Expression> expression;
 
 	while (true) {
 		//keep appending stuff to expression
-		ENode *expr = NULL;
+		ENode *expr = nullptr;
 
 		Token tk;
 		_get_token(tk);
-		if (error_set)
-			return NULL;
+		if (error_set) {
+			return nullptr;
+		}
 
 		switch (tk.type) {
 			case TK_CURLY_BRACKET_OPEN: {
@@ -676,7 +632,6 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 				DictionaryNode *dn = alloc_node<DictionaryNode>();
 
 				while (true) {
-
 					int cofs = str_ofs;
 					_get_token(tk);
 					if (tk.type == TK_CURLY_BRACKET_CLOSE) {
@@ -684,22 +639,24 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 					}
 					str_ofs = cofs; //revert
 					//parse an expression
-					ENode *expr = _parse_expression();
-					if (!expr)
-						return NULL;
-					dn->dict.push_back(expr);
+					ENode *expr2 = _parse_expression();
+					if (!expr2) {
+						return nullptr;
+					}
+					dn->dict.push_back(expr2);
 
 					_get_token(tk);
 					if (tk.type != TK_COLON) {
 						_set_error("Expected ':'");
-						return NULL;
+						return nullptr;
 					}
 
-					expr = _parse_expression();
-					if (!expr)
-						return NULL;
+					expr2 = _parse_expression();
+					if (!expr2) {
+						return nullptr;
+					}
 
-					dn->dict.push_back(expr);
+					dn->dict.push_back(expr2);
 
 					cofs = str_ofs;
 					_get_token(tk);
@@ -720,7 +677,6 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 				ArrayNode *an = alloc_node<ArrayNode>();
 
 				while (true) {
-
 					int cofs = str_ofs;
 					_get_token(tk);
 					if (tk.type == TK_BRACKET_CLOSE) {
@@ -728,10 +684,11 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 					}
 					str_ofs = cofs; //revert
 					//parse an expression
-					ENode *expr = _parse_expression();
-					if (!expr)
-						return NULL;
-					an->array.push_back(expr);
+					ENode *expr2 = _parse_expression();
+					if (!expr2) {
+						return nullptr;
+					}
+					an->array.push_back(expr2);
 
 					cofs = str_ofs;
 					_get_token(tk);
@@ -749,19 +706,19 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 			case TK_PARENTHESIS_OPEN: {
 				//a suexpression
 				ENode *e = _parse_expression();
-				if (error_set)
-					return NULL;
+				if (error_set) {
+					return nullptr;
+				}
 				_get_token(tk);
 				if (tk.type != TK_PARENTHESIS_CLOSE) {
 					_set_error("Expected ')'");
-					return NULL;
+					return nullptr;
 				}
 
 				expr = e;
 
 			} break;
 			case TK_IDENTIFIER: {
-
 				String what = tk.value;
 				int index = -1;
 				for (int i = 0; i < inputs.size(); i++) {
@@ -777,11 +734,10 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 					expr = input;
 				} else {
 					_set_error("Invalid input identifier '" + what + "'. For script variables, use self (locals are for inputs)." + what);
-					return NULL;
+					return nullptr;
 				}
 			} break;
 			case TK_SELF: {
-
 				SelfNode *self = alloc_node<SelfNode>();
 				expr = self;
 			} break;
@@ -797,14 +753,13 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 				_get_token(tk);
 				if (tk.type != TK_PARENTHESIS_OPEN) {
 					_set_error("Expected '('");
-					return NULL;
+					return nullptr;
 				}
 
 				ConstructorNode *constructor = alloc_node<ConstructorNode>();
 				constructor->data_type = bt;
 
 				while (true) {
-
 					int cofs = str_ofs;
 					_get_token(tk);
 					if (tk.type == TK_PARENTHESIS_CLOSE) {
@@ -812,11 +767,12 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 					}
 					str_ofs = cofs; //revert
 					//parse an expression
-					ENode *expr = _parse_expression();
-					if (!expr)
-						return NULL;
+					ENode *expr2 = _parse_expression();
+					if (!expr2) {
+						return nullptr;
+					}
 
-					constructor->arguments.push_back(expr);
+					constructor->arguments.push_back(expr2);
 
 					cofs = str_ofs;
 					_get_token(tk);
@@ -838,14 +794,13 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 				_get_token(tk);
 				if (tk.type != TK_PARENTHESIS_OPEN) {
 					_set_error("Expected '('");
-					return NULL;
+					return nullptr;
 				}
 
 				BuiltinFuncNode *bifunc = alloc_node<BuiltinFuncNode>();
 				bifunc->func = VisualScriptBuiltinFunc::BuiltinFunc(int(tk.value));
 
 				while (true) {
-
 					int cofs = str_ofs;
 					_get_token(tk);
 					if (tk.type == TK_PARENTHESIS_CLOSE) {
@@ -853,11 +808,12 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 					}
 					str_ofs = cofs; //revert
 					//parse an expression
-					ENode *expr = _parse_expression();
-					if (!expr)
-						return NULL;
+					ENode *expr2 = _parse_expression();
+					if (!expr2) {
+						return nullptr;
+					}
 
-					bifunc->arguments.push_back(expr);
+					bifunc->arguments.push_back(expr2);
 
 					cofs = str_ofs;
 					_get_token(tk);
@@ -879,7 +835,6 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 
 			} break;
 			case TK_OP_SUB: {
-
 				Expression e;
 				e.is_op = true;
 				e.op = Variant::OP_NEGATE;
@@ -887,7 +842,6 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 				continue;
 			} break;
 			case TK_OP_NOT: {
-
 				Expression e;
 				e.is_op = true;
 				e.op = Variant::OP_NOT;
@@ -897,7 +851,7 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 
 			default: {
 				_set_error("Expected expression.");
-				return NULL;
+				return nullptr;
 			} break;
 		}
 
@@ -906,8 +860,9 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 		while (true) {
 			int cofs2 = str_ofs;
 			_get_token(tk);
-			if (error_set)
-				return NULL;
+			if (error_set) {
+				return nullptr;
+			}
 
 			bool done = false;
 
@@ -919,15 +874,16 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 					index->base = expr;
 
 					ENode *what = _parse_expression();
-					if (!what)
-						return NULL;
+					if (!what) {
+						return nullptr;
+					}
 
 					index->index = what;
 
 					_get_token(tk);
 					if (tk.type != TK_BRACKET_CLOSE) {
 						_set_error("Expected ']' at end of index.");
-						return NULL;
+						return nullptr;
 					}
 					expr = index;
 
@@ -937,7 +893,7 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 					_get_token(tk);
 					if (tk.type != TK_IDENTIFIER) {
 						_set_error("Expected identifier after '.'");
-						return NULL;
+						return nullptr;
 					}
 
 					StringName identifier = tk.value;
@@ -951,26 +907,26 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 						func_call->base = expr;
 
 						while (true) {
-
-							int cofs = str_ofs;
+							int cofs3 = str_ofs;
 							_get_token(tk);
 							if (tk.type == TK_PARENTHESIS_CLOSE) {
 								break;
 							}
-							str_ofs = cofs; //revert
+							str_ofs = cofs3; //revert
 							//parse an expression
-							ENode *expr = _parse_expression();
-							if (!expr)
-								return NULL;
+							ENode *expr2 = _parse_expression();
+							if (!expr2) {
+								return nullptr;
+							}
 
-							func_call->arguments.push_back(expr);
+							func_call->arguments.push_back(expr2);
 
-							cofs = str_ofs;
+							cofs3 = str_ofs;
 							_get_token(tk);
 							if (tk.type == TK_COMMA) {
 								//all good
 							} else if (tk.type == TK_PARENTHESIS_CLOSE) {
-								str_ofs = cofs;
+								str_ofs = cofs3;
 							} else {
 								_set_error("Expected ',' or ')'");
 							}
@@ -994,8 +950,9 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 				} break;
 			}
 
-			if (done)
+			if (done) {
 				break;
+			}
 		}
 
 		//push expression
@@ -1010,34 +967,78 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 
 		int cofs = str_ofs;
 		_get_token(tk);
-		if (error_set)
-			return NULL;
+		if (error_set) {
+			return nullptr;
+		}
 
 		Variant::Operator op = Variant::OP_MAX;
 
 		switch (tk.type) {
-			case TK_OP_IN: op = Variant::OP_IN; break;
-			case TK_OP_EQUAL: op = Variant::OP_EQUAL; break;
-			case TK_OP_NOT_EQUAL: op = Variant::OP_NOT_EQUAL; break;
-			case TK_OP_LESS: op = Variant::OP_LESS; break;
-			case TK_OP_LESS_EQUAL: op = Variant::OP_LESS_EQUAL; break;
-			case TK_OP_GREATER: op = Variant::OP_GREATER; break;
-			case TK_OP_GREATER_EQUAL: op = Variant::OP_GREATER_EQUAL; break;
-			case TK_OP_AND: op = Variant::OP_AND; break;
-			case TK_OP_OR: op = Variant::OP_OR; break;
-			case TK_OP_NOT: op = Variant::OP_NOT; break;
-			case TK_OP_ADD: op = Variant::OP_ADD; break;
-			case TK_OP_SUB: op = Variant::OP_SUBTRACT; break;
-			case TK_OP_MUL: op = Variant::OP_MULTIPLY; break;
-			case TK_OP_DIV: op = Variant::OP_DIVIDE; break;
-			case TK_OP_MOD: op = Variant::OP_MODULE; break;
-			case TK_OP_SHIFT_LEFT: op = Variant::OP_SHIFT_LEFT; break;
-			case TK_OP_SHIFT_RIGHT: op = Variant::OP_SHIFT_RIGHT; break;
-			case TK_OP_BIT_AND: op = Variant::OP_BIT_AND; break;
-			case TK_OP_BIT_OR: op = Variant::OP_BIT_OR; break;
-			case TK_OP_BIT_XOR: op = Variant::OP_BIT_XOR; break;
-			case TK_OP_BIT_INVERT: op = Variant::OP_BIT_NEGATE; break;
-			default: {};
+			case TK_OP_IN:
+				op = Variant::OP_IN;
+				break;
+			case TK_OP_EQUAL:
+				op = Variant::OP_EQUAL;
+				break;
+			case TK_OP_NOT_EQUAL:
+				op = Variant::OP_NOT_EQUAL;
+				break;
+			case TK_OP_LESS:
+				op = Variant::OP_LESS;
+				break;
+			case TK_OP_LESS_EQUAL:
+				op = Variant::OP_LESS_EQUAL;
+				break;
+			case TK_OP_GREATER:
+				op = Variant::OP_GREATER;
+				break;
+			case TK_OP_GREATER_EQUAL:
+				op = Variant::OP_GREATER_EQUAL;
+				break;
+			case TK_OP_AND:
+				op = Variant::OP_AND;
+				break;
+			case TK_OP_OR:
+				op = Variant::OP_OR;
+				break;
+			case TK_OP_NOT:
+				op = Variant::OP_NOT;
+				break;
+			case TK_OP_ADD:
+				op = Variant::OP_ADD;
+				break;
+			case TK_OP_SUB:
+				op = Variant::OP_SUBTRACT;
+				break;
+			case TK_OP_MUL:
+				op = Variant::OP_MULTIPLY;
+				break;
+			case TK_OP_DIV:
+				op = Variant::OP_DIVIDE;
+				break;
+			case TK_OP_MOD:
+				op = Variant::OP_MODULE;
+				break;
+			case TK_OP_SHIFT_LEFT:
+				op = Variant::OP_SHIFT_LEFT;
+				break;
+			case TK_OP_SHIFT_RIGHT:
+				op = Variant::OP_SHIFT_RIGHT;
+				break;
+			case TK_OP_BIT_AND:
+				op = Variant::OP_BIT_AND;
+				break;
+			case TK_OP_BIT_OR:
+				op = Variant::OP_BIT_OR;
+				break;
+			case TK_OP_BIT_XOR:
+				op = Variant::OP_BIT_XOR;
+				break;
+			case TK_OP_BIT_INVERT:
+				op = Variant::OP_BIT_NEGATE;
+				break;
+			default: {
+			};
 		}
 
 		if (op == Variant::OP_MAX) { //stop appending stuff
@@ -1057,15 +1058,12 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 	/* Reduce the set set of expressions and place them in an operator tree, respecting precedence */
 
 	while (expression.size() > 1) {
-
 		int next_op = -1;
 		int min_priority = 0xFFFFF;
 		bool is_unary = false;
 
 		for (int i = 0; i < expression.size(); i++) {
-
 			if (!expression[i].is_op) {
-
 				continue;
 			}
 
@@ -1074,7 +1072,6 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 			bool unary = false;
 
 			switch (expression[i].op) {
-
 				case Variant::OP_BIT_NEGATE:
 					priority = 0;
 					unary = true;
@@ -1084,40 +1081,78 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 					unary = true;
 					break;
 
-				case Variant::OP_MULTIPLY: priority = 2; break;
-				case Variant::OP_DIVIDE: priority = 2; break;
-				case Variant::OP_MODULE: priority = 2; break;
+				case Variant::OP_MULTIPLY:
+					priority = 2;
+					break;
+				case Variant::OP_DIVIDE:
+					priority = 2;
+					break;
+				case Variant::OP_MODULE:
+					priority = 2;
+					break;
 
-				case Variant::OP_ADD: priority = 3; break;
-				case Variant::OP_SUBTRACT: priority = 3; break;
+				case Variant::OP_ADD:
+					priority = 3;
+					break;
+				case Variant::OP_SUBTRACT:
+					priority = 3;
+					break;
 
-				case Variant::OP_SHIFT_LEFT: priority = 4; break;
-				case Variant::OP_SHIFT_RIGHT: priority = 4; break;
+				case Variant::OP_SHIFT_LEFT:
+					priority = 4;
+					break;
+				case Variant::OP_SHIFT_RIGHT:
+					priority = 4;
+					break;
 
-				case Variant::OP_BIT_AND: priority = 5; break;
-				case Variant::OP_BIT_XOR: priority = 6; break;
-				case Variant::OP_BIT_OR: priority = 7; break;
+				case Variant::OP_BIT_AND:
+					priority = 5;
+					break;
+				case Variant::OP_BIT_XOR:
+					priority = 6;
+					break;
+				case Variant::OP_BIT_OR:
+					priority = 7;
+					break;
 
-				case Variant::OP_LESS: priority = 8; break;
-				case Variant::OP_LESS_EQUAL: priority = 8; break;
-				case Variant::OP_GREATER: priority = 8; break;
-				case Variant::OP_GREATER_EQUAL: priority = 8; break;
+				case Variant::OP_LESS:
+					priority = 8;
+					break;
+				case Variant::OP_LESS_EQUAL:
+					priority = 8;
+					break;
+				case Variant::OP_GREATER:
+					priority = 8;
+					break;
+				case Variant::OP_GREATER_EQUAL:
+					priority = 8;
+					break;
 
-				case Variant::OP_EQUAL: priority = 8; break;
-				case Variant::OP_NOT_EQUAL: priority = 8; break;
+				case Variant::OP_EQUAL:
+					priority = 8;
+					break;
+				case Variant::OP_NOT_EQUAL:
+					priority = 8;
+					break;
 
-				case Variant::OP_IN: priority = 10; break;
+				case Variant::OP_IN:
+					priority = 10;
+					break;
 
 				case Variant::OP_NOT:
 					priority = 11;
 					unary = true;
 					break;
-				case Variant::OP_AND: priority = 12; break;
-				case Variant::OP_OR: priority = 13; break;
+				case Variant::OP_AND:
+					priority = 12;
+					break;
+				case Variant::OP_OR:
+					priority = 13;
+					break;
 
 				default: {
 					_set_error("Parser bug, invalid operator in expression: " + itos(expression[i].op));
-					return NULL;
+					return nullptr;
 				}
 			}
 
@@ -1132,51 +1167,45 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 		}
 
 		if (next_op == -1) {
-
 			_set_error("Yet another parser bug....");
-			ERR_FAIL_COND_V(next_op == -1, NULL);
+			ERR_FAIL_V(nullptr);
 		}
 
 		// OK! create operator..
 		if (is_unary) {
-
 			int expr_pos = next_op;
 			while (expression[expr_pos].is_op) {
-
 				expr_pos++;
 				if (expr_pos == expression.size()) {
 					//can happen..
 					_set_error("Unexpected end of expression...");
-					return NULL;
+					return nullptr;
 				}
 			}
 
-			//consecutively do unary opeators
+			//consecutively do unary operators
 			for (int i = expr_pos - 1; i >= next_op; i--) {
-
 				OperatorNode *op = alloc_node<OperatorNode>();
 				op->op = expression[i].op;
 				op->nodes[0] = expression[i + 1].node;
-				op->nodes[1] = NULL;
-				expression[i].is_op = false;
-				expression[i].node = op;
+				op->nodes[1] = nullptr;
+				expression.write[i].is_op = false;
+				expression.write[i].node = op;
 				expression.remove(i + 1);
 			}
 
 		} else {
-
 			if (next_op < 1 || next_op >= (expression.size() - 1)) {
 				_set_error("Parser bug...");
-				ERR_FAIL_V(NULL);
+				ERR_FAIL_V(nullptr);
 			}
 
 			OperatorNode *op = alloc_node<OperatorNode>();
 			op->op = expression[next_op].op;
 
 			if (expression[next_op - 1].is_op) {
-
 				_set_error("Parser bug...");
-				ERR_FAIL_V(NULL);
+				ERR_FAIL_V(nullptr);
 			}
 
 			if (expression[next_op + 1].is_op) {
@@ -1186,14 +1215,14 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 				// due to how precedence works, unaries will always disappear first
 
 				_set_error("Unexpected two consecutive operators.");
-				return NULL;
+				return nullptr;
 			}
 
 			op->nodes[0] = expression[next_op - 1].node; //expression goes as left
 			op->nodes[1] = expression[next_op + 1].node; //next expression goes as right
 
 			//replace all 3 nodes by this operator and make it an expression
-			expression[next_op - 1].node = op;
+			expression.write[next_op - 1].node = op;
 			expression.remove(next_op);
 			expression.remove(next_op);
 		}
@@ -1203,14 +1232,14 @@ VisualScriptExpression::ENode *VisualScriptExpression::_parse_expression() {
 }
 
 bool VisualScriptExpression::_compile_expression() {
-
-	if (!expression_dirty)
+	if (!expression_dirty) {
 		return error_set;
+	}
 
 	if (nodes) {
 		memdelete(nodes);
-		nodes = NULL;
-		root = NULL;
+		nodes = nullptr;
+		root = nullptr;
 	}
 
 	error_str = String();
@@ -1220,11 +1249,11 @@ bool VisualScriptExpression::_compile_expression() {
 	root = _parse_expression();
 
 	if (error_set) {
-		root = NULL;
+		root = nullptr;
 		if (nodes) {
 			memdelete(nodes);
 		}
-		nodes = NULL;
+		nodes = nullptr;
 		return true;
 	}
 
@@ -1240,38 +1269,35 @@ public:
 	//virtual int get_working_memory_size() const { return 0; }
 	//execute by parsing the tree directly
 	virtual bool _execute(const Variant **p_inputs, VisualScriptExpression::ENode *p_node, Variant &r_ret, String &r_error_str, Variant::CallError &ce) {
-
 		switch (p_node->type) {
 			case VisualScriptExpression::ENode::TYPE_INPUT: {
-
 				const VisualScriptExpression::InputNode *in = static_cast<const VisualScriptExpression::InputNode *>(p_node);
 				r_ret = *p_inputs[in->index];
 			} break;
 			case VisualScriptExpression::ENode::TYPE_CONSTANT: {
-
 				const VisualScriptExpression::ConstantNode *c = static_cast<const VisualScriptExpression::ConstantNode *>(p_node);
 				r_ret = c->value;
 
 			} break;
 			case VisualScriptExpression::ENode::TYPE_SELF: {
-
 				r_ret = instance->get_owner_ptr();
 			} break;
 			case VisualScriptExpression::ENode::TYPE_OPERATOR: {
-
 				const VisualScriptExpression::OperatorNode *op = static_cast<const VisualScriptExpression::OperatorNode *>(p_node);
 
 				Variant a;
 				bool ret = _execute(p_inputs, op->nodes[0], a, r_error_str, ce);
-				if (ret)
+				if (ret) {
 					return true;
+				}
 
 				Variant b;
 
 				if (op->nodes[1]) {
 					ret = _execute(p_inputs, op->nodes[1], b, r_error_str, ce);
-					if (ret)
+					if (ret) {
 						return true;
+					}
 				}
 
 				bool valid = true;
@@ -1283,19 +1309,20 @@ public:
 
 			} break;
 			case VisualScriptExpression::ENode::TYPE_INDEX: {
-
 				const VisualScriptExpression::IndexNode *index = static_cast<const VisualScriptExpression::IndexNode *>(p_node);
 
 				Variant base;
 				bool ret = _execute(p_inputs, index->base, base, r_error_str, ce);
-				if (ret)
+				if (ret) {
 					return true;
+				}
 
 				Variant idx;
 
 				ret = _execute(p_inputs, index->index, idx, r_error_str, ce);
-				if (ret)
+				if (ret) {
 					return true;
+				}
 
 				bool valid;
 				r_ret = base.get(idx, &valid);
@@ -1306,13 +1333,13 @@ public:
 
 			} break;
 			case VisualScriptExpression::ENode::TYPE_NAMED_INDEX: {
-
 				const VisualScriptExpression::NamedIndexNode *index = static_cast<const VisualScriptExpression::NamedIndexNode *>(p_node);
 
 				Variant base;
 				bool ret = _execute(p_inputs, index->base, base, r_error_str, ce);
-				if (ret)
+				if (ret) {
 					return true;
+				}
 
 				bool valid;
 				r_ret = base.get_named(index->name, &valid);
@@ -1328,11 +1355,11 @@ public:
 				Array arr;
 				arr.resize(array->array.size());
 				for (int i = 0; i < array->array.size(); i++) {
-
 					Variant value;
 					bool ret = _execute(p_inputs, array->array[i], value, r_error_str, ce);
-					if (ret)
+					if (ret) {
 						return true;
+					}
 					arr[i] = value;
 				}
 
@@ -1344,16 +1371,17 @@ public:
 
 				Dictionary d;
 				for (int i = 0; i < dictionary->dict.size(); i += 2) {
-
 					Variant key;
 					bool ret = _execute(p_inputs, dictionary->dict[i + 0], key, r_error_str, ce);
-					if (ret)
+					if (ret) {
 						return true;
+					}
 
 					Variant value;
 					ret = _execute(p_inputs, dictionary->dict[i + 1], value, r_error_str, ce);
-					if (ret)
+					if (ret) {
 						return true;
+					}
 
 					d[key] = value;
 				}
@@ -1361,7 +1389,6 @@ public:
 				r_ret = d;
 			} break;
 			case VisualScriptExpression::ENode::TYPE_CONSTRUCTOR: {
-
 				const VisualScriptExpression::ConstructorNode *constructor = static_cast<const VisualScriptExpression::ConstructorNode *>(p_node);
 
 				Vector<Variant> arr;
@@ -1370,13 +1397,13 @@ public:
 				argp.resize(constructor->arguments.size());
 
 				for (int i = 0; i < constructor->arguments.size(); i++) {
-
 					Variant value;
 					bool ret = _execute(p_inputs, constructor->arguments[i], value, r_error_str, ce);
-					if (ret)
+					if (ret) {
 						return true;
-					arr[i] = value;
-					argp[i] = &arr[i];
+					}
+					arr.write[i] = value;
+					argp.write[i] = &arr[i];
 				}
 
 				r_ret = Variant::construct(constructor->data_type, (const Variant **)argp.ptr(), argp.size(), ce);
@@ -1388,7 +1415,6 @@ public:
 
 			} break;
 			case VisualScriptExpression::ENode::TYPE_BUILTIN_FUNC: {
-
 				const VisualScriptExpression::BuiltinFuncNode *bifunc = static_cast<const VisualScriptExpression::BuiltinFuncNode *>(p_node);
 
 				Vector<Variant> arr;
@@ -1397,13 +1423,13 @@ public:
 				argp.resize(bifunc->arguments.size());
 
 				for (int i = 0; i < bifunc->arguments.size(); i++) {
-
 					Variant value;
 					bool ret = _execute(p_inputs, bifunc->arguments[i], value, r_error_str, ce);
-					if (ret)
+					if (ret) {
 						return true;
-					arr[i] = value;
-					argp[i] = &arr[i];
+					}
+					arr.write[i] = value;
+					argp.write[i] = &arr[i];
 				}
 
 				VisualScriptBuiltinFunc::exec_func(bifunc->func, (const Variant **)argp.ptr(), &r_ret, ce, r_error_str);
@@ -1415,13 +1441,13 @@ public:
 
 			} break;
 			case VisualScriptExpression::ENode::TYPE_CALL: {
-
 				const VisualScriptExpression::CallNode *call = static_cast<const VisualScriptExpression::CallNode *>(p_node);
 
 				Variant base;
 				bool ret = _execute(p_inputs, call->base, base, r_error_str, ce);
-				if (ret)
+				if (ret) {
 					return true;
+				}
 
 				Vector<Variant> arr;
 				Vector<const Variant *> argp;
@@ -1429,13 +1455,13 @@ public:
 				argp.resize(call->arguments.size());
 
 				for (int i = 0; i < call->arguments.size(); i++) {
-
 					Variant value;
-					bool ret = _execute(p_inputs, call->arguments[i], value, r_error_str, ce);
-					if (ret)
+					bool ret2 = _execute(p_inputs, call->arguments[i], value, r_error_str, ce);
+					if (ret2) {
 						return true;
-					arr[i] = value;
-					argp[i] = &arr[i];
+					}
+					arr.write[i] = value;
+					argp.write[i] = &arr[i];
 				}
 
 				r_ret = base.call(call->method, (const Variant **)argp.ptr(), argp.size(), ce);
@@ -1451,7 +1477,6 @@ public:
 	}
 
 	virtual int step(const Variant **p_inputs, Variant **p_outputs, StartMode p_start_mode, Variant *p_working_mem, Variant::CallError &r_error, String &r_error_str) {
-
 		if (!expression->root || expression->error_set) {
 			r_error_str = expression->error_str;
 			r_error.error = Variant::CallError::CALL_ERROR_INVALID_METHOD;
@@ -1465,7 +1490,6 @@ public:
 
 #ifdef DEBUG_ENABLED
 		if (!error && expression->output_type != Variant::NIL && !Variant::can_convert_strict(p_outputs[0]->get_type(), expression->output_type)) {
-
 			r_error_str += "Can't convert expression result from " + Variant::get_type_name(p_outputs[0]->get_type()) + " to " + Variant::get_type_name(expression->output_type) + ".";
 			r_error.error = Variant::CallError::CALL_ERROR_INVALID_METHOD;
 		}
@@ -1476,7 +1500,6 @@ public:
 };
 
 VisualScriptNodeInstance *VisualScriptExpression::instance(VisualScriptInstance *p_instance) {
-
 	_compile_expression();
 	VisualScriptNodeInstanceExpression *instance = memnew(VisualScriptNodeInstanceExpression);
 	instance->instance = p_instance;
@@ -1488,19 +1511,17 @@ VisualScriptExpression::VisualScriptExpression() {
 	output_type = Variant::NIL;
 	expression_dirty = true;
 	error_set = true;
-	root = NULL;
-	nodes = NULL;
+	root = nullptr;
+	nodes = nullptr;
 	sequenced = false;
 }
 
 VisualScriptExpression::~VisualScriptExpression() {
-
 	if (nodes) {
 		memdelete(nodes);
 	}
 }
 
 void register_visual_script_expression_node() {
-
 	VisualScriptLanguage::singleton->add_register_func("operators/expression", create_node_generic<VisualScriptExpression>);
 }
