@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -34,19 +34,20 @@
 #include "visual_script.h"
 
 class VisualScriptFunction : public VisualScriptNode {
-
-	GDCLASS(VisualScriptFunction, VisualScriptNode)
+	GDCLASS(VisualScriptFunction, VisualScriptNode);
 
 	struct Argument {
 		String name;
 		Variant::Type type;
+		PropertyHint hint;
+		String hint_string;
 	};
 
 	Vector<Argument> arguments;
 
 	bool stack_less;
 	int stack_size;
-	ScriptInstance::RPCMode rpc_mode;
+	MultiplayerAPI::RPCMode rpc_mode;
 	bool sequenced;
 
 protected:
@@ -70,7 +71,7 @@ public:
 	virtual String get_text() const;
 	virtual String get_category() const { return "flow_control"; }
 
-	void add_argument(Variant::Type p_type, const String &p_name, int p_index = -1);
+	void add_argument(Variant::Type p_type, const String &p_name, int p_index = -1, const PropertyHint p_hint = PROPERTY_HINT_NONE, const String &p_hint_string = String(""));
 	void set_argument_type(int p_argidx, Variant::Type p_type);
 	Variant::Type get_argument_type(int p_argidx) const;
 	void set_argument_name(int p_argidx, const String &p_name);
@@ -93,17 +94,111 @@ public:
 	void set_return_type(Variant::Type p_type);
 	Variant::Type get_return_type() const;
 
-	void set_rpc_mode(ScriptInstance::RPCMode p_mode);
-	ScriptInstance::RPCMode get_rpc_mode() const;
+	void set_rpc_mode(MultiplayerAPI::RPCMode p_mode);
+	MultiplayerAPI::RPCMode get_rpc_mode() const;
 
 	virtual VisualScriptNodeInstance *instance(VisualScriptInstance *p_instance);
 
 	VisualScriptFunction();
 };
 
-class VisualScriptOperator : public VisualScriptNode {
+class VisualScriptLists : public VisualScriptNode {
+	GDCLASS(VisualScriptLists, VisualScriptNode)
 
-	GDCLASS(VisualScriptOperator, VisualScriptNode)
+	struct Port {
+		String name;
+		Variant::Type type;
+	};
+
+protected:
+	Vector<Port> inputports;
+	Vector<Port> outputports;
+
+	enum {
+		OUTPUT_EDITABLE = 0x0001,
+		OUTPUT_NAME_EDITABLE = 0x0002,
+		OUTPUT_TYPE_EDITABLE = 0x0004,
+		INPUT_EDITABLE = 0x0008,
+		INPUT_NAME_EDITABLE = 0x000F,
+		INPUT_TYPE_EDITABLE = 0x0010,
+	};
+
+	int flags;
+
+	bool sequenced;
+
+	bool _set(const StringName &p_name, const Variant &p_value);
+	bool _get(const StringName &p_name, Variant &r_ret) const;
+	void _get_property_list(List<PropertyInfo> *p_list) const;
+
+	static void _bind_methods();
+
+public:
+	virtual bool is_output_port_editable() const;
+	virtual bool is_output_port_name_editable() const;
+	virtual bool is_output_port_type_editable() const;
+
+	virtual bool is_input_port_editable() const;
+	virtual bool is_input_port_name_editable() const;
+	virtual bool is_input_port_type_editable() const;
+
+	virtual int get_output_sequence_port_count() const;
+	virtual bool has_input_sequence_port() const;
+
+	virtual String get_output_sequence_port_text(int p_port) const;
+
+	virtual int get_input_value_port_count() const;
+	virtual int get_output_value_port_count() const;
+
+	virtual PropertyInfo get_input_value_port_info(int p_idx) const;
+	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
+
+	virtual String get_caption() const = 0;
+	virtual String get_text() const = 0;
+	virtual String get_category() const = 0;
+
+	void add_input_data_port(Variant::Type p_type, const String &p_name, int p_index = -1);
+	void set_input_data_port_type(int p_idx, Variant::Type p_type);
+	void set_input_data_port_name(int p_idx, const String &p_name);
+	void remove_input_data_port(int p_argidx);
+
+	void add_output_data_port(Variant::Type p_type, const String &p_name, int p_index = -1);
+	void set_output_data_port_type(int p_idx, Variant::Type p_type);
+	void set_output_data_port_name(int p_idx, const String &p_name);
+	void remove_output_data_port(int p_argidx);
+
+	void set_sequenced(bool p_enable);
+	bool is_sequenced() const;
+
+	VisualScriptLists();
+};
+
+class VisualScriptComposeArray : public VisualScriptLists {
+	GDCLASS(VisualScriptComposeArray, VisualScriptLists)
+
+public:
+	virtual int get_output_sequence_port_count() const;
+	virtual bool has_input_sequence_port() const;
+
+	virtual String get_output_sequence_port_text(int p_port) const;
+
+	virtual int get_input_value_port_count() const;
+	virtual int get_output_value_port_count() const;
+
+	virtual PropertyInfo get_input_value_port_info(int p_idx) const;
+	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
+
+	virtual String get_caption() const;
+	virtual String get_text() const;
+	virtual String get_category() const { return "functions"; }
+
+	virtual VisualScriptNodeInstance *instance(VisualScriptInstance *p_instance);
+
+	VisualScriptComposeArray();
+};
+
+class VisualScriptOperator : public VisualScriptNode {
+	GDCLASS(VisualScriptOperator, VisualScriptNode);
 
 	Variant::Type typed;
 	Variant::Operator op;
@@ -124,7 +219,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const { return "operators"; }
 
 	void set_operator(Variant::Operator p_op);
@@ -139,8 +233,7 @@ public:
 };
 
 class VisualScriptSelect : public VisualScriptNode {
-
-	GDCLASS(VisualScriptSelect, VisualScriptNode)
+	GDCLASS(VisualScriptSelect, VisualScriptNode);
 
 	Variant::Type typed;
 
@@ -172,8 +265,7 @@ public:
 };
 
 class VisualScriptVariableGet : public VisualScriptNode {
-
-	GDCLASS(VisualScriptVariableGet, VisualScriptNode)
+	GDCLASS(VisualScriptVariableGet, VisualScriptNode);
 
 	StringName variable;
 
@@ -194,7 +286,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const { return "data"; }
 
 	void set_variable(StringName p_variable);
@@ -206,8 +297,7 @@ public:
 };
 
 class VisualScriptVariableSet : public VisualScriptNode {
-
-	GDCLASS(VisualScriptVariableSet, VisualScriptNode)
+	GDCLASS(VisualScriptVariableSet, VisualScriptNode);
 
 	StringName variable;
 
@@ -228,7 +318,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const { return "data"; }
 
 	void set_variable(StringName p_variable);
@@ -240,8 +329,7 @@ public:
 };
 
 class VisualScriptConstant : public VisualScriptNode {
-
-	GDCLASS(VisualScriptConstant, VisualScriptNode)
+	GDCLASS(VisualScriptConstant, VisualScriptNode);
 
 	Variant::Type type;
 	Variant value;
@@ -263,7 +351,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const { return "constants"; }
 
 	void set_constant_type(Variant::Type p_type);
@@ -278,8 +365,7 @@ public:
 };
 
 class VisualScriptPreload : public VisualScriptNode {
-
-	GDCLASS(VisualScriptPreload, VisualScriptNode)
+	GDCLASS(VisualScriptPreload, VisualScriptNode);
 
 	Ref<Resource> preload;
 
@@ -299,7 +385,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const { return "data"; }
 
 	void set_preload(const Ref<Resource> &p_preload);
@@ -311,8 +396,7 @@ public:
 };
 
 class VisualScriptIndexGet : public VisualScriptNode {
-
-	GDCLASS(VisualScriptIndexGet, VisualScriptNode)
+	GDCLASS(VisualScriptIndexGet, VisualScriptNode);
 
 public:
 	virtual int get_output_sequence_port_count() const;
@@ -327,7 +411,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const { return "operators"; }
 
 	virtual VisualScriptNodeInstance *instance(VisualScriptInstance *p_instance);
@@ -336,8 +419,7 @@ public:
 };
 
 class VisualScriptIndexSet : public VisualScriptNode {
-
-	GDCLASS(VisualScriptIndexSet, VisualScriptNode)
+	GDCLASS(VisualScriptIndexSet, VisualScriptNode);
 
 public:
 	virtual int get_output_sequence_port_count() const;
@@ -352,7 +434,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const { return "operators"; }
 
 	virtual VisualScriptNodeInstance *instance(VisualScriptInstance *p_instance);
@@ -361,8 +442,7 @@ public:
 };
 
 class VisualScriptGlobalConstant : public VisualScriptNode {
-
-	GDCLASS(VisualScriptGlobalConstant, VisualScriptNode)
+	GDCLASS(VisualScriptGlobalConstant, VisualScriptNode);
 
 	int index;
 
@@ -381,7 +461,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const { return "constants"; }
 
 	void set_global_constant(int p_which);
@@ -393,8 +472,7 @@ public:
 };
 
 class VisualScriptClassConstant : public VisualScriptNode {
-
-	GDCLASS(VisualScriptClassConstant, VisualScriptNode)
+	GDCLASS(VisualScriptClassConstant, VisualScriptNode);
 
 	StringName base_type;
 	StringName name;
@@ -416,7 +494,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const { return "constants"; }
 
 	void set_class_constant(const StringName &p_which);
@@ -431,8 +508,7 @@ public:
 };
 
 class VisualScriptBasicTypeConstant : public VisualScriptNode {
-
-	GDCLASS(VisualScriptBasicTypeConstant, VisualScriptNode)
+	GDCLASS(VisualScriptBasicTypeConstant, VisualScriptNode);
 
 	Variant::Type type;
 	StringName name;
@@ -469,8 +545,8 @@ public:
 };
 
 class VisualScriptMathConstant : public VisualScriptNode {
+	GDCLASS(VisualScriptMathConstant, VisualScriptNode);
 
-	GDCLASS(VisualScriptMathConstant, VisualScriptNode)
 public:
 	enum MathConstant {
 		MATH_CONSTANT_ONE,
@@ -505,7 +581,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const { return "constants"; }
 
 	void set_math_constant(MathConstant p_which);
@@ -519,11 +594,13 @@ public:
 VARIANT_ENUM_CAST(VisualScriptMathConstant::MathConstant)
 
 class VisualScriptEngineSingleton : public VisualScriptNode {
-
-	GDCLASS(VisualScriptEngineSingleton, VisualScriptNode)
+	GDCLASS(VisualScriptEngineSingleton, VisualScriptNode);
 
 	String singleton;
 
+protected:
+	void _validate_property(PropertyInfo &property) const;
+
 	static void _bind_methods();
 
 public:
@@ -539,7 +616,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const { return "data"; }
 
 	void set_singleton(const String &p_string);
@@ -553,8 +629,7 @@ public:
 };
 
 class VisualScriptSceneNode : public VisualScriptNode {
-
-	GDCLASS(VisualScriptSceneNode, VisualScriptNode)
+	GDCLASS(VisualScriptSceneNode, VisualScriptNode);
 
 	NodePath path;
 
@@ -575,7 +650,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const { return "data"; }
 
 	void set_node_path(const NodePath &p_path);
@@ -589,8 +663,7 @@ public:
 };
 
 class VisualScriptSceneTree : public VisualScriptNode {
-
-	GDCLASS(VisualScriptSceneTree, VisualScriptNode)
+	GDCLASS(VisualScriptSceneTree, VisualScriptNode);
 
 protected:
 	virtual void _validate_property(PropertyInfo &property) const;
@@ -609,7 +682,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const { return "data"; }
 
 	virtual VisualScriptNodeInstance *instance(VisualScriptInstance *p_instance);
@@ -620,8 +692,7 @@ public:
 };
 
 class VisualScriptResourcePath : public VisualScriptNode {
-
-	GDCLASS(VisualScriptResourcePath, VisualScriptNode)
+	GDCLASS(VisualScriptResourcePath, VisualScriptNode);
 
 	String path;
 
@@ -641,7 +712,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const { return "data"; }
 
 	void set_resource_path(const String &p_path);
@@ -653,8 +723,7 @@ public:
 };
 
 class VisualScriptSelf : public VisualScriptNode {
-
-	GDCLASS(VisualScriptSelf, VisualScriptNode)
+	GDCLASS(VisualScriptSelf, VisualScriptNode);
 
 protected:
 	static void _bind_methods();
@@ -672,7 +741,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const { return "data"; }
 
 	virtual VisualScriptNodeInstance *instance(VisualScriptInstance *p_instance);
@@ -683,8 +751,7 @@ public:
 };
 
 class VisualScriptCustomNode : public VisualScriptNode {
-
-	GDCLASS(VisualScriptCustomNode, VisualScriptNode)
+	GDCLASS(VisualScriptCustomNode, VisualScriptNode);
 
 protected:
 	static void _bind_methods();
@@ -723,6 +790,8 @@ public:
 
 	virtual VisualScriptNodeInstance *instance(VisualScriptInstance *p_instance);
 
+	virtual TypeGuess guess_output_type(TypeGuess *p_inputs, int p_output) const;
+
 	void _script_changed();
 
 	VisualScriptCustomNode();
@@ -731,8 +800,7 @@ public:
 VARIANT_ENUM_CAST(VisualScriptCustomNode::StartMode);
 
 class VisualScriptSubCall : public VisualScriptNode {
-
-	GDCLASS(VisualScriptSubCall, VisualScriptNode)
+	GDCLASS(VisualScriptSubCall, VisualScriptNode);
 
 protected:
 	static void _bind_methods();
@@ -759,8 +827,7 @@ public:
 };
 
 class VisualScriptComment : public VisualScriptNode {
-
-	GDCLASS(VisualScriptComment, VisualScriptNode)
+	GDCLASS(VisualScriptComment, VisualScriptNode);
 
 	String title;
 	String description;
@@ -800,8 +867,7 @@ public:
 };
 
 class VisualScriptConstructor : public VisualScriptNode {
-
-	GDCLASS(VisualScriptConstructor, VisualScriptNode)
+	GDCLASS(VisualScriptConstructor, VisualScriptNode);
 
 	Variant::Type type;
 	MethodInfo constructor;
@@ -822,7 +888,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const;
 
 	void set_constructor_type(Variant::Type p_type);
@@ -837,8 +902,7 @@ public:
 };
 
 class VisualScriptLocalVar : public VisualScriptNode {
-
-	GDCLASS(VisualScriptLocalVar, VisualScriptNode)
+	GDCLASS(VisualScriptLocalVar, VisualScriptNode);
 
 	StringName name;
 	Variant::Type type;
@@ -859,7 +923,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const;
 
 	void set_var_name(const StringName &p_name);
@@ -874,8 +937,7 @@ public:
 };
 
 class VisualScriptLocalVarSet : public VisualScriptNode {
-
-	GDCLASS(VisualScriptLocalVarSet, VisualScriptNode)
+	GDCLASS(VisualScriptLocalVarSet, VisualScriptNode);
 
 	StringName name;
 	Variant::Type type;
@@ -911,8 +973,8 @@ public:
 };
 
 class VisualScriptInputAction : public VisualScriptNode {
+	GDCLASS(VisualScriptInputAction, VisualScriptNode);
 
-	GDCLASS(VisualScriptInputAction, VisualScriptNode)
 public:
 	enum Mode {
 		MODE_PRESSED,
@@ -942,7 +1004,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const;
 
 	void set_action_name(const StringName &p_name);
@@ -959,8 +1020,7 @@ public:
 VARIANT_ENUM_CAST(VisualScriptInputAction::Mode)
 
 class VisualScriptDeconstruct : public VisualScriptNode {
-
-	GDCLASS(VisualScriptDeconstruct, VisualScriptNode)
+	GDCLASS(VisualScriptDeconstruct, VisualScriptNode);
 
 	struct Element {
 		StringName name;
@@ -993,7 +1053,6 @@ public:
 	virtual PropertyInfo get_output_value_port_info(int p_idx) const;
 
 	virtual String get_caption() const;
-	virtual String get_text() const;
 	virtual String get_category() const;
 
 	void set_deconstruct_type(Variant::Type p_type);
