@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -28,20 +28,20 @@
 /* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                */
 /*************************************************************************/
 
-#include "java_class_wrapper.h"
+#include "api/java_class_wrapper.h"
+#include "string_android.h"
 #include "thread_jandroid.h"
 
 bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method, const Variant **p_args, int p_argcount, Variant::CallError &r_error, Variant &ret) {
-
-	Map<StringName, List<MethodInfo> >::Element *M = methods.find(p_method);
+	Map<StringName, List<MethodInfo>>::Element *M = methods.find(p_method);
 	if (!M)
 		return false;
 
-	JNIEnv *env = ThreadAndroid::get_env();
+	JNIEnv *env = get_jni_env();
+	ERR_FAIL_COND_V(env == nullptr, false);
 
 	MethodInfo *method = NULL;
 	for (List<MethodInfo>::Element *E = M->get().front(); E; E = E->next()) {
-
 		if (!p_instance && !E->get()._static) {
 			r_error.error = Variant::CallError::CALL_ERROR_INSTANCE_IS_NULL;
 			continue;
@@ -49,13 +49,11 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 
 		int pc = E->get().param_types.size();
 		if (pc > p_argcount) {
-
 			r_error.error = Variant::CallError::CALL_ERROR_TOO_FEW_ARGUMENTS;
 			r_error.argument = pc;
 			continue;
 		}
 		if (pc < p_argcount) {
-
 			r_error.error = Variant::CallError::CALL_ERROR_TOO_MANY_ARGUMENTS;
 			r_error.argument = pc;
 			continue;
@@ -64,10 +62,8 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 		bool valid = true;
 
 		for (int i = 0; i < pc; i++) {
-
 			Variant::Type arg_expected = Variant::NIL;
 			switch (ptypes[i]) {
-
 				case ARG_TYPE_VOID: {
 					//bug?
 				} break;
@@ -85,7 +81,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 				case ARG_TYPE_SHORT:
 				case ARG_TYPE_INT:
 				case ARG_TYPE_LONG: {
-
 					if (!p_args[i]->is_num())
 						arg_expected = Variant::INT;
 
@@ -94,32 +89,26 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 				case ARG_NUMBER_CLASS_BIT | ARG_TYPE_DOUBLE:
 				case ARG_TYPE_FLOAT:
 				case ARG_TYPE_DOUBLE: {
-
 					if (!p_args[i]->is_num())
 						arg_expected = Variant::REAL;
 
 				} break;
 				case ARG_TYPE_STRING: {
-
 					if (p_args[i]->get_type() != Variant::STRING)
 						arg_expected = Variant::STRING;
 
 				} break;
 				case ARG_TYPE_CLASS: {
-
 					if (p_args[i]->get_type() != Variant::OBJECT)
 						arg_expected = Variant::OBJECT;
 					else {
-
 						Ref<Reference> ref = *p_args[i];
 						if (!ref.is_null()) {
 							if (Object::cast_to<JavaObject>(ref.ptr())) {
-
 								Ref<JavaObject> jo = ref;
 								//could be faster
 								jclass c = env->FindClass(E->get().param_sigs[i].operator String().utf8().get_data());
 								if (!c || !env->IsInstanceOf(jo->instance, c)) {
-
 									arg_expected = Variant::OBJECT;
 								} else {
 									//ok
@@ -132,7 +121,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 
 				} break;
 				default: {
-
 					if (p_args[i]->get_type() != Variant::ARRAY)
 						arg_expected = Variant::ARRAY;
 
@@ -162,13 +150,11 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 	jvalue *argv = NULL;
 
 	if (method->param_types.size()) {
-
 		argv = (jvalue *)alloca(sizeof(jvalue) * method->param_types.size());
 	}
 
 	List<jobject> to_free;
 	for (int i = 0; i < method->param_types.size(); i++) {
-
 		switch (method->param_types[i]) {
 			case ARG_TYPE_VOID: {
 				//can't happen
@@ -278,10 +264,8 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 				to_free.push_back(jStr);
 			} break;
 			case ARG_TYPE_CLASS: {
-
 				Ref<JavaObject> jo = *p_args[i];
 				if (jo.is_valid()) {
-
 					argv[i].l = jo->instance;
 				} else {
 					argv[i].l = NULL; //I hope this works
@@ -289,7 +273,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 
 			} break;
 			case ARG_ARRAY_BIT | ARG_TYPE_BOOLEAN: {
-
 				Array arr = *p_args[i];
 				jbooleanArray a = env->NewBooleanArray(arr.size());
 				for (int j = 0; j < arr.size(); j++) {
@@ -301,7 +284,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 
 			} break;
 			case ARG_ARRAY_BIT | ARG_TYPE_BYTE: {
-
 				Array arr = *p_args[i];
 				jbyteArray a = env->NewByteArray(arr.size());
 				for (int j = 0; j < arr.size(); j++) {
@@ -313,7 +295,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 
 			} break;
 			case ARG_ARRAY_BIT | ARG_TYPE_CHAR: {
-
 				Array arr = *p_args[i];
 				jcharArray a = env->NewCharArray(arr.size());
 				for (int j = 0; j < arr.size(); j++) {
@@ -325,7 +306,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 
 			} break;
 			case ARG_ARRAY_BIT | ARG_TYPE_SHORT: {
-
 				Array arr = *p_args[i];
 				jshortArray a = env->NewShortArray(arr.size());
 				for (int j = 0; j < arr.size(); j++) {
@@ -337,7 +317,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 
 			} break;
 			case ARG_ARRAY_BIT | ARG_TYPE_INT: {
-
 				Array arr = *p_args[i];
 				jintArray a = env->NewIntArray(arr.size());
 				for (int j = 0; j < arr.size(); j++) {
@@ -359,7 +338,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 
 			} break;
 			case ARG_ARRAY_BIT | ARG_TYPE_FLOAT: {
-
 				Array arr = *p_args[i];
 				jfloatArray a = env->NewFloatArray(arr.size());
 				for (int j = 0; j < arr.size(); j++) {
@@ -371,7 +349,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 
 			} break;
 			case ARG_ARRAY_BIT | ARG_TYPE_DOUBLE: {
-
 				Array arr = *p_args[i];
 				jdoubleArray a = env->NewDoubleArray(arr.size());
 				for (int j = 0; j < arr.size(); j++) {
@@ -383,11 +360,9 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 
 			} break;
 			case ARG_ARRAY_BIT | ARG_TYPE_STRING: {
-
 				Array arr = *p_args[i];
 				jobjectArray a = env->NewObjectArray(arr.size(), env->FindClass("java/lang/String"), NULL);
 				for (int j = 0; j < arr.size(); j++) {
-
 					String s = arr[j];
 					jstring jStr = env->NewStringUTF(s.utf8().get_data());
 					env->SetObjectArrayElement(a, j, jStr);
@@ -398,7 +373,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 				to_free.push_back(a);
 			} break;
 			case ARG_ARRAY_BIT | ARG_TYPE_CLASS: {
-
 				argv[i].l = NULL;
 			} break;
 		}
@@ -408,7 +382,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 	bool success = true;
 
 	switch (method->return_type) {
-
 		case ARG_TYPE_VOID: {
 			if (method->_static) {
 				env->CallStaticVoidMethodA(_class, method->method, argv);
@@ -433,7 +406,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 			}
 		} break;
 		case ARG_TYPE_CHAR: {
-
 			if (method->_static) {
 				ret = env->CallStaticCharMethodA(_class, method->method, argv);
 			} else {
@@ -441,7 +413,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 			}
 		} break;
 		case ARG_TYPE_SHORT: {
-
 			if (method->_static) {
 				ret = env->CallStaticShortMethodA(_class, method->method, argv);
 			} else {
@@ -450,7 +421,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 
 		} break;
 		case ARG_TYPE_INT: {
-
 			if (method->_static) {
 				ret = env->CallStaticIntMethodA(_class, method->method, argv);
 			} else {
@@ -459,7 +429,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 
 		} break;
 		case ARG_TYPE_LONG: {
-
 			if (method->_static) {
 				ret = (int64_t)env->CallStaticLongMethodA(_class, method->method, argv);
 			} else {
@@ -468,7 +437,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 
 		} break;
 		case ARG_TYPE_FLOAT: {
-
 			if (method->_static) {
 				ret = env->CallStaticFloatMethodA(_class, method->method, argv);
 			} else {
@@ -477,7 +445,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 
 		} break;
 		case ARG_TYPE_DOUBLE: {
-
 			if (method->_static) {
 				ret = env->CallStaticDoubleMethodA(_class, method->method, argv);
 			} else {
@@ -486,7 +453,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 
 		} break;
 		default: {
-
 			jobject obj;
 			if (method->_static) {
 				obj = env->CallStaticObjectMethodA(_class, method->method, argv);
@@ -497,7 +463,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 			if (!obj) {
 				ret = Variant();
 			} else {
-
 				if (!_convert_object_to_variant(env, obj, ret, method->return_type)) {
 					ret = Variant();
 					r_error.error = Variant::CallError::CALL_ERROR_INVALID_METHOD;
@@ -517,7 +482,6 @@ bool JavaClass::_call_method(JavaObject *p_instance, const StringName &p_method,
 }
 
 Variant JavaClass::call(const StringName &p_method, const Variant **p_args, int p_argcount, Variant::CallError &r_error) {
-
 	Variant ret;
 	bool found = _call_method(NULL, p_method, p_args, p_argcount, r_error, ret);
 	if (found) {
@@ -533,7 +497,6 @@ JavaClass::JavaClass() {
 /////////////////////
 
 Variant JavaObject::call(const StringName &p_method, const Variant **p_args, int p_argcount, Variant::CallError &r_error) {
-
 	return Variant();
 }
 
@@ -545,21 +508,13 @@ JavaObject::~JavaObject() {
 
 ////////////////////
 
-void JavaClassWrapper::_bind_methods() {
-
-	ClassDB::bind_method(D_METHOD("wrap", "name"), &JavaClassWrapper::wrap);
-}
-
 bool JavaClassWrapper::_get_type_sig(JNIEnv *env, jobject obj, uint32_t &sig, String &strsig) {
-
 	jstring name2 = (jstring)env->CallObjectMethod(obj, Class_getName);
-	String str_type = env->GetStringUTFChars(name2, NULL);
-	print_line("name: " + str_type);
+	String str_type = jstring_to_string(name2, env);
 	env->DeleteLocalRef(name2);
 	uint32_t t = 0;
 
 	if (str_type.begins_with("[")) {
-
 		t = JavaClass::ARG_ARRAY_BIT;
 		strsig = "[";
 		str_type = str_type.substr(1, str_type.length() - 1);
@@ -638,87 +593,71 @@ bool JavaClassWrapper::_get_type_sig(JNIEnv *env, jobject obj, uint32_t &sig, St
 }
 
 bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &var, uint32_t p_sig) {
-
 	if (!obj) {
 		var = Variant(); //seems null is just null...
 		return true;
 	}
 
 	switch (p_sig) {
-
 		case ARG_TYPE_VOID: {
-
 			return Variant();
 		} break;
 		case ARG_TYPE_BOOLEAN | ARG_NUMBER_CLASS_BIT: {
-
 			var = env->CallBooleanMethod(obj, JavaClassWrapper::singleton->Boolean_booleanValue);
 			return true;
 		} break;
 		case ARG_TYPE_BYTE | ARG_NUMBER_CLASS_BIT: {
-
 			var = env->CallByteMethod(obj, JavaClassWrapper::singleton->Byte_byteValue);
 			return true;
 
 		} break;
 		case ARG_TYPE_CHAR | ARG_NUMBER_CLASS_BIT: {
-
 			var = env->CallCharMethod(obj, JavaClassWrapper::singleton->Character_characterValue);
 			return true;
 
 		} break;
 		case ARG_TYPE_SHORT | ARG_NUMBER_CLASS_BIT: {
-
 			var = env->CallShortMethod(obj, JavaClassWrapper::singleton->Short_shortValue);
 			return true;
 
 		} break;
 		case ARG_TYPE_INT | ARG_NUMBER_CLASS_BIT: {
-
 			var = env->CallIntMethod(obj, JavaClassWrapper::singleton->Integer_integerValue);
 			return true;
 
 		} break;
 		case ARG_TYPE_LONG | ARG_NUMBER_CLASS_BIT: {
-
 			var = (int64_t)env->CallLongMethod(obj, JavaClassWrapper::singleton->Long_longValue);
 			return true;
 
 		} break;
 		case ARG_TYPE_FLOAT | ARG_NUMBER_CLASS_BIT: {
-
 			var = env->CallFloatMethod(obj, JavaClassWrapper::singleton->Float_floatValue);
 			return true;
 
 		} break;
 		case ARG_TYPE_DOUBLE | ARG_NUMBER_CLASS_BIT: {
-
 			var = env->CallDoubleMethod(obj, JavaClassWrapper::singleton->Double_doubleValue);
 			return true;
 		} break;
 		case ARG_TYPE_STRING: {
-
-			var = String::utf8(env->GetStringUTFChars((jstring)obj, NULL));
+			var = jstring_to_string((jstring)obj, env);
 			return true;
 		} break;
 		case ARG_TYPE_CLASS: {
-
 			return false;
 		} break;
 		case ARG_ARRAY_BIT | ARG_TYPE_VOID: {
-
 			var = Array(); // ?
 			return true;
 		} break;
 		case ARG_ARRAY_BIT | ARG_TYPE_BOOLEAN: {
-
 			Array ret;
 			jobjectArray arr = (jobjectArray)obj;
 
 			int count = env->GetArrayLength(arr);
 
 			for (int i = 0; i < count; i++) {
-
 				jboolean val;
 				env->GetBooleanArrayRegion((jbooleanArray)arr, 0, 1, &val);
 				ret.push_back(val);
@@ -729,14 +668,12 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 
 		} break;
 		case ARG_ARRAY_BIT | ARG_TYPE_BYTE: {
-
 			Array ret;
 			jobjectArray arr = (jobjectArray)obj;
 
 			int count = env->GetArrayLength(arr);
 
 			for (int i = 0; i < count; i++) {
-
 				jbyte val;
 				env->GetByteArrayRegion((jbyteArray)arr, 0, 1, &val);
 				ret.push_back(val);
@@ -752,7 +689,6 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 			int count = env->GetArrayLength(arr);
 
 			for (int i = 0; i < count; i++) {
-
 				jchar val;
 				env->GetCharArrayRegion((jcharArray)arr, 0, 1, &val);
 				ret.push_back(val);
@@ -768,7 +704,6 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 			int count = env->GetArrayLength(arr);
 
 			for (int i = 0; i < count; i++) {
-
 				jshort val;
 				env->GetShortArrayRegion((jshortArray)arr, 0, 1, &val);
 				ret.push_back(val);
@@ -784,7 +719,6 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 			int count = env->GetArrayLength(arr);
 
 			for (int i = 0; i < count; i++) {
-
 				jint val;
 				env->GetIntArrayRegion((jintArray)arr, 0, 1, &val);
 				ret.push_back(val);
@@ -800,7 +734,6 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 			int count = env->GetArrayLength(arr);
 
 			for (int i = 0; i < count; i++) {
-
 				jlong val;
 				env->GetLongArrayRegion((jlongArray)arr, 0, 1, &val);
 				ret.push_back((int64_t)val);
@@ -816,7 +749,6 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 			int count = env->GetArrayLength(arr);
 
 			for (int i = 0; i < count; i++) {
-
 				jfloat val;
 				env->GetFloatArrayRegion((jfloatArray)arr, 0, 1, &val);
 				ret.push_back(val);
@@ -832,7 +764,6 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 			int count = env->GetArrayLength(arr);
 
 			for (int i = 0; i < count; i++) {
-
 				jdouble val;
 				env->GetDoubleArrayRegion((jdoubleArray)arr, 0, 1, &val);
 				ret.push_back(val);
@@ -842,14 +773,12 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 			return true;
 		} break;
 		case ARG_NUMBER_CLASS_BIT | ARG_ARRAY_BIT | ARG_TYPE_BOOLEAN: {
-
 			Array ret;
 			jobjectArray arr = (jobjectArray)obj;
 
 			int count = env->GetArrayLength(arr);
 
 			for (int i = 0; i < count; i++) {
-
 				jobject o = env->GetObjectArrayElement(arr, i);
 				if (!o)
 					ret.push_back(Variant());
@@ -865,14 +794,12 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 
 		} break;
 		case ARG_NUMBER_CLASS_BIT | ARG_ARRAY_BIT | ARG_TYPE_BYTE: {
-
 			Array ret;
 			jobjectArray arr = (jobjectArray)obj;
 
 			int count = env->GetArrayLength(arr);
 
 			for (int i = 0; i < count; i++) {
-
 				jobject o = env->GetObjectArrayElement(arr, i);
 				if (!o)
 					ret.push_back(Variant());
@@ -887,14 +814,12 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 			return true;
 		} break;
 		case ARG_NUMBER_CLASS_BIT | ARG_ARRAY_BIT | ARG_TYPE_CHAR: {
-
 			Array ret;
 			jobjectArray arr = (jobjectArray)obj;
 
 			int count = env->GetArrayLength(arr);
 
 			for (int i = 0; i < count; i++) {
-
 				jobject o = env->GetObjectArrayElement(arr, i);
 				if (!o)
 					ret.push_back(Variant());
@@ -909,14 +834,12 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 			return true;
 		} break;
 		case ARG_NUMBER_CLASS_BIT | ARG_ARRAY_BIT | ARG_TYPE_SHORT: {
-
 			Array ret;
 			jobjectArray arr = (jobjectArray)obj;
 
 			int count = env->GetArrayLength(arr);
 
 			for (int i = 0; i < count; i++) {
-
 				jobject o = env->GetObjectArrayElement(arr, i);
 				if (!o)
 					ret.push_back(Variant());
@@ -931,14 +854,12 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 			return true;
 		} break;
 		case ARG_NUMBER_CLASS_BIT | ARG_ARRAY_BIT | ARG_TYPE_INT: {
-
 			Array ret;
 			jobjectArray arr = (jobjectArray)obj;
 
 			int count = env->GetArrayLength(arr);
 
 			for (int i = 0; i < count; i++) {
-
 				jobject o = env->GetObjectArrayElement(arr, i);
 				if (!o)
 					ret.push_back(Variant());
@@ -953,14 +874,12 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 			return true;
 		} break;
 		case ARG_NUMBER_CLASS_BIT | ARG_ARRAY_BIT | ARG_TYPE_LONG: {
-
 			Array ret;
 			jobjectArray arr = (jobjectArray)obj;
 
 			int count = env->GetArrayLength(arr);
 
 			for (int i = 0; i < count; i++) {
-
 				jobject o = env->GetObjectArrayElement(arr, i);
 				if (!o)
 					ret.push_back(Variant());
@@ -975,14 +894,12 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 			return true;
 		} break;
 		case ARG_NUMBER_CLASS_BIT | ARG_ARRAY_BIT | ARG_TYPE_FLOAT: {
-
 			Array ret;
 			jobjectArray arr = (jobjectArray)obj;
 
 			int count = env->GetArrayLength(arr);
 
 			for (int i = 0; i < count; i++) {
-
 				jobject o = env->GetObjectArrayElement(arr, i);
 				if (!o)
 					ret.push_back(Variant());
@@ -1003,7 +920,6 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 			int count = env->GetArrayLength(arr);
 
 			for (int i = 0; i < count; i++) {
-
 				jobject o = env->GetObjectArrayElement(arr, i);
 				if (!o)
 					ret.push_back(Variant());
@@ -1019,19 +935,17 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 		} break;
 
 		case ARG_ARRAY_BIT | ARG_TYPE_STRING: {
-
 			Array ret;
 			jobjectArray arr = (jobjectArray)obj;
 
 			int count = env->GetArrayLength(arr);
 
 			for (int i = 0; i < count; i++) {
-
 				jobject o = env->GetObjectArrayElement(arr, i);
 				if (!o)
 					ret.push_back(Variant());
 				else {
-					String val = String::utf8(env->GetStringUTFChars((jstring)o, NULL));
+					String val = jstring_to_string((jstring)o, env);
 					ret.push_back(val);
 				}
 				env->DeleteLocalRef(o);
@@ -1041,7 +955,6 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 			return true;
 		} break;
 		case ARG_ARRAY_BIT | ARG_TYPE_CLASS: {
-
 		} break;
 	}
 
@@ -1049,11 +962,11 @@ bool JavaClass::_convert_object_to_variant(JNIEnv *env, jobject obj, Variant &va
 }
 
 Ref<JavaClass> JavaClassWrapper::wrap(const String &p_class) {
-
 	if (class_cache.has(p_class))
 		return class_cache[p_class];
 
-	JNIEnv *env = ThreadAndroid::get_env();
+	JNIEnv *env = get_jni_env();
+	ERR_FAIL_COND_V(env == nullptr, Ref<JavaClass>());
 
 	jclass bclass = env->FindClass(p_class.utf8().get_data());
 	ERR_FAIL_COND_V(!bclass, Ref<JavaClass>());
@@ -1071,12 +984,11 @@ Ref<JavaClass> JavaClassWrapper::wrap(const String &p_class) {
 	int count = env->GetArrayLength(methods);
 
 	for (int i = 0; i < count; i++) {
-
 		jobject obj = env->GetObjectArrayElement(methods, i);
 		ERR_CONTINUE(!obj);
 
 		jstring name = (jstring)env->CallObjectMethod(obj, getName);
-		String str_method = env->GetStringUTFChars(name, NULL);
+		String str_method = jstring_to_string(name, env);
 		env->DeleteLocalRef(name);
 
 		Vector<String> params;
@@ -1101,7 +1013,6 @@ Ref<JavaClass> JavaClassWrapper::wrap(const String &p_class) {
 		String signature = "(";
 
 		for (int j = 0; j < count2; j++) {
-
 			jobject obj2 = env->GetObjectArrayElement(param_types, j);
 			String strsig;
 			uint32_t sig = 0;
@@ -1117,7 +1028,7 @@ Ref<JavaClass> JavaClassWrapper::wrap(const String &p_class) {
 		}
 
 		if (!valid) {
-			print_line("Method Can't be bound (unsupported arguments): " + p_class + "::" + str_method);
+			print_line("Method can't be bound (unsupported arguments): " + p_class + "::" + str_method);
 			env->DeleteLocalRef(obj);
 			env->DeleteLocalRef(param_types);
 			continue;
@@ -1130,7 +1041,7 @@ Ref<JavaClass> JavaClassWrapper::wrap(const String &p_class) {
 		String strsig;
 		uint32_t sig = 0;
 		if (!_get_type_sig(env, return_type, sig, strsig)) {
-			print_line("Method Can't be bound (unsupported return type): " + p_class + "::" + str_method);
+			print_line("Method can't be bound (unsupported return type): " + p_class + "::" + str_method);
 			env->DeleteLocalRef(obj);
 			env->DeleteLocalRef(param_types);
 			env->DeleteLocalRef(return_type);
@@ -1140,12 +1051,9 @@ Ref<JavaClass> JavaClassWrapper::wrap(const String &p_class) {
 		signature += strsig;
 		mi.return_type = sig;
 
-		print_line("METHOD: " + str_method + " SIG: " + signature + " static: " + itos(mi._static));
-
 		bool discard = false;
 
 		for (List<JavaClass::MethodInfo>::Element *E = java_class->methods[str_method].front(); E; E = E->next()) {
-
 			float new_likeliness = 0;
 			float existing_likeliness = 0;
 
@@ -1153,7 +1061,6 @@ Ref<JavaClass> JavaClassWrapper::wrap(const String &p_class) {
 				continue;
 			bool valid = true;
 			for (int j = 0; j < E->get().param_types.size(); j++) {
-
 				Variant::Type _new;
 				float new_l;
 				Variant::Type existing;
@@ -1173,11 +1080,9 @@ Ref<JavaClass> JavaClassWrapper::wrap(const String &p_class) {
 
 			if (new_likeliness > existing_likeliness) {
 				java_class->methods[str_method].erase(E);
-				print_line("replace old");
 				break;
 			} else {
 				discard = true;
-				print_line("old is better");
 			}
 		}
 
@@ -1195,9 +1100,6 @@ Ref<JavaClass> JavaClassWrapper::wrap(const String &p_class) {
 		env->DeleteLocalRef(obj);
 		env->DeleteLocalRef(param_types);
 		env->DeleteLocalRef(return_type);
-
-		//args[i] = _jobject_to_variant(env, obj);
-		//print_line("\targ"+itos(i)+": "+Variant::get_type_name(args[i].get_type()));
 	};
 
 	env->DeleteLocalRef(methods);
@@ -1207,30 +1109,24 @@ Ref<JavaClass> JavaClassWrapper::wrap(const String &p_class) {
 	count = env->GetArrayLength(fields);
 
 	for (int i = 0; i < count; i++) {
-
 		jobject obj = env->GetObjectArrayElement(fields, i);
 		ERR_CONTINUE(!obj);
 
 		jstring name = (jstring)env->CallObjectMethod(obj, Field_getName);
-		String str_field = env->GetStringUTFChars(name, NULL);
+		String str_field = jstring_to_string(name, env);
 		env->DeleteLocalRef(name);
-		print_line("FIELD: " + str_field);
 		int mods = env->CallIntMethod(obj, Field_getModifiers);
 		if ((mods & 0x8) && (mods & 0x10) && (mods & 0x1)) { //static final public!
 
 			jobject objc = env->CallObjectMethod(obj, Field_get, NULL);
 			if (objc) {
-
 				uint32_t sig;
 				String strsig;
 				jclass cl = env->GetObjectClass(objc);
 				if (JavaClassWrapper::_get_type_sig(env, cl, sig, strsig)) {
-
 					if ((sig & JavaClass::ARG_TYPE_MASK) <= JavaClass::ARG_TYPE_STRING) {
-
 						Variant value;
 						if (JavaClass::_convert_object_to_variant(env, objc, value, sig)) {
-
 							java_class->constant_map[str_field] = value;
 						}
 					}
@@ -1252,12 +1148,12 @@ Ref<JavaClass> JavaClassWrapper::wrap(const String &p_class) {
 JavaClassWrapper *JavaClassWrapper::singleton = NULL;
 
 JavaClassWrapper::JavaClassWrapper(jobject p_activity) {
-
 	singleton = this;
 
-	JNIEnv *env = ThreadAndroid::get_env();
+	JNIEnv *env = get_jni_env();
+	ERR_FAIL_COND(env == nullptr);
 
-	jclass activityClass = env->FindClass("org/godotengine/godot/Godot");
+	jclass activityClass = env->FindClass("android/app/Activity");
 	jmethodID getClassLoader = env->GetMethodID(activityClass, "getClassLoader", "()Ljava/lang/ClassLoader;");
 	classLoader = env->CallObjectMethod(p_activity, getClassLoader);
 	classLoader = (jclass)env->NewGlobalRef(classLoader);
