@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -33,7 +33,6 @@
 /***** SCAN CODE CONVERSION ******/
 
 struct _XTranslatePair {
-
 	KeySym keysym;
 	unsigned int keycode;
 };
@@ -180,35 +179,166 @@ static _XTranslatePair _xkeysym_to_keycode[] = {
 	{ 0, 0 }
 };
 
-unsigned int KeyMappingX11::get_keycode(KeySym p_keysym) {
+struct _TranslatePair {
+	unsigned int keysym;
+	unsigned int keycode;
+};
 
+static _TranslatePair _scancode_to_keycode[] = {
+
+	{ KEY_ESCAPE, 0x09 },
+	{ KEY_1, 0x0A },
+	{ KEY_2, 0x0B },
+	{ KEY_3, 0x0C },
+	{ KEY_4, 0x0D },
+	{ KEY_5, 0x0E },
+	{ KEY_6, 0x0F },
+	{ KEY_7, 0x10 },
+	{ KEY_8, 0x11 },
+	{ KEY_9, 0x12 },
+	{ KEY_0, 0x13 },
+	{ KEY_MINUS, 0x14 },
+	{ KEY_EQUAL, 0x15 },
+	{ KEY_BACKSPACE, 0x16 },
+	{ KEY_TAB, 0x17 },
+	{ KEY_Q, 0x18 },
+	{ KEY_W, 0x19 },
+	{ KEY_E, 0x1A },
+	{ KEY_R, 0x1B },
+	{ KEY_T, 0x1C },
+	{ KEY_Y, 0x1D },
+	{ KEY_U, 0x1E },
+	{ KEY_I, 0x1F },
+	{ KEY_O, 0x20 },
+	{ KEY_P, 0x21 },
+	{ KEY_BRACELEFT, 0x22 },
+	{ KEY_BRACERIGHT, 0x23 },
+	{ KEY_ENTER, 0x24 },
+	{ KEY_CONTROL, 0x25 },
+	{ KEY_A, 0x26 },
+	{ KEY_S, 0x27 },
+	{ KEY_D, 0x28 },
+	{ KEY_F, 0x29 },
+	{ KEY_G, 0x2A },
+	{ KEY_H, 0x2B },
+	{ KEY_J, 0x2C },
+	{ KEY_K, 0x2D },
+	{ KEY_L, 0x2E },
+	{ KEY_SEMICOLON, 0x2F },
+	{ KEY_APOSTROPHE, 0x30 },
+	{ KEY_QUOTELEFT, 0x31 },
+	{ KEY_SHIFT, 0x32 },
+	{ KEY_BACKSLASH, 0x33 },
+	{ KEY_Z, 0x34 },
+	{ KEY_X, 0x35 },
+	{ KEY_C, 0x36 },
+	{ KEY_V, 0x37 },
+	{ KEY_B, 0x38 },
+	{ KEY_N, 0x39 },
+	{ KEY_M, 0x3A },
+	{ KEY_COMMA, 0x3B },
+	{ KEY_PERIOD, 0x3C },
+	{ KEY_SLASH, 0x3D },
+	{ KEY_SHIFT, 0x3E },
+	{ KEY_KP_MULTIPLY, 0x3F },
+	{ KEY_ALT, 0x40 },
+	{ KEY_SPACE, 0x41 },
+	{ KEY_CAPSLOCK, 0x42 },
+	{ KEY_F1, 0x43 },
+	{ KEY_F2, 0x44 },
+	{ KEY_F3, 0x45 },
+	{ KEY_F4, 0x46 },
+	{ KEY_F5, 0x47 },
+	{ KEY_F6, 0x48 },
+	{ KEY_F7, 0x49 },
+	{ KEY_F8, 0x4A },
+	{ KEY_F9, 0x4B },
+	{ KEY_F10, 0x4C },
+	{ KEY_NUMLOCK, 0x4D },
+	{ KEY_SCROLLLOCK, 0x4E },
+	{ KEY_KP_7, 0x4F },
+	{ KEY_KP_8, 0x50 },
+	{ KEY_KP_9, 0x51 },
+	{ KEY_KP_SUBTRACT, 0x52 },
+	{ KEY_KP_4, 0x53 },
+	{ KEY_KP_5, 0x54 },
+	{ KEY_KP_6, 0x55 },
+	{ KEY_KP_ADD, 0x56 },
+	{ KEY_KP_1, 0x57 },
+	{ KEY_KP_2, 0x58 },
+	{ KEY_KP_3, 0x59 },
+	{ KEY_KP_0, 0x5A },
+	{ KEY_KP_PERIOD, 0x5B },
+	//{ KEY_???, 0x5E }, //NON US BACKSLASH
+	{ KEY_F11, 0x5F },
+	{ KEY_F12, 0x60 },
+	{ KEY_KP_ENTER, 0x68 },
+	{ KEY_CONTROL, 0x69 },
+	{ KEY_KP_DIVIDE, 0x6A },
+	{ KEY_PRINT, 0x6B },
+	{ KEY_ALT, 0x6C },
+	{ KEY_ENTER, 0x6D },
+	{ KEY_HOME, 0x6E },
+	{ KEY_UP, 0x6F },
+	{ KEY_PAGEUP, 0x70 },
+	{ KEY_LEFT, 0x71 },
+	{ KEY_RIGHT, 0x72 },
+	{ KEY_END, 0x73 },
+	{ KEY_DOWN, 0x74 },
+	{ KEY_PAGEDOWN, 0x75 },
+	{ KEY_INSERT, 0x76 },
+	{ KEY_DELETE, 0x77 },
+	{ KEY_VOLUMEMUTE, 0x79 },
+	{ KEY_VOLUMEDOWN, 0x7A },
+	{ KEY_VOLUMEUP, 0x7B },
+	{ KEY_PAUSE, 0x7F },
+	{ KEY_SUPER_L, 0x85 },
+	{ KEY_SUPER_R, 0x86 },
+	{ KEY_MENU, 0x87 },
+	{ KEY_UNKNOWN, 0 }
+};
+
+unsigned int KeyMappingX11::get_scancode(unsigned int p_code) {
+	unsigned int keycode = KEY_UNKNOWN;
+	for (int i = 0; _scancode_to_keycode[i].keysym != KEY_UNKNOWN; i++) {
+		if (_scancode_to_keycode[i].keycode == p_code) {
+			keycode = _scancode_to_keycode[i].keysym;
+			break;
+		}
+	}
+
+	return keycode;
+}
+
+unsigned int KeyMappingX11::get_keycode(KeySym p_keysym) {
 	// kinda bruteforce.. could optimize.
 
-	if (p_keysym < 0x100) // Latin 1, maps 1-1
+	if (p_keysym < 0x100) { // Latin 1, maps 1-1
 		return p_keysym;
+	}
 
 	// look for special key
 	for (int idx = 0; _xkeysym_to_keycode[idx].keysym != 0; idx++) {
-
-		if (_xkeysym_to_keycode[idx].keysym == p_keysym)
+		if (_xkeysym_to_keycode[idx].keysym == p_keysym) {
 			return _xkeysym_to_keycode[idx].keycode;
+		}
 	}
 
 	return 0;
 }
 
 KeySym KeyMappingX11::get_keysym(unsigned int p_code) {
-
 	// kinda bruteforce.. could optimize.
 
-	if (p_code < 0x100) // Latin 1, maps 1-1
+	if (p_code < 0x100) { // Latin 1, maps 1-1
 		return p_code;
+	}
 
 	// look for special key
 	for (int idx = 0; _xkeysym_to_keycode[idx].keysym != 0; idx++) {
-
-		if (_xkeysym_to_keycode[idx].keycode == p_code)
+		if (_xkeysym_to_keycode[idx].keycode == p_code) {
 			return _xkeysym_to_keycode[idx].keysym;
+		}
 	}
 
 	return 0;
@@ -219,7 +349,6 @@ KeySym KeyMappingX11::get_keysym(unsigned int p_code) {
 // Tables taken from FOX toolkit
 
 struct _XTranslateUnicodePair {
-
 	KeySym keysym;
 	unsigned int unicode;
 };
@@ -991,37 +1120,41 @@ static _XTranslateUnicodePair _xkeysym_to_unicode[_KEYSYM_MAX] = {
 };
 
 unsigned int KeyMappingX11::get_unicode_from_keysym(KeySym p_keysym) {
-
 	/* Latin-1 */
-	if (p_keysym >= 0x20 && p_keysym <= 0x7e)
+	if (p_keysym >= 0x20 && p_keysym <= 0x7e) {
 		return p_keysym;
-	if (p_keysym >= 0xa0 && p_keysym <= 0xff)
+	}
+	if (p_keysym >= 0xa0 && p_keysym <= 0xff) {
 		return p_keysym;
+	}
 	// keypad to latin1 is easy
-	if (p_keysym >= 0xffaa && p_keysym <= 0xffb9)
+	if (p_keysym >= 0xffaa && p_keysym <= 0xffb9) {
 		return p_keysym - 0xff80;
+	}
 
 	/* Unicode (may be present)*/
 
-	if ((p_keysym & 0xff000000) == 0x01000000)
+	if ((p_keysym & 0xff000000) == 0x01000000) {
 		return p_keysym & 0x00ffffff;
+	}
 
 	int middle, low = 0, high = _KEYSYM_MAX - 1;
 	do {
 		middle = (high + low) / 2;
-		if (_xkeysym_to_unicode[middle].keysym == p_keysym)
+		if (_xkeysym_to_unicode[middle].keysym == p_keysym) {
 			return _xkeysym_to_unicode[middle].unicode;
-		if (_xkeysym_to_unicode[middle].keysym <= p_keysym)
+		}
+		if (_xkeysym_to_unicode[middle].keysym <= p_keysym) {
 			low = middle + 1;
-		else
+		} else {
 			high = middle - 1;
+		}
 	} while (high >= low);
 
 	return 0;
 }
 
 struct _XTranslateUnicodePairReverse {
-
 	unsigned int unicode;
 	KeySym keysym;
 };
@@ -1785,24 +1918,27 @@ static _XTranslateUnicodePairReverse _unicode_to_xkeysym[_UNICODE_MAX] = {
 };
 
 KeySym KeyMappingX11::get_keysym_from_unicode(unsigned int p_unicode) {
-
 	/* Latin 1 */
 
-	if (p_unicode >= 0x20 && p_unicode <= 0x7e)
+	if (p_unicode >= 0x20 && p_unicode <= 0x7e) {
 		return p_unicode;
+	}
 
-	if (p_unicode >= 0xa0 && p_unicode <= 0xff)
+	if (p_unicode >= 0xa0 && p_unicode <= 0xff) {
 		return p_unicode;
+	}
 
 	int middle, low = 0, high = _UNICODE_MAX - 1;
 	do {
 		middle = (high + low) / 2;
-		if (_unicode_to_xkeysym[middle].keysym == p_unicode)
+		if (_unicode_to_xkeysym[middle].keysym == p_unicode) {
 			return _unicode_to_xkeysym[middle].keysym;
-		if (_unicode_to_xkeysym[middle].keysym <= p_unicode)
+		}
+		if (_unicode_to_xkeysym[middle].keysym <= p_unicode) {
 			low = middle + 1;
-		else
+		} else {
 			high = middle - 1;
+		}
 	} while (high >= low);
 
 	// if not found, let's hope X understands it as unicode
