@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -29,30 +29,32 @@
 /*************************************************************************/
 
 #include "camera_2d.h"
+
+#include "core/engine.h"
 #include "core/math/math_funcs.h"
 #include "scene/scene_string_names.h"
 #include "servers/visual_server.h"
-#include <editor/editor_node.h>
 
 void Camera2D::_update_scroll() {
-
-	if (!is_inside_tree())
+	if (!is_inside_tree()) {
 		return;
+	}
 
 	if (Engine::get_singleton()->is_editor_hint()) {
 		update(); //will just be drawn
 		return;
 	}
 
-	if (current) {
+	if (!viewport) {
+		return;
+	}
 
+	if (current) {
 		ERR_FAIL_COND(custom_viewport && !ObjectDB::get_instance(custom_viewport_id));
 
 		Transform2D xform = get_camera_transform();
 
-		if (viewport) {
-			viewport->set_canvas_transform(xform);
-		}
+		viewport->set_canvas_transform(xform);
 
 		Size2 screen_size = viewport->get_visible_rect().size;
 		Point2 screen_offset = (anchor_mode == ANCHOR_MODE_DRAG_CENTER ? (screen_size * 0.5) : Point2());
@@ -61,7 +63,41 @@ void Camera2D::_update_scroll() {
 	};
 }
 
+void Camera2D::_update_process_mode() {
+	// smoothing can be enabled in the editor but will never be active
+	if (process_mode == CAMERA2D_PROCESS_IDLE) {
+		set_process_internal(smoothing_active);
+		set_physics_process_internal(false);
+	} else {
+		set_process_internal(false);
+		set_physics_process_internal(smoothing_active);
+	}
+}
+
+void Camera2D::_setup_viewport() {
+	// Disconnect signal on previous viewport if there's one.
+	if (viewport && viewport->is_connected("size_changed", this, "_update_scroll")) {
+		viewport->disconnect("size_changed", this, "_update_scroll");
+	}
+
+	if (custom_viewport && ObjectDB::get_instance(custom_viewport_id)) {
+		viewport = custom_viewport;
+	} else {
+		viewport = get_viewport();
+	}
+
+	RID vp = viewport->get_viewport_rid();
+	group_name = "__cameras_" + itos(vp.get_id());
+	canvas_group_name = "__cameras_c" + itos(canvas.get_id());
+	add_to_group(group_name);
+	add_to_group(canvas_group_name);
+
+	viewport->connect("size_changed", this, "_update_scroll");
+}
+
 void Camera2D::set_zoom(const Vector2 &p_zoom) {
+	// Setting zoom to zero causes 'affine_invert' issues
+	ERR_FAIL_COND_MSG(Math::is_zero_approx(p_zoom.x) || Math::is_zero_approx(p_zoom.y), "Zoom level must be different from 0 (can be negative).");
 
 	zoom = p_zoom;
 	Point2 old_smoothed_camera_pos = smoothed_camera_pos;
@@ -70,14 +106,13 @@ void Camera2D::set_zoom(const Vector2 &p_zoom) {
 };
 
 Vector2 Camera2D::get_zoom() const {
-
 	return zoom;
 };
 
 Transform2D Camera2D::get_camera_transform() {
-
-	if (!get_tree())
+	if (!get_tree() || !viewport) {
 		return Transform2D();
+	}
 
 	ERR_FAIL_COND_V(custom_viewport && !ObjectDB::get_instance(custom_viewport_id), Transform2D());
 
@@ -87,68 +122,64 @@ Transform2D Camera2D::get_camera_transform() {
 	Point2 ret_camera_pos;
 
 	if (!first) {
-
 		if (anchor_mode == ANCHOR_MODE_DRAG_CENTER) {
-
-			if (h_drag_enabled && !Engine::get_singleton()->is_editor_hint()) {
-				camera_pos.x = MIN(camera_pos.x, (new_camera_pos.x + screen_size.x * 0.5 * drag_margin[MARGIN_LEFT]));
-				camera_pos.x = MAX(camera_pos.x, (new_camera_pos.x - screen_size.x * 0.5 * drag_margin[MARGIN_RIGHT]));
+			if (h_drag_enabled && !Engine::get_singleton()->is_editor_hint() && !h_offset_changed) {
+				camera_pos.x = MIN(camera_pos.x, (new_camera_pos.x + screen_size.x * 0.5 * zoom.x * drag_margin[MARGIN_LEFT]));
+				camera_pos.x = MAX(camera_pos.x, (new_camera_pos.x - screen_size.x * 0.5 * zoom.x * drag_margin[MARGIN_RIGHT]));
 			} else {
-
 				if (h_ofs < 0) {
 					camera_pos.x = new_camera_pos.x + screen_size.x * 0.5 * drag_margin[MARGIN_RIGHT] * h_ofs;
 				} else {
 					camera_pos.x = new_camera_pos.x + screen_size.x * 0.5 * drag_margin[MARGIN_LEFT] * h_ofs;
 				}
+
+				h_offset_changed = false;
 			}
 
-			if (v_drag_enabled && !Engine::get_singleton()->is_editor_hint()) {
-
-				camera_pos.y = MIN(camera_pos.y, (new_camera_pos.y + screen_size.y * 0.5 * drag_margin[MARGIN_TOP]));
-				camera_pos.y = MAX(camera_pos.y, (new_camera_pos.y - screen_size.y * 0.5 * drag_margin[MARGIN_BOTTOM]));
+			if (v_drag_enabled && !Engine::get_singleton()->is_editor_hint() && !v_offset_changed) {
+				camera_pos.y = MIN(camera_pos.y, (new_camera_pos.y + screen_size.y * 0.5 * zoom.y * drag_margin[MARGIN_TOP]));
+				camera_pos.y = MAX(camera_pos.y, (new_camera_pos.y - screen_size.y * 0.5 * zoom.y * drag_margin[MARGIN_BOTTOM]));
 
 			} else {
-
 				if (v_ofs < 0) {
-					camera_pos.y = new_camera_pos.y + screen_size.y * 0.5 * drag_margin[MARGIN_TOP] * v_ofs;
-				} else {
 					camera_pos.y = new_camera_pos.y + screen_size.y * 0.5 * drag_margin[MARGIN_BOTTOM] * v_ofs;
+				} else {
+					camera_pos.y = new_camera_pos.y + screen_size.y * 0.5 * drag_margin[MARGIN_TOP] * v_ofs;
 				}
+
+				v_offset_changed = false;
 			}
 
 		} else if (anchor_mode == ANCHOR_MODE_FIXED_TOP_LEFT) {
-
 			camera_pos = new_camera_pos;
 		}
 
 		Point2 screen_offset = (anchor_mode == ANCHOR_MODE_DRAG_CENTER ? (screen_size * 0.5 * zoom) : Point2());
 		Rect2 screen_rect(-screen_offset + camera_pos, screen_size * zoom);
 
-		if (offset != Vector2())
-			screen_rect.position += offset;
-
 		if (limit_smoothing_enabled) {
-			if (screen_rect.position.x < limit[MARGIN_LEFT])
+			if (screen_rect.position.x < limit[MARGIN_LEFT]) {
 				camera_pos.x -= screen_rect.position.x - limit[MARGIN_LEFT];
+			}
 
-			if (screen_rect.position.x + screen_rect.size.x > limit[MARGIN_RIGHT])
+			if (screen_rect.position.x + screen_rect.size.x > limit[MARGIN_RIGHT]) {
 				camera_pos.x -= screen_rect.position.x + screen_rect.size.x - limit[MARGIN_RIGHT];
+			}
 
-			if (screen_rect.position.y + screen_rect.size.y > limit[MARGIN_BOTTOM])
+			if (screen_rect.position.y + screen_rect.size.y > limit[MARGIN_BOTTOM]) {
 				camera_pos.y -= screen_rect.position.y + screen_rect.size.y - limit[MARGIN_BOTTOM];
+			}
 
-			if (screen_rect.position.y < limit[MARGIN_TOP])
+			if (screen_rect.position.y < limit[MARGIN_TOP]) {
 				camera_pos.y -= screen_rect.position.y - limit[MARGIN_TOP];
+			}
 		}
 
-		if (smoothing_enabled && !Engine::get_singleton()->is_editor_hint()) {
-
-			float c = smoothing * get_process_delta_time();
+		if (smoothing_active) {
+			float c = smoothing * (process_mode == CAMERA2D_PROCESS_PHYSICS ? get_physics_process_delta_time() : get_process_delta_time());
 			smoothed_camera_pos = ((camera_pos - smoothed_camera_pos) * c) + smoothed_camera_pos;
 			ret_camera_pos = smoothed_camera_pos;
-			//camera_pos=camera_pos*(1.0-smoothing)+new_camera_pos*smoothing;
 		} else {
-
 			ret_camera_pos = smoothed_camera_pos = camera_pos;
 		}
 
@@ -165,117 +196,94 @@ Transform2D Camera2D::get_camera_transform() {
 	}
 
 	Rect2 screen_rect(-screen_offset + ret_camera_pos, screen_size * zoom);
-	if (screen_rect.position.x < limit[MARGIN_LEFT])
-		screen_rect.position.x = limit[MARGIN_LEFT];
 
-	if (screen_rect.position.x + screen_rect.size.x > limit[MARGIN_RIGHT])
-		screen_rect.position.x = limit[MARGIN_RIGHT] - screen_rect.size.x;
+	if (!smoothing_enabled || !limit_smoothing_enabled) {
+		if (screen_rect.position.x < limit[MARGIN_LEFT]) {
+			screen_rect.position.x = limit[MARGIN_LEFT];
+		}
 
-	if (screen_rect.position.y + screen_rect.size.y > limit[MARGIN_BOTTOM])
-		screen_rect.position.y = limit[MARGIN_BOTTOM] - screen_rect.size.y;
+		if (screen_rect.position.x + screen_rect.size.x > limit[MARGIN_RIGHT]) {
+			screen_rect.position.x = limit[MARGIN_RIGHT] - screen_rect.size.x;
+		}
 
-	if (screen_rect.position.y < limit[MARGIN_TOP])
-		screen_rect.position.y = limit[MARGIN_TOP];
+		if (screen_rect.position.y + screen_rect.size.y > limit[MARGIN_BOTTOM]) {
+			screen_rect.position.y = limit[MARGIN_BOTTOM] - screen_rect.size.y;
+		}
+
+		if (screen_rect.position.y < limit[MARGIN_TOP]) {
+			screen_rect.position.y = limit[MARGIN_TOP];
+		}
+	}
 
 	if (offset != Vector2()) {
-
 		screen_rect.position += offset;
-		if (screen_rect.position.x + screen_rect.size.x > limit[MARGIN_RIGHT])
-			screen_rect.position.x = limit[MARGIN_RIGHT] - screen_rect.size.x;
-
-		if (screen_rect.position.y + screen_rect.size.y > limit[MARGIN_BOTTOM])
-			screen_rect.position.y = limit[MARGIN_BOTTOM] - screen_rect.size.y;
-
-		if (screen_rect.position.x < limit[MARGIN_LEFT])
-			screen_rect.position.x = limit[MARGIN_LEFT];
-
-		if (screen_rect.position.y < limit[MARGIN_TOP])
-			screen_rect.position.y = limit[MARGIN_TOP];
 	}
 
 	camera_screen_center = screen_rect.position + screen_rect.size * 0.5;
 
 	Transform2D xform;
+	xform.scale_basis(zoom);
 	if (rotating) {
 		xform.set_rotation(angle);
 	}
-	xform.scale_basis(zoom);
-	xform.set_origin(screen_rect.position /*.floor()*/);
-
-	/*
-	if (0) {
-		xform = get_global_transform() * xform;
-	} else {
-		xform.elements[2]+=get_global_transform().get_origin();
-	}
-*/
+	xform.set_origin(screen_rect.position);
 
 	return (xform).affine_inverse();
 }
 
 void Camera2D::_notification(int p_what) {
-
 	switch (p_what) {
-
-		case NOTIFICATION_INTERNAL_PROCESS: {
-
+		case NOTIFICATION_INTERNAL_PROCESS:
+		case NOTIFICATION_INTERNAL_PHYSICS_PROCESS: {
 			_update_scroll();
 
 		} break;
 		case NOTIFICATION_TRANSFORM_CHANGED: {
-
-			if (!is_processing_internal())
+			if (!smoothing_active) {
 				_update_scroll();
+			}
 
 		} break;
 		case NOTIFICATION_ENTER_TREE: {
-
-			if (custom_viewport && ObjectDB::get_instance(custom_viewport_id)) {
-				viewport = custom_viewport;
-			} else {
-				viewport = get_viewport();
-			}
-
 			canvas = get_canvas();
 
-			RID vp = viewport->get_viewport_rid();
+			_setup_viewport();
+			_update_process_mode();
 
-			group_name = "__cameras_" + itos(vp.get_id());
-			canvas_group_name = "__cameras_c" + itos(canvas.get_id());
-			add_to_group(group_name);
-			add_to_group(canvas_group_name);
+			// if a camera enters the tree that is set to current,
+			// it should take over as the current camera, and mark
+			// all other cameras as non current
+			_set_current(current);
 
-			if (Engine::get_singleton()->is_editor_hint()) {
-				set_process_internal(false);
-			}
-
-			_update_scroll();
 			first = true;
 
 		} break;
 		case NOTIFICATION_EXIT_TREE: {
-
+			const bool viewport_valid = !custom_viewport || ObjectDB::get_instance(custom_viewport_id);
 			if (is_current()) {
-				if (viewport && !(custom_viewport && !ObjectDB::get_instance(custom_viewport_id))) {
+				if (viewport && viewport_valid) {
 					viewport->set_canvas_transform(Transform2D());
 				}
 			}
+			if (viewport && viewport_valid) {
+				viewport->disconnect("size_changed", this, "_update_scroll");
+			}
 			remove_from_group(group_name);
 			remove_from_group(canvas_group_name);
-			viewport = NULL;
+			viewport = nullptr;
 
 		} break;
+#ifdef TOOLS_ENABLED
 		case NOTIFICATION_DRAW: {
-
-			if (!is_inside_tree() || !Engine::get_singleton()->is_editor_hint())
+			if (!is_inside_tree() || !Engine::get_singleton()->is_editor_hint()) {
 				break;
+			}
 
 			if (screen_drawing_enabled) {
-				Color area_axis_color(0.5, 0.42, 0.87, 0.63);
+				Color area_axis_color(1, 0.4, 1, 0.63);
 				float area_axis_width = 1;
 				if (is_current()) {
 					area_axis_width = 3;
-					area_axis_color.a = 0.83;
 				}
 
 				Transform2D inv_camera_transform = get_camera_transform().affine_inverse();
@@ -296,10 +304,9 @@ void Camera2D::_notification(int p_what) {
 			}
 
 			if (limit_drawing_enabled) {
-				Color limit_drawing_color(1, 1, 0, 0.63);
+				Color limit_drawing_color(1, 1, 0.25, 0.63);
 				float limit_drawing_width = 1;
 				if (is_current()) {
-					limit_drawing_color.a = 0.83;
 					limit_drawing_width = 3;
 				}
 
@@ -318,11 +325,10 @@ void Camera2D::_notification(int p_what) {
 			}
 
 			if (margin_drawing_enabled) {
-				Color margin_drawing_color(0, 1, 1, 0.63);
+				Color margin_drawing_color(0.25, 1, 1, 0.63);
 				float margin_drawing_width = 1;
 				if (is_current()) {
 					margin_drawing_width = 3;
-					margin_drawing_color.a = 0.83;
 				}
 
 				Transform2D inv_camera_transform = get_camera_transform().affine_inverse();
@@ -341,48 +347,57 @@ void Camera2D::_notification(int p_what) {
 					draw_line(inv_transform.xform(margin_endpoints[i]), inv_transform.xform(margin_endpoints[(i + 1) % 4]), margin_drawing_color, margin_drawing_width);
 				}
 			}
-
 		} break;
+#endif
 	}
 }
 
 void Camera2D::set_offset(const Vector2 &p_offset) {
-
 	offset = p_offset;
+	Point2 old_smoothed_camera_pos = smoothed_camera_pos;
 	_update_scroll();
+	smoothed_camera_pos = old_smoothed_camera_pos;
 }
 
 Vector2 Camera2D::get_offset() const {
-
 	return offset;
 }
 
 void Camera2D::set_anchor_mode(AnchorMode p_anchor_mode) {
-
 	anchor_mode = p_anchor_mode;
 	_update_scroll();
 }
 
 Camera2D::AnchorMode Camera2D::get_anchor_mode() const {
-
 	return anchor_mode;
 }
 
 void Camera2D::set_rotating(bool p_rotating) {
-
 	rotating = p_rotating;
+	Point2 old_smoothed_camera_pos = smoothed_camera_pos;
 	_update_scroll();
+	smoothed_camera_pos = old_smoothed_camera_pos;
 }
 
 bool Camera2D::is_rotating() const {
-
 	return rotating;
 }
 
+void Camera2D::set_process_mode(Camera2DProcessMode p_mode) {
+	if (process_mode == p_mode) {
+		return;
+	}
+
+	process_mode = p_mode;
+	_update_process_mode();
+}
+
+Camera2D::Camera2DProcessMode Camera2D::get_process_mode() const {
+	return process_mode;
+}
+
 void Camera2D::_make_current(Object *p_which) {
-
 	if (p_which == this) {
-
 		current = true;
 	} else {
 		current = false;
@@ -390,91 +405,80 @@ void Camera2D::_make_current(Object *p_which) {
 }
 
 void Camera2D::_set_current(bool p_current) {
-
-	if (p_current)
+	if (p_current) {
 		make_current();
+	}
 
 	current = p_current;
 	update();
 }
 
 bool Camera2D::is_current() const {
-
 	return current;
 }
 
 void Camera2D::make_current() {
-
 	if (!is_inside_tree()) {
 		current = true;
 	} else {
 		get_tree()->call_group_flags(SceneTree::GROUP_CALL_REALTIME, group_name, "_make_current", this);
 	}
+	_update_scroll();
 }
 
 void Camera2D::clear_current() {
-
 	current = false;
 	if (is_inside_tree()) {
-		get_tree()->call_group_flags(SceneTree::GROUP_CALL_REALTIME, group_name, "_make_current", (Object *)(NULL));
+		get_tree()->call_group_flags(SceneTree::GROUP_CALL_REALTIME, group_name, "_make_current", (Object *)nullptr);
 	}
 }
 
 void Camera2D::set_limit(Margin p_margin, int p_limit) {
-
-	ERR_FAIL_INDEX(p_margin, 4);
+	ERR_FAIL_INDEX((int)p_margin, 4);
 	limit[p_margin] = p_limit;
 	update();
 }
 
 int Camera2D::get_limit(Margin p_margin) const {
-
-	ERR_FAIL_INDEX_V(p_margin, 4, 0);
+	ERR_FAIL_INDEX_V((int)p_margin, 4, 0);
 	return limit[p_margin];
 }
 
 void Camera2D::set_limit_smoothing_enabled(bool enable) {
-
 	limit_smoothing_enabled = enable;
 	_update_scroll();
 }
 
 bool Camera2D::is_limit_smoothing_enabled() const {
-
 	return limit_smoothing_enabled;
 }
 
 void Camera2D::set_drag_margin(Margin p_margin, float p_drag_margin) {
-
-	ERR_FAIL_INDEX(p_margin, 4);
+	ERR_FAIL_INDEX((int)p_margin, 4);
 	drag_margin[p_margin] = p_drag_margin;
 	update();
 }
 
 float Camera2D::get_drag_margin(Margin p_margin) const {
-
-	ERR_FAIL_INDEX_V(p_margin, 4, 0);
+	ERR_FAIL_INDEX_V((int)p_margin, 4, 0);
 	return drag_margin[p_margin];
 }
 
 Vector2 Camera2D::get_camera_position() const {
-
 	return camera_pos;
 }
 
 void Camera2D::force_update_scroll() {
-
 	_update_scroll();
 }
 
 void Camera2D::reset_smoothing() {
-
-	smoothed_camera_pos = camera_pos;
 	_update_scroll();
+	smoothed_camera_pos = camera_pos;
 }
 
 void Camera2D::align() {
-
+	ERR_FAIL_COND(!is_inside_tree() || !viewport);
 	ERR_FAIL_COND(custom_viewport && !ObjectDB::get_instance(custom_viewport_id));
 
 	Size2 screen_size = viewport->get_visible_rect().size;
@@ -492,7 +496,6 @@ void Camera2D::align() {
 			camera_pos.y = current_camera_pos.y + screen_size.y * 0.5 * drag_margin[MARGIN_BOTTOM] * v_ofs;
 		}
 	} else if (anchor_mode == ANCHOR_MODE_FIXED_TOP_LEFT) {
-
 		camera_pos = current_camera_pos;
 	}
 
@@ -500,80 +503,75 @@ void Camera2D::align() {
 }
 
 void Camera2D::set_follow_smoothing(float p_speed) {
-
 	smoothing = p_speed;
-	if (smoothing > 0 && !(is_inside_tree() && Engine::get_singleton()->is_editor_hint()))
-		set_process_internal(true);
-	else
-		set_process_internal(false);
 }
 
 float Camera2D::get_follow_smoothing() const {
-
 	return smoothing;
 }
 
 Point2 Camera2D::get_camera_screen_center() const {
-
 	return camera_screen_center;
 }
 
 void Camera2D::set_h_drag_enabled(bool p_enabled) {
-
 	h_drag_enabled = p_enabled;
 }
 
 bool Camera2D::is_h_drag_enabled() const {
-
 	return h_drag_enabled;
 }
 
 void Camera2D::set_v_drag_enabled(bool p_enabled) {
-
 	v_drag_enabled = p_enabled;
 }
 
 bool Camera2D::is_v_drag_enabled() const {
-
 	return v_drag_enabled;
 }
 
 void Camera2D::set_v_offset(float p_offset) {
-
 	v_ofs = p_offset;
+	v_offset_changed = true;
+	Point2 old_smoothed_camera_pos = smoothed_camera_pos;
 	_update_scroll();
+	smoothed_camera_pos = old_smoothed_camera_pos;
 }
 
 float Camera2D::get_v_offset() const {
-
 	return v_ofs;
 }
 
 void Camera2D::set_h_offset(float p_offset) {
-
 	h_ofs = p_offset;
+	h_offset_changed = true;
+	Point2 old_smoothed_camera_pos = smoothed_camera_pos;
 	_update_scroll();
+	smoothed_camera_pos = old_smoothed_camera_pos;
 }
 float Camera2D::get_h_offset() const {
-
 	return h_ofs;
 }
 
-void Camera2D::_set_old_smoothing(float p_enable) {
-	//compatibility
-	if (p_enable > 0) {
-		smoothing_enabled = true;
-		set_follow_smoothing(p_enable);
-	}
-}
-
 void Camera2D::set_enable_follow_smoothing(bool p_enabled) {
+	// watch for situation where an pre-enabled camera is added to the tree
+	// processing must be resumed and bypass this noop check
+	// (this currently works but is a possible future bug)
+	if (smoothing_enabled == p_enabled) {
+		return;
+	}
 
+	// Separate the logic between enabled and active, because the smoothing
+	// cannot be active in the editor. This can be done without a separate flag
+	// but is bug prone so this approach is easier to follow.
 	smoothing_enabled = p_enabled;
+	smoothing_active = smoothing_enabled && !Engine::get_singleton()->is_editor_hint();
+
+	// keep the processing up to date after each change
+	_update_process_mode();
 }
 
 bool Camera2D::is_follow_smoothing_enabled() const {
-
 	return smoothing_enabled;
 }
 
@@ -593,28 +591,19 @@ void Camera2D::set_custom_viewport(Node *p_viewport) {
 	}
 
 	if (is_inside_tree()) {
-
-		if (custom_viewport)
-			viewport = custom_viewport;
-		else
-			viewport = get_viewport();
-
-		RID vp = viewport->get_viewport_rid();
-		group_name = "__cameras_" + itos(vp.get_id());
-		canvas_group_name = "__cameras_c" + itos(canvas.get_id());
-		add_to_group(group_name);
-		add_to_group(canvas_group_name);
+		_setup_viewport();
 	}
 }
 
 Node *Camera2D::get_custom_viewport() const {
-
 	return custom_viewport;
 }
 
 void Camera2D::set_screen_drawing_enabled(bool enable) {
 	screen_drawing_enabled = enable;
+#ifdef TOOLS_ENABLED
 	update();
+#endif
 }
 
 bool Camera2D::is_screen_drawing_enabled() const {
@@ -623,7 +612,9 @@ bool Camera2D::is_screen_drawing_enabled() const {
 
 void Camera2D::set_limit_drawing_enabled(bool enable) {
 	limit_drawing_enabled = enable;
+#ifdef TOOLS_ENABLED
 	update();
+#endif
 }
 
 bool Camera2D::is_limit_drawing_enabled() const {
@@ -632,7 +623,9 @@ bool Camera2D::is_limit_drawing_enabled() const {
 
 void Camera2D::set_margin_drawing_enabled(bool enable) {
 	margin_drawing_enabled = enable;
+#ifdef TOOLS_ENABLED
 	update();
+#endif
 }
 
 bool Camera2D::is_margin_drawing_enabled() const {
@@ -640,7 +633,6 @@ bool Camera2D::is_margin_drawing_enabled() const {
 }
 
 void Camera2D::_bind_methods() {
-
 	ClassDB::bind_method(D_METHOD("set_offset", "offset"), &Camera2D::set_offset);
 	ClassDB::bind_method(D_METHOD("get_offset"), &Camera2D::get_offset);
 
@@ -656,6 +648,9 @@ void Camera2D::_bind_methods() {
 
 	ClassDB::bind_method(D_METHOD("_update_scroll"), &Camera2D::_update_scroll);
 
+	ClassDB::bind_method(D_METHOD("set_process_mode", "mode"), &Camera2D::set_process_mode);
+	ClassDB::bind_method(D_METHOD("get_process_mode"), &Camera2D::get_process_mode);
+
 	ClassDB::bind_method(D_METHOD("_set_current", "current"), &Camera2D::_set_current);
 	ClassDB::bind_method(D_METHOD("is_current"), &Camera2D::is_current);
 
@@ -699,8 +694,6 @@ void Camera2D::_bind_methods() {
 	ClassDB::bind_method(D_METHOD("reset_smoothing"), &Camera2D::reset_smoothing);
 	ClassDB::bind_method(D_METHOD("align"), &Camera2D::align);
 
-	ClassDB::bind_method(D_METHOD("_set_old_smoothing", "follow_smoothing"), &Camera2D::_set_old_smoothing);
-
 	ClassDB::bind_method(D_METHOD("set_screen_drawing_enabled", "screen_drawing_enabled"), &Camera2D::set_screen_drawing_enabled);
 	ClassDB::bind_method(D_METHOD("is_screen_drawing_enabled"), &Camera2D::is_screen_drawing_enabled);
 
@@ -710,12 +703,13 @@ void Camera2D::_bind_methods() {
 	ClassDB::bind_method(D_METHOD("set_margin_drawing_enabled", "margin_drawing_enabled"), &Camera2D::set_margin_drawing_enabled);
 	ClassDB::bind_method(D_METHOD("is_margin_drawing_enabled"), &Camera2D::is_margin_drawing_enabled);
 
-	ADD_PROPERTYNZ(PropertyInfo(Variant::VECTOR2, "offset"), "set_offset", "get_offset");
+	ADD_PROPERTY(PropertyInfo(Variant::VECTOR2, "offset"), "set_offset", "get_offset");
 	ADD_PROPERTY(PropertyInfo(Variant::INT, "anchor_mode", PROPERTY_HINT_ENUM, "Fixed TopLeft,Drag Center"), "set_anchor_mode", "get_anchor_mode");
 	ADD_PROPERTY(PropertyInfo(Variant::BOOL, "rotating"), "set_rotating", "is_rotating");
 	ADD_PROPERTY(PropertyInfo(Variant::BOOL, "current"), "_set_current", "is_current");
 	ADD_PROPERTY(PropertyInfo(Variant::VECTOR2, "zoom"), "set_zoom", "get_zoom");
 	ADD_PROPERTY(PropertyInfo(Variant::OBJECT, "custom_viewport", PROPERTY_HINT_RESOURCE_TYPE, "Viewport", 0), "set_custom_viewport", "get_custom_viewport");
+	ADD_PROPERTY(PropertyInfo(Variant::INT, "process_mode", PROPERTY_HINT_ENUM, "Physics,Idle"), "set_process_mode", "get_process_mode");
 
 	ADD_GROUP("Limit", "limit_");
 	ADD_PROPERTYI(PropertyInfo(Variant::INT, "limit_left"), "set_limit", "get_limit", MARGIN_LEFT);
@@ -733,8 +727,8 @@ void Camera2D::_bind_methods() {
 	ADD_PROPERTY(PropertyInfo(Variant::REAL, "smoothing_speed"), "set_follow_smoothing", "get_follow_smoothing");
 
 	ADD_GROUP("Offset", "offset_");
-	ADD_PROPERTY(PropertyInfo(Variant::REAL, "offset_v", PROPERTY_HINT_RANGE, "-1,1,0.01"), "set_v_offset", "get_v_offset");
 	ADD_PROPERTY(PropertyInfo(Variant::REAL, "offset_h", PROPERTY_HINT_RANGE, "-1,1,0.01"), "set_h_offset", "get_h_offset");
+	ADD_PROPERTY(PropertyInfo(Variant::REAL, "offset_v", PROPERTY_HINT_RANGE, "-1,1,0.01"), "set_v_offset", "get_v_offset");
 
 	ADD_GROUP("Drag Margin", "drag_margin_");
 	ADD_PROPERTYI(PropertyInfo(Variant::REAL, "drag_margin_left", PROPERTY_HINT_RANGE, "0,1,0.01"), "set_drag_margin", "get_drag_margin", MARGIN_LEFT);
@@ -749,10 +743,11 @@ void Camera2D::_bind_methods() {
 
 	BIND_ENUM_CONSTANT(ANCHOR_MODE_FIXED_TOP_LEFT);
 	BIND_ENUM_CONSTANT(ANCHOR_MODE_DRAG_CENTER);
+	BIND_ENUM_CONSTANT(CAMERA2D_PROCESS_PHYSICS);
+	BIND_ENUM_CONSTANT(CAMERA2D_PROCESS_IDLE);
 }
 
 Camera2D::Camera2D() {
-
 	anchor_mode = ANCHOR_MODE_DRAG_CENTER;
 	rotating = false;
 	current = false;
@@ -768,9 +763,13 @@ Camera2D::Camera2D() {
 	camera_pos = Vector2();
 	first = true;
 	smoothing_enabled = false;
+	smoothing_active = false;
 	limit_smoothing_enabled = false;
-	custom_viewport = NULL;
+
+	viewport = nullptr;
+	custom_viewport = nullptr;
 	custom_viewport_id = 0;
+	process_mode = CAMERA2D_PROCESS_IDLE;
 
 	smoothing = 5.0;
 	zoom = Vector2(1, 1);
@@ -779,9 +778,12 @@ Camera2D::Camera2D() {
 	limit_drawing_enabled = false;
 	margin_drawing_enabled = false;
 
-	h_drag_enabled = true;
-	v_drag_enabled = true;
+	h_drag_enabled = false;
+	v_drag_enabled = false;
 	h_ofs = 0;
 	v_ofs = 0;
+	h_offset_changed = false;
+	v_offset_changed = false;
+
 	set_notify_transform(true);
 }
