@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -30,8 +30,9 @@
 
 #include "navigation.h"
 
-void Navigation::_navmesh_link(int p_id) {
+#define USE_ENTRY_POINT
 
+void Navigation::_navmesh_link(int p_id) {
 	ERR_FAIL_COND(!navmesh_map.has(p_id));
 	NavMesh &nm = navmesh_map[p_id];
 	ERR_FAIL_COND(nm.linked);
@@ -39,13 +40,13 @@ void Navigation::_navmesh_link(int p_id) {
 
 	PoolVector<Vector3> vertices = nm.navmesh->get_vertices();
 	int len = vertices.size();
-	if (len == 0)
+	if (len == 0) {
 		return;
+	}
 
 	PoolVector<Vector3>::Read r = vertices.read();
 
 	for (int i = 0; i < nm.navmesh->get_polygon_count(); i++) {
-
 		//build
 
 		List<Polygon>::Element *P = nm.polygons.push_back(Polygon());
@@ -62,7 +63,6 @@ void Navigation::_navmesh_link(int p_id) {
 		float sum = 0;
 
 		for (int j = 0; j < plen; j++) {
-
 			int idx = indices[j];
 			if (idx < 0 || idx >= len) {
 				valid = false;
@@ -73,7 +73,7 @@ void Navigation::_navmesh_link(int p_id) {
 			Vector3 ep = nm.xform.xform(r[idx]);
 			center += ep;
 			e.point = _get_point(ep);
-			p.edges[j] = e;
+			p.edges.write[j] = e;
 
 			if (j >= 2) {
 				Vector3 epa = nm.xform.xform(r[indices[j - 2]]);
@@ -88,7 +88,6 @@ void Navigation::_navmesh_link(int p_id) {
 		if (!valid) {
 			nm.polygons.pop_back();
 			ERR_CONTINUE(!valid);
-			continue;
 		}
 
 		p.center = center;
@@ -99,37 +98,32 @@ void Navigation::_navmesh_link(int p_id) {
 		//connect
 
 		for (int j = 0; j < plen; j++) {
-
 			int next = (j + 1) % plen;
 			EdgeKey ek(p.edges[j].point, p.edges[next].point);
 
 			Map<EdgeKey, Connection>::Element *C = connections.find(ek);
 			if (!C) {
-
 				Connection c;
 				c.A = &p;
 				c.A_edge = j;
-				c.B = NULL;
+				c.B = nullptr;
 				c.B_edge = -1;
 				connections[ek] = c;
 			} else {
-
-				if (C->get().B != NULL) {
+				if (C->get().B != nullptr) {
 					ConnectionPending pending;
 					pending.polygon = &p;
 					pending.edge = j;
-					p.edges[j].P = C->get().pending.push_back(pending);
+					p.edges.write[j].P = C->get().pending.push_back(pending);
 					continue;
-					//print_line(String()+_get_vertex(ek.a)+" -> "+_get_vertex(ek.b));
 				}
-				//ERR_CONTINUE(C->get().B!=NULL); //wut
 
 				C->get().B = &p;
 				C->get().B_edge = j;
-				C->get().A->edges[C->get().A_edge].C = &p;
-				C->get().A->edges[C->get().A_edge].C_edge = j;
-				p.edges[j].C = C->get().A;
-				p.edges[j].C_edge = C->get().A_edge;
+				C->get().A->edges.write[C->get().A_edge].C = &p;
+				C->get().A->edges.write[C->get().A_edge].C_edge = j;
+				p.edges.write[j].C = C->get().A;
+				p.edges.write[j].C_edge = C->get().A_edge;
 				//connection successful.
 			}
 		}
@@ -139,21 +133,18 @@ void Navigation::_navmesh_link(int p_id) {
 }
 
 void Navigation::_navmesh_unlink(int p_id) {
-
 	ERR_FAIL_COND(!navmesh_map.has(p_id));
 	NavMesh &nm = navmesh_map[p_id];
 	ERR_FAIL_COND(!nm.linked);
 
 	for (List<Polygon>::Element *E = nm.polygons.front(); E; E = E->next()) {
-
 		Polygon &p = E->get();
 
-		int ec = p.edges.size();
+		int edge_count = p.edges.size();
 		Polygon::Edge *edges = p.edges.ptrw();
 
-		for (int i = 0; i < ec; i++) {
-			int next = (i + 1) % ec;
-
+		for (int i = 0; i < edge_count; i++) {
+			int next = (i + 1) % edge_count;
 			EdgeKey ek(edges[i].point, edges[next].point);
 			Map<EdgeKey, Connection>::Element *C = connections.find(ek);
 
@@ -161,21 +152,20 @@ void Navigation::_navmesh_unlink(int p_id) {
 
 			if (edges[i].P) {
 				C->get().pending.erase(edges[i].P);
-				edges[i].P = NULL;
+				edges[i].P = nullptr;
 			} else if (C->get().B) {
 				//disconnect
 
-				C->get().B->edges[C->get().B_edge].C = NULL;
-				C->get().B->edges[C->get().B_edge].C_edge = -1;
-				C->get().A->edges[C->get().A_edge].C = NULL;
-				C->get().A->edges[C->get().A_edge].C_edge = -1;
+				C->get().B->edges.write[C->get().B_edge].C = nullptr;
+				C->get().B->edges.write[C->get().B_edge].C_edge = -1;
+				C->get().A->edges.write[C->get().A_edge].C = nullptr;
+				C->get().A->edges.write[C->get().A_edge].C_edge = -1;
 
 				if (C->get().A == &E->get()) {
-
 					C->get().A = C->get().B;
 					C->get().A_edge = C->get().B_edge;
 				}
-				C->get().B = NULL;
+				C->get().B = nullptr;
 				C->get().B_edge = -1;
 
 				if (C->get().pending.size()) {
@@ -185,11 +175,11 @@ void Navigation::_navmesh_unlink(int p_id) {
 
 					C->get().B = cp.polygon;
 					C->get().B_edge = cp.edge;
-					C->get().A->edges[C->get().A_edge].C = cp.polygon;
-					C->get().A->edges[C->get().A_edge].C_edge = cp.edge;
-					cp.polygon->edges[cp.edge].C = C->get().A;
-					cp.polygon->edges[cp.edge].C_edge = C->get().A_edge;
-					cp.polygon->edges[cp.edge].P = NULL;
+					C->get().A->edges.write[C->get().A_edge].C = cp.polygon;
+					C->get().A->edges.write[C->get().A_edge].C_edge = cp.edge;
+					cp.polygon->edges.write[cp.edge].C = C->get().A;
+					cp.polygon->edges.write[cp.edge].C_edge = C->get().A_edge;
+					cp.polygon->edges.write[cp.edge].P = nullptr;
 				}
 
 			} else {
@@ -205,7 +195,6 @@ void Navigation::_navmesh_unlink(int p_id) {
 }
 
 int Navigation::navmesh_add(const Ref<NavigationMesh> &p_mesh, const Transform &p_xform, Object *p_owner) {
-
 	int id = last_id++;
 	NavMesh nm;
 	nm.linked = false;
@@ -220,46 +209,47 @@ int Navigation::navmesh_add(const Ref<NavigationMesh> &p_mesh, const Transform &
 }
 
 void Navigation::navmesh_set_transform(int p_id, const Transform &p_xform) {
-
 	ERR_FAIL_COND(!navmesh_map.has(p_id));
 	NavMesh &nm = navmesh_map[p_id];
-	if (nm.xform == p_xform)
+	if (nm.xform == p_xform) {
 		return; //bleh
+	}
 	_navmesh_unlink(p_id);
 	nm.xform = p_xform;
 	_navmesh_link(p_id);
 }
 void Navigation::navmesh_remove(int p_id) {
-
-	ERR_FAIL_COND(!navmesh_map.has(p_id));
+	ERR_FAIL_COND_MSG(!navmesh_map.has(p_id), "Trying to remove nonexisting navmesh with id: " + itos(p_id));
 	_navmesh_unlink(p_id);
 	navmesh_map.erase(p_id);
 }
 
 void Navigation::_clip_path(Vector<Vector3> &path, Polygon *from_poly, const Vector3 &p_to_point, Polygon *p_to_poly) {
-
 	Vector3 from = path[path.size() - 1];
 
-	if (from.distance_to(p_to_point) < CMP_EPSILON)
+	if (from.distance_to(p_to_point) < CMP_EPSILON) {
 		return;
+	}
 	Plane cut_plane;
 	cut_plane.normal = (from - p_to_point).cross(up);
-	if (cut_plane.normal == Vector3())
+	if (cut_plane.normal == Vector3()) {
 		return;
+	}
 	cut_plane.normal.normalize();
 	cut_plane.d = cut_plane.normal.dot(from);
 
 	while (from_poly != p_to_poly) {
-
+		int edge_count = from_poly->edges.size();
+		ERR_FAIL_COND_MSG(edge_count == 0, "Polygon has no edges.");
 		int pe = from_poly->prev_edge;
+		int next = (pe + 1) % edge_count;
 		Vector3 a = _get_vertex(from_poly->edges[pe].point);
-		Vector3 b = _get_vertex(from_poly->edges[(pe + 1) % from_poly->edges.size()].point);
+		Vector3 b = _get_vertex(from_poly->edges[next].point);
 
 		from_poly = from_poly->edges[pe].C;
 		ERR_FAIL_COND(!from_poly);
 
 		if (a.distance_to(b) > CMP_EPSILON) {
-
 			Vector3 inters;
 			if (cut_plane.intersects_segment(a, b, &inters)) {
 				if (inters.distance_to(p_to_point) > CMP_EPSILON && inters.distance_to(path[path.size() - 1]) > CMP_EPSILON) {
@@ -271,23 +261,20 @@ void Navigation::_clip_path(Vector<Vector3> &path, Polygon *from_poly, const Vec
 }
 
 Vector<Vector3> Navigation::get_simple_path(const Vector3 &p_start, const Vector3 &p_end, bool p_optimize) {
-
-	Polygon *begin_poly = NULL;
-	Polygon *end_poly = NULL;
+	Polygon *begin_poly = nullptr;
+	Polygon *end_poly = nullptr;
 	Vector3 begin_point;
 	Vector3 end_point;
 	float begin_d = 1e20;
 	float end_d = 1e20;
 
 	for (Map<int, NavMesh>::Element *E = navmesh_map.front(); E; E = E->next()) {
-
-		if (!E->get().linked)
+		if (!E->get().linked) {
 			continue;
+		}
 		for (List<Polygon>::Element *F = E->get().polygons.front(); F; F = F->next()) {
-
 			Polygon &p = F->get();
 			for (int i = 2; i < p.edges.size(); i++) {
-
 				Face3 f(_get_vertex(p.edges[0].point), _get_vertex(p.edges[i - 1].point), _get_vertex(p.edges[i].point));
 				Vector3 spoint = f.get_closest_point_to(p_start);
 				float dpoint = spoint.distance_to(p_start);
@@ -311,60 +298,62 @@ Vector<Vector3> Navigation::get_simple_path(const Vector3 &p_start, const Vector
 	}
 
 	if (!begin_poly || !end_poly) {
-
-		//print_line("No Path Path");
 		return Vector<Vector3>(); //no path
 	}
 
 	if (begin_poly == end_poly) {
-
 		Vector<Vector3> path;
 		path.resize(2);
-		path[0] = begin_point;
-		path[1] = end_point;
-		//print_line("Direct Path");
+		path.write[0] = begin_point;
+		path.write[1] = end_point;
 		return path;
 	}
 
 	bool found_route = false;
 
 	List<Polygon *> open_list;
+	int begin_edge_count = begin_poly->edges.size();
 
-	for (int i = 0; i < begin_poly->edges.size(); i++) {
-
+	for (int i = 0; i < begin_edge_count; i++) {
 		if (begin_poly->edges[i].C) {
-
 			begin_poly->edges[i].C->prev_edge = begin_poly->edges[i].C_edge;
-			begin_poly->edges[i].C->distance = begin_poly->center.distance_to(begin_poly->edges[i].C->center);
-			open_list.push_back(begin_poly->edges[i].C);
+#ifdef USE_ENTRY_POINT
+			int next = (i + 1) % begin_edge_count;
+			Vector3 edge[2] = {
+				_get_vertex(begin_poly->edges[i].point),
+				_get_vertex(begin_poly->edges[next].point)
+			};
 
-			if (begin_poly->edges[i].C == end_poly) {
-				found_route = true;
-			}
+			Vector3 entry = Geometry::get_closest_point_to_segment(begin_poly->entry, edge);
+			begin_poly->edges[i].C->distance = begin_point.distance_to(entry);
+			begin_poly->edges[i].C->entry = entry;
+#else
+			begin_poly->edges[i].C->distance = begin_poly->center.distance_to(begin_poly->edges[i].C->center);
+#endif
+			open_list.push_back(begin_poly->edges[i].C);
 		}
 	}
 
 	while (!found_route) {
-
 		if (open_list.size() == 0) {
-			//print_line("NOU OPEN LIST");
 			break;
 		}
 		//check open list
 
-		List<Polygon *>::Element *least_cost_poly = NULL;
+		List<Polygon *>::Element *least_cost_poly = nullptr;
 		float least_cost = 1e30;
 
 		//this could be faster (cache previous results)
 		for (List<Polygon *>::Element *E = open_list.front(); E; E = E->next()) {
-
 			Polygon *p = E->get();
 
 			float cost = p->distance;
+#ifdef USE_ENTRY_POINT
+			cost += p->entry.distance_to(end_point);
+#else
 			cost += p->center.distance_to(end_point);
-
+#endif
 			if (cost < least_cost) {
-
 				least_cost_poly = E;
 				least_cost = cost;
 			}
@@ -373,46 +362,59 @@ Vector<Vector3> Navigation::get_simple_path(const Vector3 &p_start, const Vector
 		Polygon *p = least_cost_poly->get();
 		//open the neighbours for search
 
-		for (int i = 0; i < p->edges.size(); i++) {
+		if (p == end_poly) {
+			//oh my reached end! stop algorithm
+			found_route = true;
+			break;
+		}
 
-			Polygon::Edge &e = p->edges[i];
+		int edge_count = p->edges.size();
+		for (int i = 0; i < edge_count; i++) {
+			Polygon::Edge &e = p->edges.write[i];
 
-			if (!e.C)
+			if (!e.C) {
 				continue;
+			}
 
+#ifdef USE_ENTRY_POINT
+			int next = (i + 1) % edge_count;
+			Vector3 edge[2] = {
+				_get_vertex(p->edges[i].point),
+				_get_vertex(p->edges[next].point)
+			};
+
+			Vector3 entry = Geometry::get_closest_point_to_segment(p->entry, edge);
+			float distance = p->entry.distance_to(entry) + p->distance;
+#else
 			float distance = p->center.distance_to(e.C->center) + p->distance;
+#endif
 
 			if (e.C->prev_edge != -1) {
 				//oh this was visited already, can we win the cost?
 
 				if (e.C->distance > distance) {
-
 					e.C->prev_edge = e.C_edge;
 					e.C->distance = distance;
+#ifdef USE_ENTRY_POINT
+					e.C->entry = entry;
+#endif
 				}
 			} else {
 				//add to open neighbours
 
 				e.C->prev_edge = e.C_edge;
 				e.C->distance = distance;
+#ifdef USE_ENTRY_POINT
+				e.C->entry = entry;
+#endif
 				open_list.push_back(e.C);
-
-				if (e.C == end_poly) {
-					//oh my reached end! stop algorithm
-					found_route = true;
-					break;
-				}
 			}
 		}
 
-		if (found_route)
-			break;
-
 		open_list.erase(least_cost_poly);
 	}
 
 	if (found_route) {
-
 		Vector<Vector3> path;
 
 		if (p_optimize) {
@@ -428,7 +430,6 @@ Vector<Vector3> Navigation::get_simple_path(const Vector3 &p_start, const Vector
 			path.push_back(end_point);
 
 			while (p) {
-
 				Vector3 left;
 				Vector3 right;
 
@@ -438,8 +439,10 @@ Vector<Vector3> Navigation::get_simple_path(const Vector3 &p_start, const Vector
 					left = begin_point;
 					right = begin_point;
 				} else {
+					int edge_count = p->edges.size();
+					ERR_FAIL_COND_V_MSG(edge_count == 0, Vector<Vector3>(), "Polygon has no edges.");
 					int prev = p->prev_edge;
-					int prev_n = (p->prev_edge + 1) % p->edges.size();
+					int prev_n = (p->prev_edge + 1) % edge_count;
 					left = _get_vertex(p->edges[prev].point);
 					right = _get_vertex(p->edges[prev_n].point);
 
@@ -457,7 +460,6 @@ Vector<Vector3> Navigation::get_simple_path(const Vector3 &p_start, const Vector
 						left_poly = p;
 						portal_left = left;
 					} else {
-
 						_clip_path(path, apex_poly, portal_right, right_poly);
 
 						apex_point = portal_right;
@@ -477,7 +479,6 @@ Vector<Vector3> Navigation::get_simple_path(const Vector3 &p_start, const Vector
 						right_poly = p;
 						portal_right = right;
 					} else {
-
 						_clip_path(path, apex_poly, portal_left, left_poly);
 
 						apex_point = portal_left;
@@ -490,14 +491,16 @@ Vector<Vector3> Navigation::get_simple_path(const Vector3 &p_start, const Vector
 					}
 				}
 
-				if (p != begin_poly)
+				if (p != begin_poly) {
 					p = p->edges[p->prev_edge].C;
-				else
-					p = NULL;
+				} else {
+					p = nullptr;
+				}
 			}
 
-			if (path[path.size() - 1] != begin_point)
+			if (path[path.size() - 1] != begin_point) {
 				path.push_back(begin_point);
+			}
 
 			path.invert();
 
@@ -508,12 +511,19 @@ Vector<Vector3> Navigation::get_simple_path(const Vector3 &p_start, const Vector
 			path.push_back(end_point);
 			while (true) {
 				int prev = p->prev_edge;
-				int prev_n = (p->prev_edge + 1) % p->edges.size();
+#ifdef USE_ENTRY_POINT
+				Vector3 point = p->entry;
+#else
+				int edge_count = p->edges.size();
+				ERR_FAIL_COND_V_MSG(edge_count == 0, Vector<Vector3>(), "Polygon has no edges.");
+				int prev_n = (p->prev_edge + 1) % edge_count;
 				Vector3 point = (_get_vertex(p->edges[prev].point) + _get_vertex(p->edges[prev_n].point)) * 0.5;
+#endif
 				path.push_back(point);
 				p = p->edges[prev].C;
-				if (p == begin_poly)
+				if (p == begin_poly) {
 					break;
+				}
 			}
 
 			path.push_back(begin_point);
@@ -528,80 +538,62 @@ Vector<Vector3> Navigation::get_simple_path(const Vector3 &p_start, const Vector
 }
 
 Vector3 Navigation::get_closest_point_to_segment(const Vector3 &p_from, const Vector3 &p_to, const bool &p_use_collision) {
-
 	bool use_collision = p_use_collision;
 	Vector3 closest_point;
 	float closest_point_d = 1e20;
-	NavMesh *closest_navmesh = NULL;
 
 	for (Map<int, NavMesh>::Element *E = navmesh_map.front(); E; E = E->next()) {
-
-		if (!E->get().linked)
+		if (!E->get().linked) {
 			continue;
+		}
 		for (List<Polygon>::Element *F = E->get().polygons.front(); F; F = F->next()) {
-
 			Polygon &p = F->get();
 			for (int i = 2; i < p.edges.size(); i++) {
-
 				Face3 f(_get_vertex(p.edges[0].point), _get_vertex(p.edges[i - 1].point), _get_vertex(p.edges[i].point));
 				Vector3 inters;
 				if (f.intersects_segment(p_from, p_to, &inters)) {
-
 					if (!use_collision) {
 						closest_point = inters;
 						use_collision = true;
 						closest_point_d = p_from.distance_to(inters);
-						closest_navmesh = p.owner;
 					} else if (closest_point_d > inters.distance_to(p_from)) {
-
 						closest_point = inters;
 						closest_point_d = p_from.distance_to(inters);
-						closest_navmesh = p.owner;
 					}
 				}
 			}
 
 			if (!use_collision) {
-
-				for (int i = 0; i < p.edges.size(); i++) {
-
+				int edge_count = p.edges.size();
+				for (int i = 0; i < edge_count; i++) {
 					Vector3 a, b;
-
-					Geometry::get_closest_points_between_segments(p_from, p_to, _get_vertex(p.edges[i].point), _get_vertex(p.edges[(i + 1) % p.edges.size()].point), a, b);
+					int next = (i + 1) % edge_count;
+					Geometry::get_closest_points_between_segments(p_from, p_to, _get_vertex(p.edges[i].point), _get_vertex(p.edges[next].point), a, b);
 
 					float d = a.distance_to(b);
 					if (d < closest_point_d) {
-
 						closest_point_d = d;
 						closest_point = b;
-						closest_navmesh = p.owner;
 					}
 				}
 			}
 		}
 	}
 
-	if (closest_navmesh && closest_navmesh->owner) {
-		//print_line("navmesh is: "+Object::cast_to<Node>(closest_navmesh->owner)->get_name());
-	}
-
 	return closest_point;
 }
 
 Vector3 Navigation::get_closest_point(const Vector3 &p_point) {
-
 	Vector3 closest_point;
 	float closest_point_d = 1e20;
 
 	for (Map<int, NavMesh>::Element *E = navmesh_map.front(); E; E = E->next()) {
-
-		if (!E->get().linked)
+		if (!E->get().linked) {
 			continue;
+		}
 		for (List<Polygon>::Element *F = E->get().polygons.front(); F; F = F->next()) {
-
 			Polygon &p = F->get();
 			for (int i = 2; i < p.edges.size(); i++) {
-
 				Face3 f(_get_vertex(p.edges[0].point), _get_vertex(p.edges[i - 1].point), _get_vertex(p.edges[i].point));
 				Vector3 inters = f.get_closest_point_to(p_point);
 				float d = inters.distance_to(p_point);
@@ -617,20 +609,17 @@ Vector3 Navigation::get_closest_point(const Vector3 &p_point) {
 }
 
 Vector3 Navigation::get_closest_point_normal(const Vector3 &p_point) {
-
 	Vector3 closest_point;
 	Vector3 closest_normal;
 	float closest_point_d = 1e20;
 
 	for (Map<int, NavMesh>::Element *E = navmesh_map.front(); E; E = E->next()) {
-
-		if (!E->get().linked)
+		if (!E->get().linked) {
 			continue;
+		}
 		for (List<Polygon>::Element *F = E->get().polygons.front(); F; F = F->next()) {
-
 			Polygon &p = F->get();
 			for (int i = 2; i < p.edges.size(); i++) {
-
 				Face3 f(_get_vertex(p.edges[0].point), _get_vertex(p.edges[i - 1].point), _get_vertex(p.edges[i].point));
 				Vector3 inters = f.get_closest_point_to(p_point);
 				float d = inters.distance_to(p_point);
@@ -647,20 +636,17 @@ Vector3 Navigation::get_closest_point_normal(const Vector3 &p_point) {
 }
 
 Object *Navigation::get_closest_point_owner(const Vector3 &p_point) {
-
 	Vector3 closest_point;
-	Object *owner = NULL;
+	Object *owner = nullptr;
 	float closest_point_d = 1e20;
 
 	for (Map<int, NavMesh>::Element *E = navmesh_map.front(); E; E = E->next()) {
-
-		if (!E->get().linked)
+		if (!E->get().linked) {
 			continue;
+		}
 		for (List<Polygon>::Element *F = E->get().polygons.front(); F; F = F->next()) {
-
 			Polygon &p = F->get();
 			for (int i = 2; i < p.edges.size(); i++) {
-
 				Face3 f(_get_vertex(p.edges[0].point), _get_vertex(p.edges[i - 1].point), _get_vertex(p.edges[i].point));
 				Vector3 inters = f.get_closest_point_to(p_point);
 				float d = inters.distance_to(p_point);
@@ -677,17 +663,14 @@ Object *Navigation::get_closest_point_owner(const Vector3 &p_point) {
 }
 
 void Navigation::set_up_vector(const Vector3 &p_up) {
-
 	up = p_up;
 }
 
 Vector3 Navigation::get_up_vector() const {
-
 	return up;
 }
 
 void Navigation::_bind_methods() {
-
 	ClassDB::bind_method(D_METHOD("navmesh_add", "mesh", "xform", "owner"), &Navigation::navmesh_add, DEFVAL(Variant()));
 	ClassDB::bind_method(D_METHOD("navmesh_set_transform", "id", "xform"), &Navigation::navmesh_set_transform);
 	ClassDB::bind_method(D_METHOD("navmesh_remove", "id"), &Navigation::navmesh_remove);
@@ -705,7 +688,6 @@ void Navigation::_bind_methods() {
 }
 
 Navigation::Navigation() {
-
 	ERR_FAIL_COND(sizeof(Point) != 8);
 	cell_size = 0.01; //one centimeter
 	last_id = 1;
