@@ -5,8 +5,8 @@
 /*                           GODOT ENGINE                                */
 /*                      https://godotengine.org                          */
 /*************************************************************************/
-/* Copyright (c) 2007-2018 Juan Linietsky, Ariel Manzur.                 */
-/* Copyright (c) 2014-2018 Godot Engine contributors (cf. AUTHORS.md)    */
+/* Copyright (c) 2007-2022 Juan Linietsky, Ariel Manzur.                 */
+/* Copyright (c) 2014-2022 Godot Engine contributors (cf. AUTHORS.md).   */
 /*                                                                       */
 /* Permission is hereby granted, free of charge, to any person obtaining */
 /* a copy of this software and associated documentation files (the       */
@@ -29,44 +29,46 @@
 /*************************************************************************/
 
 #include "label.h"
-#include "print_string.h"
-#include "project_settings.h"
-#include "translation.h"
+#include "core/print_string.h"
+#include "core/project_settings.h"
+#include "core/translation.h"
 
 void Label::set_autowrap(bool p_autowrap) {
+	if (autowrap == p_autowrap) {
+		return;
+	}
 
 	autowrap = p_autowrap;
 	word_cache_dirty = true;
 	update();
+
+	if (clip) {
+		minimum_size_changed();
+	}
 }
 bool Label::has_autowrap() const {
-
 	return autowrap;
 }
 
 void Label::set_uppercase(bool p_uppercase) {
-
 	uppercase = p_uppercase;
 	word_cache_dirty = true;
 	update();
 }
 bool Label::is_uppercase() const {
-
 	return uppercase;
 }
 
 int Label::get_line_height() const {
-
 	return get_font("font")->get_height();
 }
 
 void Label::_notification(int p_what) {
-
 	if (p_what == NOTIFICATION_TRANSLATION_CHANGED) {
-
 		String new_text = tr(text);
-		if (new_text == xl_text)
+		if (new_text == xl_text) {
 			return; //nothing new
+		}
 		xl_text = new_text;
 
 		regenerate_word_cache();
@@ -74,13 +76,13 @@ void Label::_notification(int p_what) {
 	}
 
 	if (p_what == NOTIFICATION_DRAW) {
-
-		if (clip || autowrap) {
+		if (clip) {
 			VisualServer::get_singleton()->canvas_item_set_clip(get_canvas_item(), true);
 		}
 
-		if (word_cache_dirty)
+		if (word_cache_dirty) {
 			regenerate_word_cache();
+		}
 
 		RID ci = get_canvas_item();
 
@@ -93,6 +95,7 @@ void Label::_notification(int p_what) {
 		bool use_outline = get_constant("shadow_as_outline");
 		Point2 shadow_ofs(get_constant("shadow_offset_x"), get_constant("shadow_offset_y"));
 		int line_spacing = get_constant("line_spacing");
+		Color font_outline_modulate = get_color("font_outline_modulate");
 
 		style->draw(ci, Rect2(Point2(0, 0), get_size()));
 
@@ -102,7 +105,7 @@ void Label::_notification(int p_what) {
 
 		int lines_visible = (size.y + line_spacing) / font_h;
 
-		int space_w = font->get_char_size(' ').width;
+		real_t space_w = font->get_char_size(' ').width;
 		int chars_total = 0;
 
 		int vbegin = 0, vsep = 0;
@@ -115,27 +118,29 @@ void Label::_notification(int p_what) {
 			lines_visible = max_lines_visible;
 		}
 
+		float total_h = 0;
+		total_h += lines_visible * font_h;
+		total_h += style->get_margin(MARGIN_TOP) + style->get_margin(MARGIN_BOTTOM);
+
 		if (lines_visible > 0) {
-
 			switch (valign) {
-
 				case VALIGN_TOP: {
 					//nothing
 				} break;
 				case VALIGN_CENTER: {
-					vbegin = (size.y - (lines_visible * font_h - line_spacing)) / 2;
+					vbegin = (size.y - (total_h - line_spacing)) / 2;
 					vsep = 0;
 
 				} break;
 				case VALIGN_BOTTOM: {
-					vbegin = size.y - (lines_visible * font_h - line_spacing);
+					vbegin = size.y - (total_h - line_spacing);
 					vsep = 0;
 
 				} break;
 				case VALIGN_FILL: {
 					vbegin = 0;
 					if (lines_visible > 1) {
-						vsep = (size.y - (lines_visible * font_h - line_spacing)) / (lines_visible - 1);
+						vsep = (size.y - (total_h - line_spacing)) / (lines_visible - 1);
 					} else {
 						vsep = 0;
 					}
@@ -145,21 +150,25 @@ void Label::_notification(int p_what) {
 		}
 
 		WordCache *wc = word_cache;
-		if (!wc)
+		if (!wc) {
 			return;
+		}
 
 		int line = 0;
 		int line_to = lines_skipped + (lines_visible > 0 ? lines_visible : 1);
+		FontDrawer drawer(font, font_outline_modulate);
 		while (wc) {
 			/* handle lines not meant to be drawn quickly */
-			if (line >= line_to)
+			if (line >= line_to) {
 				break;
+			}
 			if (line < lines_skipped) {
-
-				while (wc && wc->char_pos >= 0)
+				while (wc && wc->char_pos >= 0) {
 					wc = wc->next;
-				if (wc)
+				}
+				if (wc) {
 					wc = wc->next;
+				}
 				line++;
 				continue;
 			}
@@ -179,9 +188,8 @@ void Label::_notification(int p_what) {
 			int taken = 0;
 			int spaces = 0;
 			while (to && to->char_pos >= 0) {
-
 				taken += to->pixel_width;
-				if (to != from && to->space_count) {
+				if (to->space_count) {
 					spaces += to->space_count;
 				}
 				to = to->next;
@@ -192,18 +200,14 @@ void Label::_notification(int p_what) {
 			float x_ofs = 0;
 
 			switch (align) {
-
 				case ALIGN_FILL:
 				case ALIGN_LEFT: {
-
 					x_ofs = style->get_offset().x;
 				} break;
 				case ALIGN_CENTER: {
-
 					x_ofs = int(size.width - (taken + spaces * space_w)) / 2;
 				} break;
 				case ALIGN_RIGHT: {
-
 					x_ofs = int(size.width - style->get_margin(MARGIN_RIGHT) - (taken + spaces * space_w));
 				} break;
 			}
@@ -213,11 +217,9 @@ void Label::_notification(int p_what) {
 			y_ofs += vbegin + line * vsep;
 
 			while (from != to) {
-
 				// draw a word
 				int pos = from->char_pos;
 				if (from->char_pos < 0) {
-
 					ERR_PRINT("BUG");
 					return;
 				}
@@ -225,30 +227,27 @@ void Label::_notification(int p_what) {
 					/* spacing */
 					x_ofs += space_w * from->space_count;
 					if (can_fill && align == ALIGN_FILL && spaces) {
-
 						x_ofs += int((size.width - (taken + space_w * spaces)) / spaces);
 					}
 				}
 
 				if (font_color_shadow.a > 0) {
-
 					int chars_total_shadow = chars_total; //save chars drawn
 					float x_ofs_shadow = x_ofs;
 					for (int i = 0; i < from->word_len; i++) {
-
 						if (visible_chars < 0 || chars_total_shadow < visible_chars) {
 							CharType c = xl_text[i + pos];
 							CharType n = xl_text[i + pos + 1];
 							if (uppercase) {
 								c = String::char_uppercase(c);
-								n = String::char_uppercase(c);
+								n = String::char_uppercase(n);
 							}
 
-							float move = font->draw_char(ci, Point2(x_ofs_shadow, y_ofs) + shadow_ofs, c, n, font_color_shadow);
+							float move = drawer.draw_char(ci, Point2(x_ofs_shadow, y_ofs) + shadow_ofs, c, n, font_color_shadow);
 							if (use_outline) {
-								font->draw_char(ci, Point2(x_ofs_shadow, y_ofs) + Vector2(-shadow_ofs.x, shadow_ofs.y), c, n, font_color_shadow);
-								font->draw_char(ci, Point2(x_ofs_shadow, y_ofs) + Vector2(shadow_ofs.x, -shadow_ofs.y), c, n, font_color_shadow);
-								font->draw_char(ci, Point2(x_ofs_shadow, y_ofs) + Vector2(-shadow_ofs.x, -shadow_ofs.y), c, n, font_color_shadow);
+								drawer.draw_char(ci, Point2(x_ofs_shadow, y_ofs) + Vector2(-shadow_ofs.x, shadow_ofs.y), c, n, font_color_shadow);
+								drawer.draw_char(ci, Point2(x_ofs_shadow, y_ofs) + Vector2(shadow_ofs.x, -shadow_ofs.y), c, n, font_color_shadow);
+								drawer.draw_char(ci, Point2(x_ofs_shadow, y_ofs) + Vector2(-shadow_ofs.x, -shadow_ofs.y), c, n, font_color_shadow);
 							}
 							x_ofs_shadow += move;
 							chars_total_shadow++;
@@ -256,156 +255,163 @@ void Label::_notification(int p_what) {
 					}
 				}
 				for (int i = 0; i < from->word_len; i++) {
-
 					if (visible_chars < 0 || chars_total < visible_chars) {
 						CharType c = xl_text[i + pos];
 						CharType n = xl_text[i + pos + 1];
 						if (uppercase) {
 							c = String::char_uppercase(c);
-							n = String::char_uppercase(c);
+							n = String::char_uppercase(n);
 						}
 
-						x_ofs += font->draw_char(ci, Point2(x_ofs, y_ofs), c, n, font_color);
+						x_ofs += drawer.draw_char(ci, Point2(x_ofs, y_ofs), c, n, font_color);
 						chars_total++;
 					}
 				}
 				from = from->next;
 			}
 
-			wc = to ? to->next : 0;
+			wc = to ? to->next : nullptr;
 			line++;
 		}
 	}
 
 	if (p_what == NOTIFICATION_THEME_CHANGED) {
-
 		word_cache_dirty = true;
 		update();
 	}
 	if (p_what == NOTIFICATION_RESIZED) {
-
 		word_cache_dirty = true;
 	}
 }
 
 Size2 Label::get_minimum_size() const {
-
 	Size2 min_style = get_stylebox("normal")->get_minimum_size();
 
-	if (autowrap)
+	// don't want to mutable everything
+	if (word_cache_dirty) {
+		const_cast<Label *>(this)->regenerate_word_cache();
+	}
+
+	if (autowrap) {
 		return Size2(1, clip ? 1 : minsize.height) + min_style;
-	else {
-
-		// don't want to mutable everything
-		if (word_cache_dirty)
-			const_cast<Label *>(this)->regenerate_word_cache();
-
+	} else {
 		Size2 ms = minsize;
-		if (clip)
+		if (clip) {
 			ms.width = 1;
+		}
 		return ms + min_style;
 	}
 }
 
 int Label::get_longest_line_width() const {
-
 	Ref<Font> font = get_font("font");
-	int max_line_width = 0;
-	int line_width = 0;
+	real_t max_line_width = 0;
+	real_t line_width = 0;
 
 	for (int i = 0; i < xl_text.size(); i++) {
-
 		CharType current = xl_text[i];
-		if (uppercase)
+		if (uppercase) {
 			current = String::char_uppercase(current);
+		}
 
 		if (current < 32) {
-
 			if (current == '\n') {
-
-				if (line_width > max_line_width)
+				if (line_width > max_line_width) {
 					max_line_width = line_width;
+				}
 				line_width = 0;
 			}
 		} else {
-
-			int char_width = font->get_char_size(current, xl_text[i + 1]).width;
+			real_t char_width = font->get_char_size(current, xl_text[i + 1]).width;
 			line_width += char_width;
 		}
 	}
 
-	if (line_width > max_line_width)
+	if (line_width > max_line_width) {
 		max_line_width = line_width;
+	}
 
-	return max_line_width;
+	// ceiling to ensure autowrapping does not cut text
+	return Math::ceil(max_line_width);
 }
 
 int Label::get_line_count() const {
-
-	if (!is_inside_tree())
+	if (!is_inside_tree()) {
 		return 1;
-	if (word_cache_dirty)
+	}
+	if (word_cache_dirty) {
 		const_cast<Label *>(this)->regenerate_word_cache();
+	}
 
 	return line_count;
 }
 
 int Label::get_visible_line_count() const {
-
 	int line_spacing = get_constant("line_spacing");
 	int font_h = get_font("font")->get_height() + line_spacing;
 	int lines_visible = (get_size().height - get_stylebox("normal")->get_minimum_size().height + line_spacing) / font_h;
 
-	if (lines_visible > line_count)
+	if (lines_visible > line_count) {
 		lines_visible = line_count;
+	}
 
-	if (max_lines_visible >= 0 && lines_visible > max_lines_visible)
+	if (max_lines_visible >= 0 && lines_visible > max_lines_visible) {
 		lines_visible = max_lines_visible;
+	}
 
 	return lines_visible;
 }
 
 void Label::regenerate_word_cache() {
-
 	while (word_cache) {
-
 		WordCache *current = word_cache;
 		word_cache = current->next;
 		memdelete(current);
 	}
 
-	Ref<StyleBox> style = get_stylebox("normal");
-	int width = autowrap ? (get_size().width - style->get_minimum_size().width) : get_longest_line_width();
+	int width;
+	if (autowrap) {
+		Ref<StyleBox> style = get_stylebox("normal");
+		width = MAX(get_size().width, get_custom_minimum_size().width) - style->get_minimum_size().width;
+	} else {
+		width = get_longest_line_width();
+	}
+
 	Ref<Font> font = get_font("font");
 
-	int current_word_size = 0;
+	real_t current_word_size = 0;
 	int word_pos = 0;
-	int line_width = 0;
+	real_t line_width = 0;
 	int space_count = 0;
-	int space_width = font->get_char_size(' ').width;
+	real_t space_width = font->get_char_size(' ').width;
 	int line_spacing = get_constant("line_spacing");
 	line_count = 1;
 	total_char_cache = 0;
 
-	WordCache *last = NULL;
+	WordCache *last = nullptr;
 
-	for (int i = 0; i < xl_text.size() + 1; i++) {
+	for (int i = 0; i <= xl_text.length(); i++) {
+		CharType current = i < xl_text.length() ? xl_text[i] : L' '; //always a space at the end, so the algo works
 
-		CharType current = i < xl_text.length() ? xl_text[i] : ' '; //always a space at the end, so the algo works
-
-		if (uppercase)
+		if (uppercase) {
 			current = String::char_uppercase(current);
+		}
 
-		// ranges taken from http://www.unicodemap.org/
+		// ranges taken from https://en.wikipedia.org/wiki/Plane_(Unicode)
 		// if your language is not well supported, consider helping improve
 		// the unicode support in Godot.
-		bool separatable = (current >= 0x2E08 && current <= 0xFAFF) || (current >= 0xFE30 && current <= 0xFE4F);
-		//current>=33 && (current < 65||current >90) && (current<97||current>122) && (current<48||current>57);
+		bool separatable = (current >= 0x2E08 && current <= 0x9FFF) || // CJK scripts and symbols.
+				(current >= 0xAC00 && current <= 0xD7FF) || // Hangul Syllables and Hangul Jamo Extended-B.
+				(current >= 0xF900 && current <= 0xFAFF) || // CJK Compatibility Ideographs.
+				(current >= 0xFE30 && current <= 0xFE4F) || // CJK Compatibility Forms.
+				(current >= 0xFF65 && current <= 0xFF9F) || // Halfwidth forms of katakana
+				(current >= 0xFFA0 && current <= 0xFFDC) || // Halfwidth forms of compatibility jamo characters for Hangul
+				(current >= 0x20000 && current <= 0x2FA1F) || // CJK Unified Ideographs Extension B ~ F and CJK Compatibility Ideographs Supplement.
+				(current >= 0x30000 && current <= 0x3134F); // CJK Unified Ideographs Extension G.
 		bool insert_newline = false;
-		int char_width = 0;
+		real_t char_width = 0;
 
 		if (current < 33) {
-
 			if (current_word_size > 0) {
 				WordCache *wc = memnew(WordCache);
 				if (word_cache) {
@@ -421,17 +427,32 @@ void Label::regenerate_word_cache() {
 				wc->space_count = space_count;
 				current_word_size = 0;
 				space_count = 0;
+			} else if ((i == xl_text.length() || current == '\n') && last != nullptr && space_count != 0) {
+				//in case there are trailing white spaces we add a placeholder word cache with just the spaces
+				WordCache *wc = memnew(WordCache);
+				if (word_cache) {
+					last->next = wc;
+				} else {
+					word_cache = wc;
+				}
+				last = wc;
+
+				wc->pixel_width = 0;
+				wc->char_pos = 0;
+				wc->word_len = 0;
+				wc->space_count = space_count;
+				current_word_size = 0;
+				space_count = 0;
 			}
 
 			if (current == '\n') {
 				insert_newline = true;
-			} else {
+			} else if (current != ' ') {
 				total_char_cache++;
 			}
 
 			if (i < xl_text.length() && xl_text[i] == ' ') {
-				total_char_cache--; // do not count spaces
-				if (line_width > 0 || last == NULL || last->char_pos != WordCache::CHAR_WRAPLINE) {
+				if (line_width > 0 || last == nullptr || last->char_pos != WordCache::CHAR_WRAPLINE) {
 					space_count++;
 					line_width += space_width;
 				} else {
@@ -444,11 +465,15 @@ void Label::regenerate_word_cache() {
 			if (current_word_size == 0) {
 				word_pos = i;
 			}
-
 			char_width = font->get_char_size(current, xl_text[i + 1]).width;
 			current_word_size += char_width;
 			line_width += char_width;
 			total_char_cache++;
+
+			// allow autowrap to cut words when they exceed line width
+			if (autowrap && (current_word_size > width)) {
+				separatable = true;
+			}
 		}
 
 		if ((autowrap && (line_width >= width) && ((last && last->char_pos >= 0) || separatable)) || insert_newline) {
@@ -467,7 +492,6 @@ void Label::regenerate_word_cache() {
 					wc->word_len = i - word_pos;
 					wc->space_count = space_count;
 					current_word_size = char_width;
-					space_count = 0;
 					word_pos = i;
 				}
 			}
@@ -489,8 +513,9 @@ void Label::regenerate_word_cache() {
 		}
 	}
 
-	if (!autowrap)
+	if (!autowrap) {
 		minsize.width = width;
+	}
 
 	if (max_lines_visible > 0 && line_count > max_lines_visible) {
 		minsize.height = (font->get_height() * max_lines_visible) + (line_spacing * (max_lines_visible - 1));
@@ -506,60 +531,53 @@ void Label::regenerate_word_cache() {
 }
 
 void Label::set_align(Align p_align) {
-
-	ERR_FAIL_INDEX(p_align, 4);
+	ERR_FAIL_INDEX((int)p_align, 4);
 	align = p_align;
 	update();
 }
 
 Label::Align Label::get_align() const {
-
 	return align;
 }
 
 void Label::set_valign(VAlign p_align) {
-
-	ERR_FAIL_INDEX(p_align, 4);
+	ERR_FAIL_INDEX((int)p_align, 4);
 	valign = p_align;
 	update();
 }
 
 Label::VAlign Label::get_valign() const {
-
 	return valign;
 }
 
 void Label::set_text(const String &p_string) {
-
-	if (text == p_string)
+	if (text == p_string) {
 		return;
+	}
 	text = p_string;
 	xl_text = tr(p_string);
 	word_cache_dirty = true;
-	if (percent_visible < 1)
+	if (percent_visible < 1) {
 		visible_chars = get_total_character_count() * percent_visible;
+	}
 	update();
 }
 
 void Label::set_clip_text(bool p_clip) {
-
 	clip = p_clip;
 	update();
 	minimum_size_changed();
 }
 
 bool Label::is_clipping_text() const {
-
 	return clip;
 }
 
 String Label::get_text() const {
-
 	return text;
 }
 
 void Label::set_visible_characters(int p_amount) {
-
 	visible_chars = p_amount;
 	if (get_total_character_count() > 0) {
 		percent_visible = (float)p_amount / (float)total_char_cache;
@@ -569,19 +587,15 @@ void Label::set_visible_characters(int p_amount) {
 }
 
 int Label::get_visible_characters() const {
-
 	return visible_chars;
 }
 
 void Label::set_percent_visible(float p_percent) {
-
 	if (p_percent < 0 || p_percent >= 1) {
-
 		visible_chars = -1;
 		percent_visible = 1;
 
 	} else {
-
 		visible_chars = get_total_character_count() * p_percent;
 		percent_visible = p_percent;
 	}
@@ -590,42 +604,37 @@ void Label::set_percent_visible(float p_percent) {
 }
 
 float Label::get_percent_visible() const {
-
 	return percent_visible;
 }
 
 void Label::set_lines_skipped(int p_lines) {
-
+	ERR_FAIL_COND(p_lines < 0);
 	lines_skipped = p_lines;
 	update();
 }
 
 int Label::get_lines_skipped() const {
-
 	return lines_skipped;
 }
 
 void Label::set_max_lines_visible(int p_lines) {
-
 	max_lines_visible = p_lines;
 	update();
 }
 
 int Label::get_max_lines_visible() const {
-
 	return max_lines_visible;
 }
 
 int Label::get_total_character_count() const {
-
-	if (word_cache_dirty)
+	if (word_cache_dirty) {
 		const_cast<Label *>(this)->regenerate_word_cache();
+	}
 
 	return total_char_cache;
 }
 
 void Label::_bind_methods() {
-
 	ClassDB::bind_method(D_METHOD("set_align", "align"), &Label::set_align);
 	ClassDB::bind_method(D_METHOD("get_align"), &Label::get_align);
 	ClassDB::bind_method(D_METHOD("set_valign", "valign"), &Label::set_valign);
@@ -661,12 +670,12 @@ void Label::_bind_methods() {
 	BIND_ENUM_CONSTANT(VALIGN_BOTTOM);
 	BIND_ENUM_CONSTANT(VALIGN_FILL);
 
-	ADD_PROPERTYNZ(PropertyInfo(Variant::STRING, "text", PROPERTY_HINT_MULTILINE_TEXT, "", PROPERTY_USAGE_DEFAULT_INTL), "set_text", "get_text");
-	ADD_PROPERTYNZ(PropertyInfo(Variant::INT, "align", PROPERTY_HINT_ENUM, "Left,Center,Right,Fill"), "set_align", "get_align");
-	ADD_PROPERTYNZ(PropertyInfo(Variant::INT, "valign", PROPERTY_HINT_ENUM, "Top,Center,Bottom,Fill"), "set_valign", "get_valign");
-	ADD_PROPERTYNZ(PropertyInfo(Variant::BOOL, "autowrap"), "set_autowrap", "has_autowrap");
-	ADD_PROPERTYNZ(PropertyInfo(Variant::BOOL, "clip_text"), "set_clip_text", "is_clipping_text");
-	ADD_PROPERTYNZ(PropertyInfo(Variant::BOOL, "uppercase"), "set_uppercase", "is_uppercase");
+	ADD_PROPERTY(PropertyInfo(Variant::STRING, "text", PROPERTY_HINT_MULTILINE_TEXT, "", PROPERTY_USAGE_DEFAULT_INTL), "set_text", "get_text");
+	ADD_PROPERTY(PropertyInfo(Variant::INT, "align", PROPERTY_HINT_ENUM, "Left,Center,Right,Fill"), "set_align", "get_align");
+	ADD_PROPERTY(PropertyInfo(Variant::INT, "valign", PROPERTY_HINT_ENUM, "Top,Center,Bottom,Fill"), "set_valign", "get_valign");
+	ADD_PROPERTY(PropertyInfo(Variant::BOOL, "autowrap"), "set_autowrap", "has_autowrap");
+	ADD_PROPERTY(PropertyInfo(Variant::BOOL, "clip_text"), "set_clip_text", "is_clipping_text");
+	ADD_PROPERTY(PropertyInfo(Variant::BOOL, "uppercase"), "set_uppercase", "is_uppercase");
 	ADD_PROPERTY(PropertyInfo(Variant::INT, "visible_characters", PROPERTY_HINT_RANGE, "-1,128000,1", PROPERTY_USAGE_EDITOR), "set_visible_characters", "get_visible_characters");
 	ADD_PROPERTY(PropertyInfo(Variant::REAL, "percent_visible", PROPERTY_HINT_RANGE, "0,1,0.001"), "set_percent_visible", "get_percent_visible");
 	ADD_PROPERTY(PropertyInfo(Variant::INT, "lines_skipped", PROPERTY_HINT_RANGE, "0,999,1"), "set_lines_skipped", "get_lines_skipped");
@@ -674,11 +683,10 @@ void Label::_bind_methods() {
 }
 
 Label::Label(const String &p_text) {
-
 	align = ALIGN_LEFT;
 	valign = VALIGN_TOP;
 	xl_text = "";
-	word_cache = NULL;
+	word_cache = nullptr;
 	word_cache_dirty = true;
 	autowrap = false;
 	line_count = 0;
@@ -696,9 +704,7 @@ Label::Label(const String &p_text) {
 }
 
 Label::~Label() {
-
 	while (word_cache) {
-
 		WordCache *current = word_cache;
 		word_cache = current->next;
 		memdelete(current);
