@@ -1,7 +1,9 @@
 short_name = "godot"
 name = "Godot Engine"
 major = 3
-minor = 0
-patch = 6
+minor = 4
+patch = 4
 status = "stable"
 module_config = ""
+year = 2022
+website = "https://godotengine.org"
